/*                                                                            *
 *   This file is part of the ESO X-shooter Pipeline                          *
 *   Copyright (C) 2006 European Southern Observatory                         *
 *                                                                            *
 *   This library is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU General Public License as published by     *
 *   the Free Software Foundation; either version 2 of the License, or        *
 *   (at your option) any later version.                                      *
 *                                                                            *
 *   This program is distributed in the hope that it will be useful,          *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *   GNU General Public License for more details.                             *
 *                                                                            *
 *   You should have received a copy of the GNU General Public License        *
 *   along with this program; if not, write to the Free Software              *
 *   Foundation, 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA     *
 *                                                                            */

/*
 * $Author: amodigli $
 * $Date: 2011/03/07 09:45:10 $
 * $Revision: 1.93 $
 * $Name: HEAD $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*---------------------------------------------------------------------------*/
/**
 * @defgroup xsh_mbias    xsh_mbias
 * @ingroup recipes
 *
 * This recipe calculates the master bias frame
 * See man-page for details.
 */
/*---------------------------------------------------------------------------*/
/**@{*/

/*----------------------------------------------------------------------------
                                Includes
 ----------------------------------------------------------------------------*/


/* DRL steps */

/* Error handling */
#include <xsh_error.h>
/* Utility fonctions */
#include <xsh_utils.h>
#include <xsh_msg.h>
#include <xsh_data_instrument.h>
#include <math.h>
/* DFS functions */
#include <xsh_dfs.h>
/* DRL functions */
#include <xsh_drl.h>
#include <xsh_pfits.h>
#include <xsh_detmon.h>
#include <xsh_irplib_utils.h>
#include <xsh_paf_save.h>
#include <xsh_utils_image.h>
#include <xsh_parameters.h>
#include <xsh_drl_check.h>

/* Library */
#include <cpl.h>
#include <assert.h>
/*----------------------------------------------------------------------------
                                Defines
 ----------------------------------------------------------------------------*/

#define RECIPE_ID "xsh_mbias"
#define RECIPE_AUTHOR "P.Goldoni, L.Guglielmi, R. Haigron, F. Royer, A. Modigliani"
#define RECIPE_CONTACT "amodigli@eso.org"

/*----------------------------------------------------------------------------
                            Functions prototypes
 ----------------------------------------------------------------------------*/

/*
 *   Plugin initalization, execute and cleanup handlers
 */

static int xsh_mbias_create(cpl_plugin *);
static int xsh_mbias_exec(cpl_plugin *);
static int xsh_mbias_destroy(cpl_plugin *);

/* The actual executor function */
static void xsh_mbias(cpl_parameterlist *, cpl_frameset *);

/*----------------------------------------------------------------------------
                            Static variables
 ----------------------------------------------------------------------------*/
static char xsh_mbias_description_short[] = "Create the master bias frame";

static char xsh_mbias_description[] =
    "This recipe creates a master bias frame by computing the median of all \
input bias frames.\n\
Input Frames : \n\
  - A set of n RAW frames (Format=RAW, n >=3, Tag = BIAS_arm)\n\
  - [OPTIONAL] A badpixel map (Format=QUP, Tag = MASTER_BP_MAP_arm)\n\
  - [OPTIONAL] A map on non linear pixels (Format=QUP,RAW, Tag = BP_MAP_NL_arm)\n\
Products : \n\
  - A master bias frame (Format=PRE, PRO.CATG = MASTER_BIAS_arm)\n\
  - A cold pixel map frame (Format=PRE, PRO.CATG = BP_MAP_CP_arm)\n\
  - A hot pixel map frame (Format=PRE, PRO.CATG = BP_MAP_HP_arm)\n\
  - [if given in input] the master bp map is updated frame (Format=PRE, PRO.CATG = MASTER_BP_MAP_arm)\n";

/*----------------------------------------------------------------------------
                              Functions code
 ----------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module.
  @param    list    the plugin list
  @return   0 if everything is ok, -1 otherwise

  Create the recipe instance and make it available to the application using 
  the interface. This function is exported.
 */
/*---------------------------------------------------------------------------*/

int cpl_plugin_get_info(cpl_pluginlist * list)
{
    cpl_recipe *recipe = NULL;
    cpl_plugin *plugin = NULL;

    recipe = cpl_calloc(1, sizeof(*recipe));
    if (recipe == NULL) {
	return -1;
    }

    plugin = &recipe->interface;

    cpl_plugin_init(plugin, CPL_PLUGIN_API,	/* Plugin API */
		    XSH_BINARY_VERSION,	/* Plugin version */
		    CPL_PLUGIN_TYPE_RECIPE,	/* Plugin type */
		    RECIPE_ID,	/* Plugin name */
		    xsh_mbias_description_short,	/* Short help */
		    xsh_mbias_description,	/* Detailed help */
		    RECIPE_AUTHOR,	/* Author name */
		    RECIPE_CONTACT,	/* Contact address */
		    xsh_get_license(),	/* Copyright */
		    xsh_mbias_create, 
                    xsh_mbias_exec, 
                    xsh_mbias_destroy);

    cpl_pluginlist_append(list, plugin);

    return (cpl_error_get_code() != CPL_ERROR_NONE);
}

/*---------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options
  @param    plugin  the plugin
  @return   0 if everything is ok

  Create the recipe instance and make it available to the application using 
  the interface.

 */
/*--------------------------------------------------------------------------*/

static int xsh_mbias_create(cpl_plugin * plugin)
{
  cpl_recipe *recipe = NULL;
  cpl_parameter   * p =NULL;
  xsh_clipping_param crh_clip_param = {8.0, 2, 0.7, 0, 0.3};
  xsh_hot_cold_pix_param hp_clip_param = {1, 5.0, 3, 5.0 , 3};
  xsh_fpn_param fpn_param = {10,10,1024,1024,10,100};
  xsh_ron_param ron_param = {"ALL",
                             10,100,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,3,25};

  xsh_struct_param struct_param = {-1,-1};

  /* Reset library state */
  xsh_init();

  /* Check input */
  assure(plugin != NULL, CPL_ERROR_NULL_INPUT, "Null plugin");

  /* Get the recipe out of the plugin */
  assure(cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE,
    CPL_ERROR_TYPE_MISMATCH, "Plugin is not a recipe");

  recipe = (cpl_recipe *) plugin;

  /* Create the parameter list in the cpl_recipe object */
  recipe->parameters = cpl_parameterlist_new();
  assure(recipe->parameters != NULL,
    CPL_ERROR_ILLEGAL_OUTPUT, "Memory allocation failed!");

  /* Set generic parameters (common to all recipes) */
  check( xsh_parameters_generic( RECIPE_ID, recipe->parameters ) ) ;

  p = cpl_parameter_new_enum("xsh.xsh_mbias.method",
			     CPL_TYPE_STRING,
			     "Method for master bias generation: "
			     "[median] (+CRH clip) or (clean) mean",
			     "xsh.xsh_mbias", 
			     "median", 2, "median", "mean");

  cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"method");
  cpl_parameterlist_append(recipe->parameters,p);

  /* crh clipping params */
  check(xsh_parameters_clipping_crh_create(RECIPE_ID,recipe->parameters,
    crh_clip_param));

  check(xsh_parameters_hot_cold_pix_create(RECIPE_ID,recipe->parameters,
                                           hp_clip_param));


  check(xsh_parameters_fpn_create(RECIPE_ID,recipe->parameters,fpn_param));
  check(xsh_parameters_ron_create(RECIPE_ID,recipe->parameters,ron_param));
  check(xsh_parameters_struct_create(RECIPE_ID,recipe->parameters,struct_param));


  cleanup:
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
	xsh_error_dump(CPL_MSG_ERROR);
	return 1;
    } else {
	return 0;
    }
}

/*---------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*---------------------------------------------------------------------------*/

static int xsh_mbias_exec(cpl_plugin * plugin)
{
    cpl_recipe *recipe = NULL;

    /* Check parameter */
    assure(plugin != NULL, CPL_ERROR_NULL_INPUT, "Null plugin");

    /* Get the recipe out of the plugin */
    assure(cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE,
	   CPL_ERROR_TYPE_MISMATCH, "Plugin is not a recipe");

    recipe = (cpl_recipe *) plugin;

    /* Check recipe */
    xsh_mbias(recipe->parameters, recipe->frames);

  cleanup:
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
	xsh_error_dump(CPL_MSG_ERROR);
        cpl_error_reset();
	return 1;
    } else {
	return 0;
    }
}

/*---------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*---------------------------------------------------------------------------*/
static int xsh_mbias_destroy(cpl_plugin * plugin)
{
    cpl_recipe *recipe = NULL;

    xsh_error_reset();
    /* Check parameter */
    assure(plugin != NULL, CPL_ERROR_NULL_INPUT, "Null plugin");

    /* Get the recipe out of the plugin */
    assure(cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE,
	   CPL_ERROR_TYPE_MISMATCH, "Plugin is not a recipe");

    recipe = (cpl_recipe *) plugin;

    xsh_free_parameterlist(&recipe->parameters);

  cleanup:
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
	return 1;
    } else {
	return 0;
    }
}


/*---------------------------------------------------------------------------*/
/**
  @brief    Interpret the command line options and execute the data processing
  @param    parameters     the parameters list
  @param    frameset   the frames list

  In case of failure the cpl_error_code is set.
 */
/*---------------------------------------------------------------------------*/
static void xsh_mbias(cpl_parameterlist * parameters, cpl_frameset * frameset)
{
  const char* recipe_tags[1] = {XSH_BIAS};
  int recipe_tags_size = 1;

  cpl_frameset *raws = NULL;
  cpl_frameset *calib = NULL;
  cpl_frame *bpmap = NULL;
  cpl_frame *master_bias = NULL;
  cpl_frame *product = NULL;
  xsh_instrument* instrument  = NULL;
  xsh_clipping_param* crh_clipping = NULL;
  cpl_imagelist* crh_list=NULL;
  cpl_frame* crh_frm=NULL;
  
  /*
  //RON
  cpl_image *mbias = NULL;
  cpl_imagelist* rawbiases=NULL;
  cpl_propertylist* qclist=NULL;
  */
  char* crh_name=NULL;
  //const char* crh_pro_catg=NULL;
  char prefix[80];

  double ks_low=0;
  double ks_high=0;

  //double fpn=0;
  cpl_parameter* p=NULL;
  int binx=0;
  int biny=0;
  cpl_frame* bias_frm=NULL;
  cpl_propertylist* plist=NULL;
  const char* method=NULL;
  char name[256];
  const char* ftag;
  int find_hot_cold_pix=0;

  int cold_niter=3;
  int hot_niter=3;
  cpl_frame* cpix_frm=NULL;
  cpl_frame* hpix_frm=NULL;
  cpl_image* ima=NULL;
  cpl_propertylist* head=NULL;
  cpl_frame* coadd_bp_map=NULL;
  const char* pro_catg=NULL;
  int pre_overscan_corr=0;

  //Initial check on critical parameters
  check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mbias.cold-pix-kappa"));
  check(ks_low = cpl_parameter_get_double(p));

  if(ks_low<2.0) {
    xsh_msg_error("You set %s too low: %f! ",
		  cpl_parameter_get_name(p),cpl_parameter_get_double(p));
    xsh_msg("Should be at least 2 for proper kappa-sigma clipping");
    goto cleanup;
  }

  check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mbias.hot-pix-kappa"));
  check(ks_high = cpl_parameter_get_double(p));

  if(ks_high<2.0) {
    xsh_msg_error("You set %s too low: %f! ",
		  cpl_parameter_get_name(p),cpl_parameter_get_double(p));
    xsh_msg("Should be at least 2 for proper kappa-sigma clipping");
    goto cleanup;
  }

  check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mbias.hot-cold-pix-search"));
  check(find_hot_cold_pix = cpl_parameter_get_bool(p));

  /**************************************************************************/
  /* DFS management */
  /**************************************************************************/
  check( xsh_begin( frameset, parameters, &instrument, &raws, &calib,
                    recipe_tags, recipe_tags_size,
                    RECIPE_ID, XSH_BINARY_VERSION,
                    xsh_mbias_description_short ) ) ;
  /**************************************************************************/
  /* Recipe frames */
  /**************************************************************************/
  //AModigliani: why should be prevent to reduce < 3 frames if it works?
  //XSH_ASSURE_NOT_ILLEGAL(cpl_frameset_get_size(raws) >= 3);
  check(bpmap=xsh_check_load_master_bpmap(calib,instrument));

  check(bias_frm=cpl_frameset_get_first(raws));
  check(plist=cpl_propertylist_load(cpl_frame_get_filename(bias_frm),0));
  check(binx=xsh_pfits_get_binx(plist));
  check(biny=xsh_pfits_get_biny(plist));
  xsh_free_propertylist(&plist);


  /* Sabine does not like binning in filename:
     sprintf(name,"BIAS_REMOVE_CRH_%dx%d.fits",binx,biny); */
  sprintf(name,"%s.fits",XSH_BIAS_REMOVE_CRH);
  ftag=XSH_GET_TAG_FROM_ARM(XSH_BIAS_REMOVE_CRH,instrument);


  /**************************************************************************/
  /* Recipe parameters */
  /**************************************************************************/
  check(crh_clipping = xsh_parameters_clipping_crh_get(RECIPE_ID,
    parameters));

  /**************************************************************************/
  /* Recipe code */
  /**************************************************************************/

  /* prepare RAW frames in XSH format */
  check(xsh_prepare(raws, bpmap, NULL, XSH_BIAS, instrument,pre_overscan_corr));

  /* remove cosmic rays and merge the RAWS */
  //crh_list=cpl_imagelist_new();
  check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mbias.method"));
  check(method = cpl_parameter_get_string(p));
  
  if(strcmp(method,"mean") == 0) {
     check(master_bias = xsh_remove_crh_multiple(raws,
                                                 ftag,
                                                 crh_clipping, 
                                                 instrument,
                                                 &crh_list,
                                                 NULL)); 
  } else {
     check(master_bias = xsh_create_master_bias2(raws,instrument,ftag)); 
  }

  /* create master bias */
  check(product=xsh_create_master_bias(raws,master_bias,instrument,parameters));

  /* RON computation (as detector group does) 
  check(nsampl=xsh_parameters_get_int(parameters,RECIPE_ID,"random_nsamples"));
  check(hsize=xsh_parameters_get_int(parameters,RECIPE_ID,"random_sizex"));
  check(llx=xsh_parameters_get_int(parameters,RECIPE_ID,"ref1_llx"));
  check(lly=xsh_parameters_get_int(parameters,RECIPE_ID,"ref1_lly"));
  check(urx=xsh_parameters_get_int(parameters,RECIPE_ID,"ref1_urx"));
  check(ury=xsh_parameters_get_int(parameters,RECIPE_ID,"ref1_ury"));

  check(xsh_compute_ron(raws,llx,lly,urx,ury,nsampl,hsize,1,&ron,&ron_err));
  xsh_msg("RON DIFF method region 1: val=%g stdev=%g",ron,ron_err);


  check(llx=xsh_parameters_get_int(parameters,RECIPE_ID,"ref2_llx"));
  check(lly=xsh_parameters_get_int(parameters,RECIPE_ID,"ref2_lly"));
  check(urx=xsh_parameters_get_int(parameters,RECIPE_ID,"ref2_urx"));
  check(ury=xsh_parameters_get_int(parameters,RECIPE_ID,"ref2_ury"));

  check(xsh_compute_ron(raws,llx,lly,urx,ury,nsampl,hsize,2,&ron,&ron_err));
  xsh_msg("RON DIFF method region 2: val=%g stdev=%g",ron,ron_err);
  */

  /**************************************************************************/
  /* Products */
  /**************************************************************************/

  xsh_msg("Save products");

  check( xsh_add_product_pre(product, frameset, parameters, RECIPE_ID,
    instrument));
  
  if(find_hot_cold_pix) {
    check(p=cpl_parameterlist_find(parameters,"xsh.xsh_mbias.cold-pix-niter"));
    check(cold_niter = cpl_parameter_get_int(p));

    check(p=cpl_parameterlist_find(parameters,"xsh.xsh_mbias.hot-pix-niter"));
    check(hot_niter = cpl_parameter_get_int(p));





    check(xsh_image_get_hot_cold_pixs(master_bias,instrument,
				      ks_low,cold_niter,ks_high,hot_niter,
				      &cpix_frm,&hpix_frm));



    sprintf(prefix, "%s_%s",XSH_BP_MAP_CP,
            xsh_instrument_arm_tostring(instrument) ) ;


    check(xsh_add_product_image(cpix_frm, frameset,parameters, 
				RECIPE_ID, instrument, prefix ));


    sprintf(prefix, "%s_%s", XSH_BP_MAP_HP,
            xsh_instrument_arm_tostring(instrument) ) ;


    check(xsh_add_product_image(hpix_frm, frameset,parameters, 
				RECIPE_ID, instrument, prefix ));

  
  }

  if(bpmap!=NULL) {


    xsh_msg( "Coadd non-lin and static bp map frames" ) ;
    check(coadd_bp_map=cpl_frame_duplicate(bpmap));
    sprintf(name,cpl_frame_get_filename(bpmap));
    check(ima=cpl_image_load(name,CPL_TYPE_FLOAT,0,0));
    check(head=cpl_propertylist_load(name,0));
    check(pro_catg=XSH_GET_TAG_FROM_ARM(XSH_MASTER_BP_MAP,instrument));
    sprintf(name,"%s.fits",pro_catg);
    check(cpl_image_save(ima,name,CPL_BPP_IEEE_FLOAT,head,CPL_IO_DEFAULT));
    check(cpl_frame_set_filename(coadd_bp_map,name));
    check(cpl_frame_set_tag(coadd_bp_map,pro_catg));
    check(xsh_badpixelmap_coadd(coadd_bp_map,hpix_frm));
    check( xsh_add_product_bpmap(coadd_bp_map, frameset, parameters,
				 RECIPE_ID, instrument, pro_catg ));
    xsh_free_frame(&coadd_bp_map);
    xsh_free_image(&ima);
    xsh_free_propertylist(&head);

  }



  /* Commented out: this frame is useless unless for debugging
  //CRH list: is this useful to anything?
  check(crh_name=cpl_sprintf("%s_%s_mbias_%dx%d.fits",XSH_CRH_LIST,
				    xsh_instrument_arm_tostring(instrument),
			            binx,biny)) ;

  cpl_imagelist_save(crh_list,crh_name,CPL_BPP_IEEE_FLOAT,
  	       NULL,CPL_IO_DEFAULT);
  xsh_add_temporary_file(crh_name);

  xsh_free_imagelist(&crh_list)
  check(crh_pro_catg=XSH_GET_TAG_FROM_ARM(XSH_CRH_LIST,instrument));


  check(crh_frm=xsh_frame_product(crh_name,crh_pro_catg,
                                   CPL_FRAME_TYPE_IMAGE,
                                   CPL_FRAME_GROUP_PRODUCT,
                                   CPL_FRAME_LEVEL_INTERMEDIATE));
 

  XSH_PREFIX(prefix,XSH_CRH_LIST,instrument);
  check(xsh_add_product_imagelist(crh_frm, frameset,parameters, 
			      RECIPE_ID, instrument, prefix ));
 
  */


  xsh_msg("xsh_mbias success!!"); 

  cleanup:
    xsh_end( RECIPE_ID, frameset, parameters );
    XSH_FREE(crh_clipping);
    xsh_free_propertylist(&plist);
    xsh_instrument_free(&instrument);
    xsh_free_frame(&product);
    xsh_free_frameset(&raws);
    xsh_free_frameset(&calib);
    xsh_free_frame(&master_bias);
    xsh_free_imagelist(&crh_list);
    XSH_FREE(crh_name);

    xsh_free_frame(&crh_frm);
    xsh_free_frame(&cpix_frm);
    xsh_free_frame(&hpix_frm);
    return;
}

/**@}*/


