/*                                                                            *
 *   This file is part of the ESO X-shooter Pipeline                          *
 *   Copyright (C) 2006 European Southern Observatory                         *
 *                                                                            *
 *   This library is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU General Public License as published by     *
 *   the Free Software Foundation; either version 2 of the License, or        *
 *   (at your option) any later version.                                      *
 *                                                                            *
 *   This program is distributed in the hope that it will be useful,          *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *   GNU General Public License for more details.                             *
 *                                                                            *
 *   You should have received a copy of the GNU General Public License        *
 *   along with this program; if not, write to the Free Software              *
 *   Foundation, 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA     *
 *                                                                            */

/*
 * $Author: amodigli $
 * $Date: 2011/03/04 12:59:41 $
 * $Revision: 1.86 $
 *
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*----------------------------------------------------------------------------*/
/**
 * @defgroup xsh_scired_slit_offset   xsh_scired_slit_offset
 * @ingroup recipes
 *
 * This recipe ...
 * See man-page for details.
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*-----------------------------------------------------------------------------
  Includes
  ----------------------------------------------------------------------------*/

/* DRL steps */

/* Error handling */
#include <xsh_error.h>
/* Utility fonctions */
#include <xsh_utils.h>
#include <xsh_utils_scired_slit.h>

#include <xsh_msg.h>
/* DFS functions */
#include <xsh_dfs.h>
#include <xsh_pfits.h>
#include <xsh_data_spectrum1D.h>
/* DRL functions */
#include <xsh_drl.h>
/* Library */
#include <cpl.h>

/*-----------------------------------------------------------------------------
  Defines
  ---------------------------------------------------------------------------*/

#define RECIPE_ID "xsh_scired_slit_offset"
#define RECIPE_AUTHOR "P.Goldoni, L.Guglielmi, R. Haigron, F. Royer"
#define RECIPE_CONTACT "amodigli@eso.org"

/*-----------------------------------------------------------------------------
  Functions prototypes
  ----------------------------------------------------------------------------*/

/*
 *   Plugin initalization, execute and cleanup handlers
 */

static int xsh_scired_slit_offset_create( cpl_plugin *);
static int xsh_scired_slit_offset_exec( cpl_plugin *);
static int xsh_scired_slit_offset_destroy( cpl_plugin *);

/* The actual executor function */
static void xsh_scired_slit_offset( cpl_parameterlist *, cpl_frameset *);

/*-----------------------------------------------------------------------------
  Static variables
  ----------------------------------------------------------------------------*/
static char xsh_scired_slit_offset_description_short[] =
"Reduce science exposure in SLIT configuration and on/off mode";

static char xsh_scired_slit_offset_description[] =
"This recipe reduces science exposure in SLIT configuration and on/off mode\n\
Input Frames : \n\
  - A set of n Science frames ( n even ), \
      Tag = STD_FLUX_SLIT_OFFSET_arm, SKY_SLIT_arm\n\
  - Spectral format table (Tag = SPECTRAL_FORMAT_TAB_arm)\n\
  - A master flat frame (Tag = MASTER_FLAT_SLIT_arm)\n\
  - An order table frame(Tag = ORDER_TABLE_EDGES_SLIT_arm)\n\
  - A wavelength calibration solution frame(Tag = WAVE_TAB_2D_arm)\n\
  - [OPTIONAL] Table with dispersion coefficients (Tag = DISP_TAB_arm)\n\
    It is required if -rectify-conserve-flux=TRUE\n\
  - [OPTIONAL] A badpixel map (Tag = BADPIXEL_MAP_arm)\n\
  - [OPTIONAL] The instrument response table (Tag = RESPONSE_MERGE1D_SLIT_arm)\n\
  - [OPTIONAL] An atmospheric extinction table (Tag = ATMOS_EXT_arm)\n\
Products : \n\
  - PREFIX_ORDER2D_arm extracted spectrum, order-by-order, 2D\n\
  - PREFIX_ORDER1D_arm extracted spectrum, order-by-order, 1D\n\
  - PREFIX_MERGE2D_arm merged spectrum, 2D\n\
  - PREFIX_MERGE1D_arm merged spectrum, 1D\n\
 - where PREFIX is SCI, FLUX, TELL if input raw DPR.TYPE contains OBJECT or FLUX or TELLURIC\n\
  - [OPTIONAL, if response and atm ext are provided]  PREFIX_FLUXCAL2D_MERGE2D_arm (2 dimension)\n\
  - [OPTIONAL, if response and atm ext are provided]  PREFIX_FLUXCAL1D_MERGE1D_arm (1 dimension)\n";

/*-----------------------------------------------------------------------------
  Functions code
  ----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/
/**
   @brief    Build the list of available plugins, for this module.
   @param    list    the plugin list
   @return   0 if everything is ok, -1 otherwise

   Create the recipe instance and make it available to the application using the
   interface. This function is exported.
*/
/*----------------------------------------------------------------------------*/

int cpl_plugin_get_info(cpl_pluginlist *list) {
  cpl_recipe *recipe = NULL;
  cpl_plugin *plugin = NULL;

  recipe = cpl_calloc(1, sizeof(*recipe));
  if ( recipe == NULL ){
    return -1;
  }

  plugin = &recipe->interface ;

  cpl_plugin_init(plugin,
                  CPL_PLUGIN_API,                    /* Plugin API */
                  XSH_BINARY_VERSION,             /* Plugin version */
                  CPL_PLUGIN_TYPE_RECIPE,            /* Plugin type */
                  RECIPE_ID,                         /* Plugin name */
                  xsh_scired_slit_offset_description_short,   /* Short help */
                  xsh_scired_slit_offset_description,         /* Detailed help */
                  RECIPE_AUTHOR,                     /* Author name */
                  RECIPE_CONTACT,                    /* Contact address */
                  xsh_get_license(),                 /* Copyright */
                  xsh_scired_slit_offset_create,
                  xsh_scired_slit_offset_exec,
                  xsh_scired_slit_offset_destroy);

  cpl_pluginlist_append(list, plugin);

  return (cpl_error_get_code() != CPL_ERROR_NONE);
}

/*----------------------------------------------------------------------------*/
/**
   @brief    Setup the recipe options
   @param    plugin  the plugin
   @return   0 if everything is ok

   Create the recipe instance and make it available to the application using the
   interface.

*/
/*----------------------------------------------------------------------------*/

static int xsh_scired_slit_offset_create(cpl_plugin *plugin){
  cpl_recipe *recipe = NULL;
  cpl_parameter* p=NULL;

  xsh_clipping_param crh_clip_param = {0.3, 5.0, 5, 0.7, 0};
  /* First param (conv_kernel) should be initialized correctly ! */
  xsh_remove_crh_single_param crh_single = { 0.1, 5, 2.0, 4} ;
  xsh_rectify_param rectify = { "default", 
                                CPL_KERNEL_DEFAULT, 
                                4, 
                                XSH_WAVE_BIN_SIZE_PIPE_NIR, 
                                XSH_SLIT_BIN_SIZE_PIPE_NIR,
                                1,0,0.};
  /* 2nd and 3rd params should be initialized correctly */
  xsh_localize_obj_param loc_obj = 
     {10, 0.1, 0, 0, LOC_MANUAL_METHOD, 0, 2.0,3,3,FALSE};
  xsh_extract_param extract_par = 
    { LOCALIZATION_METHOD};
  xsh_combine_nod_param nod_param = { 5, TRUE, 5, 2, 0.1, NULL, COMBINE_MEAN_METHOD } ;

  /* Reset library state */
  xsh_init();

  /* Check input */
  assure( plugin != NULL, CPL_ERROR_NULL_INPUT, "Null plugin");
  /* Get the recipe out of the plugin */
  assure( cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE,
          CPL_ERROR_TYPE_MISMATCH,
          "Plugin is not a recipe");

  recipe = (cpl_recipe *)plugin;

  /* Create the parameter list in the cpl_recipe object */
  recipe->parameters = cpl_parameterlist_new();
  assure( recipe->parameters != NULL,
          CPL_ERROR_ILLEGAL_OUTPUT,
          "Memory allocation failed!");

  /* Set generic parameters (common to all recipes) */
  check( xsh_parameters_generic( RECIPE_ID, recipe->parameters ) ) ;
  check( xsh_parameters_pre_overscan( RECIPE_ID, recipe->parameters ) ) ;

  /* crh clipping params */
  check(xsh_parameters_clipping_crh_create(RECIPE_ID,recipe->parameters,
    crh_clip_param));

  /* subtract_background_params */
  check(xsh_parameters_background_create(RECIPE_ID,recipe->parameters));
  check(p=xsh_parameters_find(recipe->parameters,RECIPE_ID,"background-method"));
  cpl_parameter_set_default_string(p,"poly");

  /* remove_crh_single */
  check(xsh_parameters_remove_crh_single_create(RECIPE_ID,recipe->parameters,
						crh_single )) ;
  /* xsh_rectify */
  check(xsh_parameters_rectify_create(RECIPE_ID,recipe->parameters,
					    rectify )) ;
  /* xsh_localize_object */
  check(xsh_parameters_localize_obj_create(RECIPE_ID,recipe->parameters,
					   loc_obj )) ;
  /* xsh_optimal_extract */
  check(xsh_parameters_optimal_extract_create(RECIPE_ID,
					      recipe->parameters,-1. )) ;
  /* trivial extract (Just temporary) */
  check(xsh_parameters_extract_create(RECIPE_ID,
				      recipe->parameters,
				      extract_par,LOCALIZATION_METHOD )) ;

  check(xsh_parameters_merge_ord_create(RECIPE_ID,
					recipe->parameters,
					WEIGHTED_MERGE_METHOD)) ;

  check(xsh_parameters_combine_nod_create(RECIPE_ID,
					  recipe->parameters,
					  nod_param )) ;
  
  check( xsh_parameters_new_boolean( recipe->parameters, RECIPE_ID,
    "do-flatfield", TRUE, 
    "TRUE if we do the flatfielding"));

  check( xsh_parameters_new_boolean( recipe->parameters, RECIPE_ID,
    "gen-sky", FALSE, 
    "if TRUE a 2D sky frame, a 2D rectified, a 2D merged sky are generated"));

  cleanup:
    if ( cpl_error_get_code() != CPL_ERROR_NONE ){
      xsh_error_dump(CPL_MSG_ERROR);
      return 1;
    }
    else {
      return 0;
    }
}

/*----------------------------------------------------------------------------*/
/**
   @brief    Execute the plugin instance given by the interface
   @param    plugin  the plugin
   @return   0 if everything is ok
*/
/*----------------------------------------------------------------------------*/

static int xsh_scired_slit_offset_exec(cpl_plugin *plugin) {
  cpl_recipe *recipe = NULL;

  /* Check parameter */

  assure( plugin != NULL, CPL_ERROR_NULL_INPUT, "Null plugin");

  /* Get the recipe out of the plugin */
  assure( cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE,
          CPL_ERROR_TYPE_MISMATCH, "Plugin is not a recipe");

  recipe = (cpl_recipe *)plugin;

  /* Check recipe */
  xsh_scired_slit_offset(recipe->parameters, recipe->frames);

  cleanup:
    if ( cpl_error_get_code() != CPL_ERROR_NONE ) {
      xsh_error_dump(CPL_MSG_ERROR);
      xsh_error_reset();
      return 1;
    }
    else {
      return 0;
    }
}

/*----------------------------------------------------------------------------*/
/**
   @brief    Destroy what has been created by the 'create' function
   @param    plugin  the plugin
   @return   0 if everything is ok
*/
/*----------------------------------------------------------------------------*/
static int xsh_scired_slit_offset_destroy(cpl_plugin *plugin)
{
  cpl_recipe *recipe = NULL;

  /* reset error state before detroying recipe */
  xsh_error_reset(); 
  /* Check parameter */
  assure( plugin != NULL, CPL_ERROR_NULL_INPUT, "Null plugin" );

  /* Get the recipe out of the plugin */
  assure( cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE,
    CPL_ERROR_TYPE_MISMATCH, "Plugin is not a recipe");

  recipe = (cpl_recipe *)plugin;

  xsh_free_parameterlist(&recipe->parameters);

 cleanup:
  if (cpl_error_get_code() != CPL_ERROR_NONE)
    {
      return 1;
    }
  else
    {
      return 0;
    }
}


static cpl_error_code 
xsh_params_monitor(xsh_background_param* backg,
                   xsh_rectify_param * rectify_par,
                   xsh_localize_obj_param * loc_obj_par)
{


  xsh_msg_dbg_low("bkg params: sampley=%d radius_y=%d smooth_y=%d",
	  backg->sampley,backg->radius_y,backg->smooth_y);

  xsh_msg_dbg_low("bkg params: radius_x=%d smooth_x=%d",
	  backg->radius_x,backg->smooth_x);

  xsh_msg_dbg_low("rectify params: radius=%g bin_lambda=%g bin_space=%g",
	  rectify_par->rectif_radius,rectify_par->rectif_bin_lambda,
	  rectify_par->rectif_bin_space);

  xsh_msg_dbg_low("localize params: chunk_nb=%d nod_step=%g",
	  loc_obj_par->loc_chunk_nb,loc_obj_par->nod_step);

  return cpl_error_get_code();

}



static cpl_error_code 
xsh_params_set_defaults(cpl_parameterlist* pars,
                        xsh_instrument* inst,
                        xsh_rectify_param * rectify_par)
{
   cpl_parameter* p=NULL;
 
  if (xsh_instrument_get_arm(inst) == XSH_ARM_NIR){

     check(p=xsh_parameters_find(pars,RECIPE_ID,"rectify-bin-slit"));
     if(xsh_parameter_get_default_flag(p) == 0) {
        /* if the user has not set the parameter use default */
        rectify_par->rectif_bin_space=XSH_SLIT_BIN_SIZE_PIPE_NIR;
        cpl_parameter_set_double(p,XSH_SLIT_BIN_SIZE_PIPE_NIR);
     }

     check(p=xsh_parameters_find(pars,RECIPE_ID,"rectify-bin-lambda"));
     if(xsh_parameter_get_default_flag(p) == 0) {
        /* if the user has not set the parameter use default */
        rectify_par->rectif_bin_lambda=XSH_WAVE_BIN_SIZE_PIPE_NIR;
        cpl_parameter_set_double(p,XSH_WAVE_BIN_SIZE_PIPE_NIR);
     }



  } else {


     check(p=xsh_parameters_find(pars,RECIPE_ID,"rectify-bin-slit"));
     if(xsh_parameter_get_default_flag(p) == 0) {
        /* if the user has not set the parameter use default */
        rectify_par->rectif_bin_space=XSH_SLIT_BIN_SIZE_PIPE_UVB;
        cpl_parameter_set_double(p,XSH_SLIT_BIN_SIZE_PIPE_UVB);
     }

     check(p=xsh_parameters_find(pars,RECIPE_ID,"rectify-bin-lambda"));
     if(xsh_parameter_get_default_flag(p) == 0) {
        /* if the user has not set the parameter use default */
        rectify_par->rectif_bin_lambda=XSH_WAVE_BIN_SIZE_PIPE_UVB;
        cpl_parameter_set_double(p,XSH_WAVE_BIN_SIZE_PIPE_UVB);

     }


  }

 cleanup:
 
  return cpl_error_get_code();

}



/*--------------------------------------------------------------------------*/
/**
  @brief    Scale relevant input parameters for binning
  @param    raws   the frames list
  @param    backg  structure to control inter-order background computation

  In case of failure the cpl_error_code is set.
 */
/*--------------------------------------------------------------------------*/

static cpl_error_code 
xsh_params_bin_scale(cpl_frameset* raws,
                     xsh_background_param* backg)
{

  cpl_frame* frame=NULL;
  const char* name=NULL;
  cpl_propertylist* plist=NULL;
  int binx=0;
  int biny=0;
  xsh_msg("Rescale relevant parameters to bin size");
  check(frame=cpl_frameset_get_first(raws));
  check(name=cpl_frame_get_filename(frame));

  check(plist=cpl_propertylist_load(name,0));
  check(binx=xsh_pfits_get_binx(plist));
  check(biny=xsh_pfits_get_biny(plist));
  xsh_free_propertylist(&plist);

  if(biny>1) {

    /*
    backg->sampley=backg->sampley/biny;
    Not bin dependent.
    */

    backg->radius_y=backg->radius_y/biny;
    /*
    if(backg->smooth_y>0) {
      backg->smooth_y=backg->smooth_y/biny;
    }
    Smoothing radius' half size in y direction
    For the moment not bin dependent, but for optimal results probably yes
    */

    /*
    rectify_par->rectif_radius=rectify_par->rectif_radius/biny;
    Rectify Interpolation radius.
    For the moment not bin dependent, but for optimal results probably yes
    */

    /*
    rectify_par->rectif_bin_lambda=rectify_par->rectif_bin_lambda/biny;
    Rectify Wavelength Step.
    For the moment not bin dependent, but for optimal results probably yes
    */

    /*
    loc_obj_par->loc_chunk_nb=loc_obj_par->loc_chunk_nb/biny;
    Localization Nb of chunks.
    Not bin dependent
    */
  }


  if(binx>1) {

    backg->radius_x=backg->radius_x/binx;
    /*
    if(backg->smooth_x>0) {
      backg->smooth_x=backg->smooth_x/binx;
    }
    Smoothing radius' half size in x direction
    For the moment not bin dependent, but for optimal results probably yes
    */

    /*
    rectify_par->rectif_bin_space=rectify_par->rectif_bin_space/binx;
    Rectify Position Step.
    For the moment not bin dependent, but for optimal results probably yes
    */

    /*
    loc_obj_par->nod_step=loc_obj_par->nod_step/binx;
    Step (arcsec) between A and B images in nodding mode.
    For the moment not bin dependent, but for optimal results probably yes
    */
  }
 
 cleanup:
  xsh_free_propertylist(&plist);
  return cpl_error_get_code();

}


/*----------------------------------------------------------------------------*/
/**
   @brief    Interpret the command line options and execute the data processing
   @param    parameters     the parameters list
   @param    frameset   the frames list

   In case of failure the cpl_error_code is set.
*/
/*----------------------------------------------------------------------------*/
static void xsh_scired_slit_offset( cpl_parameterlist* parameters, 
  cpl_frameset* frameset)
{
   const char* recipe_tags[3] = {XSH_OBJECT_SLIT_OFFSET, XSH_STD_FLUX_SLIT_OFFSET, XSH_SKY_SLIT};
  int recipe_tags_size = 3;

  /* Input frames */
  cpl_frameset* raws = NULL;

  cpl_frameset * raw_object = NULL;
  cpl_frameset * raw_sky = NULL;
  cpl_frameset * calib = NULL;
  int nb_sub_frames = 0;
  /* Beware, do not "free" the following input frames, they are part
     of the input frameset */
  cpl_frame* bpmap = NULL;
  cpl_frame* master_bias = NULL;
  cpl_frame* master_dark = NULL;
  cpl_frame* master_flat = NULL;
  cpl_frame* order_tab_edges = NULL;
  cpl_frame * wave_tab = NULL ;
  cpl_frame * model_config_frame = NULL ;
  cpl_frame * wavemap = NULL ;
  cpl_frame *disp_tab_frame = NULL;
  cpl_frame * spectral_format = NULL ;

  /* Parameters */
  xsh_clipping_param* crh_clipping_par = NULL;
  xsh_background_param* backg_par = NULL;
  xsh_remove_crh_single_param * crh_single_par = NULL ;
  xsh_rectify_param * rectify_par = NULL ;
  xsh_localize_obj_param * loc_obj_par = NULL ;

  double opt_kappa = -1. ;
  xsh_merge_param* merge_par = NULL;
  xsh_extract_param * extract_par = NULL ;
  xsh_combine_nod_param * combine_nod_param = NULL ;

  xsh_instrument* instrument = NULL;

  /* Intermediate frames */

  cpl_frame * loc_table_frame = NULL ; /**< Output of xsh_localize_object */
  cpl_frame * clean_frame = NULL ; /**< Output of remove_crh_single */
  cpl_frameset * sub_frameset = NULL ;	/**< frameset of subtracted offset frames */
  cpl_frameset * rmbkg_frameset = NULL ;	/**< frameset of subtracted background frames */
  cpl_frameset * clean_frameset = NULL ;	/**< frameset of clean offset frames */
  cpl_frameset * rect_frameset = NULL ; /**< frameset of rectified frames */
  cpl_frameset * rect2_frameset_tables = NULL ; /**< frameset of rectified frames */

  /* Output Frames (results)*/
  cpl_frame * res1D_frame = NULL ;	/**< Final (result) frame */
  cpl_frame * res2D_frame = NULL ;	/**< Final (result) frame */
  cpl_frame * ext_frame = NULL ;
  cpl_frame * ext_frame_eso = NULL ;
  cpl_frame * comb_frame = NULL ;
  cpl_frame * comb_frame_eso = NULL ;
  cpl_frame* response_ord_frame=NULL;

  char  file_name[80];
  char  arm_str[16] ;
  char  file_tag[40];

  int i ;
  cpl_frame* grid_back=NULL;
  cpl_frame* frame_backg=NULL;
  int do_flatfield=0;
  int gen_sky=0;
  char rec_prefix[80];
  char sky_prefix[80];
  char sky_tag[80];
  cpl_frame* avg_sky=NULL;
  cpl_frame* rec_sky=NULL;
  cpl_frame* rec_sky_eso=NULL;
  cpl_frame* mer_sky=NULL;
  cpl_frame * rectif_tab = NULL ;
  cpl_frame * sky_divided = NULL ;
  cpl_frameset* sky_bias=NULL;
  cpl_frameset* sky_dark=NULL;
  cpl_frame * fluxcal_rect_1D_frame = NULL ;	
  cpl_frame * fluxcal_rect_2D_frame = NULL ;	
  cpl_frame * fluxcal_1D_frame = NULL ;	
  cpl_frame * fluxcal_2D_frame = NULL ;	
  cpl_frame* frm_atmext=NULL;
  int pre_overscan_corr=0;

  /**************************************************************************/
  /* DFS management */
  /**************************************************************************/
  check( xsh_begin( frameset, parameters, &instrument, &raws, &calib,
                    recipe_tags, recipe_tags_size,
                    RECIPE_ID, XSH_BINARY_VERSION,
                    xsh_scired_slit_offset_description_short));

  check(xsh_ensure_raws_input_offset_recipe_is_proper(raws,instrument));
  assure( instrument->mode == XSH_MODE_SLIT, CPL_ERROR_ILLEGAL_INPUT,
	  "Instrument NOT in Slit Mode" ) ;

  /**************************************************************************/
  /* Recipe frames */
  /**************************************************************************/
  check(xsh_slit_offset_get_calibs(calib,instrument,&bpmap,&master_bias,
                                   &master_dark,&order_tab_edges,
                                   &model_config_frame,&wave_tab,&master_flat,
                                   &wavemap,&spectral_format));
  /* scired specifig input */
  check( response_ord_frame = xsh_find_frame_with_tag(calib,
                                                      XSH_RESPONSE_MERGE1D_SLIT,
                                                      instrument));

   if(response_ord_frame != NULL) { 
      frm_atmext=xsh_find_frame_with_tag(calib,XSH_ATMOS_EXT,instrument);
      if(frm_atmext==NULL) {
         xsh_msg_error("Provide atmospheric extinction frame");
      }
   }

  /**************************************************************************/
  /* Recipe parameters */
  /**************************************************************************/
  check(xsh_slit_offset_get_params(parameters,RECIPE_ID,
                                  &crh_clipping_par,&backg_par,
                                  &loc_obj_par,&rectify_par,&crh_single_par,
                                  &opt_kappa,&merge_par,&extract_par,
                                  &combine_nod_param,&do_flatfield,&gen_sky));

  check(xsh_params_set_defaults(parameters,instrument,rectify_par));
  /* adjust relevant parameter to binning */
  if ( xsh_instrument_get_arm( instrument ) != XSH_ARM_NIR ) {
    check(xsh_params_bin_scale(raws,backg_par));
  }
  check(xsh_params_monitor(backg_par,rectify_par,loc_obj_par));

  if ( rectify_par->conserve_flux){
     check_msg( disp_tab_frame = xsh_find_disp_tab( calib, instrument),
                "if rectify-conserve-flux=TRUE DISP_TAB_ARM is required input");
  }
  /**************************************************************************/
  /* Recipe code */
  /**************************************************************************/
  /* prepare RAW frames in PRE format */
  check(xsh_prepare(raws, bpmap, master_bias, XSH_OBJECT_SLIT_OFFSET,
		    instrument,pre_overscan_corr));

  /* Separate OBJECT and SKY frames */
  check( nb_sub_frames = xsh_dfs_split_offset( raws, XSH_OBJECT_SLIT_OFFSET,
					       XSH_SKY_SLIT, &raw_object,
					       &raw_sky ) ) ;

  check(strcpy(rec_prefix,
               xsh_set_recipe_file_prefix(raw_object,"xsh_scired_slit_offset")));

  if(gen_sky) {
     if(strstr(rec_prefix,"SCI") != NULL) {
        strcpy(sky_prefix,"SKY_SLIT");
     } else if(strstr(rec_prefix,"TELL") != NULL) {
      strcpy(sky_prefix,"TELL_SLIT_SKY");
     } else if(strstr(rec_prefix,"FLUX") != NULL) {
      strcpy(sky_prefix,"FLUX_SLIT_SKY");
     } else {
      strcpy(sky_prefix,"CAL_SLIT_SKY");
     }
  }
  if(nb_sub_frames==0) {
     xsh_msg_error("nb_sub_frames=%d something wrong check your input raws",
                   nb_sub_frames);
     goto cleanup;

  }

  check( sub_frameset = xsh_subtract_sky_offset( raw_object, raw_sky,
						 nb_sub_frames,
						 instrument )) ;

  xsh_msg_dbg_low( "Nb of Subtracted Frames: %d",
	   cpl_frameset_get_size( sub_frameset ) ) ;

  check( clean_frameset = cpl_frameset_new() ) ;
  check( rmbkg_frameset = cpl_frameset_new() ) ;

  sprintf( arm_str, "%s",  xsh_instrument_arm_tostring(instrument) ) ;

  if(gen_sky) {
    /* Not yet working.. */
    if(master_bias!= NULL && pre_overscan_corr ==0 ) {
        check(sky_bias=xsh_pre_frameset_subtract_frame(raw_sky,master_bias,"MBIAS_",instrument));
     } else {
        sky_bias=cpl_frameset_duplicate(raw_sky);
     }
     if(master_dark!= NULL) {
        check(sky_dark=xsh_pre_frameset_subtract_frame(sky_bias,master_dark,"MDARK",instrument));
     } else {
        sky_dark=cpl_frameset_duplicate(sky_bias);
     }
     sprintf(sky_tag,"%s_SKY_%s",rec_prefix,arm_str);
     check(avg_sky=xsh_frameset_average_pre(sky_dark,instrument,sky_tag));
     if(do_flatfield==1) {
        sprintf(sky_tag,"%s_FF_SKY_%s",rec_prefix,arm_str);
        check( sky_divided = xsh_divide_flat( avg_sky,master_flat, sky_tag,
                                           instrument ) ) ;
     } else {
        sky_divided=cpl_frame_duplicate(avg_sky);
     }
  }


  for( i = 0 ; i<nb_sub_frames ; i++ ) {
    cpl_frame * a_b = NULL ;
    cpl_frame * rm_bkg = NULL ;
    char str[16] ;

    sprintf( str, "_%d", i ) ;
    a_b = cpl_frameset_get_frame( sub_frameset, i ) ;

    /* subtract background */
    xsh_msg("Subtract inter-order background [%d]", i );
    sprintf(file_name,"%s%s", rec_prefix, str) ;
    check(rm_bkg = xsh_subtract_background( a_b,
					    order_tab_edges,
					    backg_par, instrument,
					    file_name,&grid_back,
					    &frame_backg));
    check( cpl_frameset_insert( rmbkg_frameset, rm_bkg ) ) ;
  }

  xsh_msg_dbg_low( "Nb of Frames Bkg Subtracted: %d",
	   cpl_frameset_get_size( rmbkg_frameset ) ) ;

  if ( nb_sub_frames < 3 && crh_single_par->nb_iter > 0 ) {
    xsh_msg( "Less than 3 frames AND removecrhsingle_niter > 0" ) ;
    for ( i = 0 ; i < nb_sub_frames ; i++ ) {
      cpl_frame * divided = NULL ;
      cpl_frame * rm_crh = NULL ;
      cpl_frame * a_b = NULL ;
      char str[16] ;

      check( a_b = cpl_frameset_get_frame( rmbkg_frameset, i ) ) ;
      sprintf( str, "%d", i ) ;
      sprintf(file_tag,"NO_CRH_SLIT_OFFSET_%s%s",arm_str,str) ;
      xsh_msg( "Remove crh (single frame)" ) ;
      check( rm_crh = xsh_remove_crh_single( a_b, wavemap,instrument,
					     crh_single_par,
					     file_tag ) ) ;
      if(do_flatfield==1) {
         xsh_msg( "Calling xsh_divide_flat" ) ;
         sprintf(file_tag,"FF_SLIT_OFFSET_%s%s", arm_str,str) ;
         check( divided = xsh_divide_flat( rm_crh,master_flat, file_tag,
                                           instrument ) ) ;
      } else {
         divided=cpl_frame_duplicate(rm_crh);
      }
      check( cpl_frameset_insert( clean_frameset, divided ) ) ;
    }
  }
  else for( i = 0 ; i < nb_sub_frames ; i++ ) {
    cpl_frame * divided = NULL ;
    cpl_frame * a_b = NULL ;
    char str[16] ;

    /* If enough frames, the CRH removal is done in xsh_combine_offset */
    a_b = cpl_frameset_get_frame( rmbkg_frameset, i ) ;
    sprintf( str, "%d", i ) ;

    if(do_flatfield==1) {
       /* Divide by flat */
       xsh_msg( "Calling xsh_divide_flat" ) ;

       sprintf(file_tag,"FF_SLIT_OFFSET_%s%s", arm_str,str) ;
       check( divided = xsh_divide_flat( a_b,master_flat, file_tag,
                                         instrument ) ) ;
    } else {
       divided=cpl_frame_duplicate(a_b);
    }
    check( cpl_frameset_insert( clean_frameset, divided ) ) ;
  }

  {
    int clean_size ;

    check( clean_size = cpl_frameset_get_size( clean_frameset ) ) ;
    xsh_msg( "Nb of clean Frames: %d", clean_size ) ;
  }

  /* Now rectify each clean frame */
  check( rect_frameset = cpl_frameset_new() ) ;
  for( i = 0 ; i < nb_sub_frames ; i++ ) {
    cpl_frame * rectif = NULL ;
    cpl_frame * rectif_eso = NULL ;
    cpl_frame * clean = NULL ;
    char str[16] ;

    sprintf( str, "%d", i ) ;
    check( clean = cpl_frameset_get_frame( clean_frameset, i ) ) ;
    xsh_msg( "Rectifying Frame '%s'", cpl_frame_get_filename( clean ) ) ;
    sprintf(file_name,"RECTIFIED_SLIT_OFFSET_%s_%s.fits",arm_str,str) ;
    check( rectif = xsh_rectify( clean, order_tab_edges, wave_tab,
				 model_config_frame, instrument,
				 rectify_par, spectral_format,
                                 disp_tab_frame,
				 file_name,&rectif_eso,&rectif_tab,
				 rec_prefix ) ) ;
    check( cpl_frameset_insert( rect_frameset, rectif ) ) ;
    check( cpl_frameset_insert( rect_frameset, rectif_eso ) ) ;
    xsh_free_frame(&rectif_tab);
  }



  if(gen_sky) {
     xsh_msg("rectify sky frame");
     sprintf(file_name,"%s_RECTIFIED_SKY_%s.fits",sky_prefix,arm_str) ;
     check( rec_sky = xsh_rectify(sky_divided, order_tab_edges, wave_tab,
                                  model_config_frame, instrument,
                                  rectify_par, spectral_format,
                                  disp_tab_frame,
                                  file_name,&rec_sky_eso,&rectif_tab,
                                  sky_prefix ) ) ;


     xsh_msg("Merge 2D sky frame");
     check( mer_sky = xsh_merge_ord( rec_sky, instrument,
                                     merge_par,sky_prefix ));


  }

  /* Combine rectified Frames */
  sprintf(file_tag,"%s_%s_%s",rec_prefix,XSH_ORDER2D, arm_str) ;
  check(rect2_frameset_tables=xsh_frameset_ext_table_frames(rect_frameset));
  check( comb_frame = xsh_combine_nod(rect2_frameset_tables, combine_nod_param,
				       file_tag,instrument,&comb_frame_eso));
  xsh_free_frameset(&rect2_frameset_tables);

  if ( extract_par->method == LOCALIZATION_METHOD ) {
    xsh_msg( "Localize before extraction" ) ;
    sprintf(file_name,"LOCALIZE_%s_ALL.fits",arm_str) ;
    check( loc_table_frame = xsh_localize_obj( comb_frame, NULL,instrument,
					       loc_obj_par, NULL,
					       file_name) ) ;
  }

  xsh_msg( "Calling xsh_extract" ) ;
  check(ext_frame = xsh_extract(comb_frame, loc_table_frame,
				instrument, extract_par,&ext_frame_eso,
				rec_prefix )) ;
  xsh_msg( "Calling xsh_merge_ord with 1D frame" ) ;
  check( res1D_frame = xsh_merge_ord( ext_frame, instrument,
				      merge_par,rec_prefix ));
  /* More todo here
-     mark_telluric
     calibrate_flux
  */

  xsh_msg( "Calling xsh_merge_ord with 2D frame" ) ;
  check( res2D_frame = xsh_merge_ord( comb_frame, instrument, merge_par,
				      rec_prefix ) ) ;

  if(response_ord_frame != NULL && frm_atmext != NULL) {
   check(xsh_flux_calibrate(comb_frame_eso,ext_frame_eso,frm_atmext,
			    response_ord_frame,merge_par,instrument,rec_prefix,
                            &fluxcal_rect_2D_frame,&fluxcal_rect_1D_frame,
			    &fluxcal_2D_frame,&fluxcal_1D_frame));

  }

  /* Save products */
  xsh_msg( "Saving products" ) ;
  /* Removed as format not easy to read
   check( xsh_add_product_table_multi( comb_frame, frameset,
  			      parameters, RECIPE_ID, instrument));
   *
   */

  check( xsh_add_product_image( comb_frame_eso, frameset, 
				parameters, RECIPE_ID, instrument,
				cpl_frame_get_tag(comb_frame_eso)));

  /* Removed as format not easy to read
  check( xsh_add_product_table_multi( ext_frame, frameset,
				      parameters, RECIPE_ID, instrument));
  */

  check( xsh_add_product_image( ext_frame_eso, frameset,
				parameters, RECIPE_ID, instrument,
				cpl_frame_get_tag(ext_frame_eso)));


  check( xsh_add_product_spectrum( res2D_frame, frameset, parameters,
				RECIPE_ID, instrument));

  check(xsh_monitor_spectrum1D_flux(res1D_frame,instrument));
  check( xsh_add_product_spectrum( res1D_frame, frameset, parameters,
			      RECIPE_ID, instrument));


  if(gen_sky) {
     check( xsh_add_product_image( avg_sky, frameset, parameters,
                                   RECIPE_ID, instrument,
                                   cpl_frame_get_tag(avg_sky)));


     sprintf(file_tag,"%s_%s_%s",rec_prefix,XSH_ORDER2D, arm_str) ;
     check( xsh_add_product_image( rec_sky_eso, frameset, parameters,
                                   RECIPE_ID, instrument,
                                   cpl_frame_get_tag(rec_sky_eso)));


     sprintf(file_tag,"%s_%s_%s",rec_prefix,XSH_MERGE2D, arm_str) ;
     check( xsh_add_product_image( mer_sky, frameset, parameters,
                                   RECIPE_ID, instrument,
                                   cpl_frame_get_tag(mer_sky)));

  }

  if(response_ord_frame != NULL) {
     check( xsh_add_product_image(fluxcal_rect_2D_frame,frameset,parameters, 
                                  RECIPE_ID, instrument,
                                  cpl_frame_get_filename(fluxcal_rect_2D_frame)));
     check( xsh_add_product_image(fluxcal_rect_1D_frame,frameset,parameters, 
                                  RECIPE_ID, instrument,
                                  cpl_frame_get_filename(fluxcal_rect_1D_frame)));
     check( xsh_add_product_spectrum( fluxcal_2D_frame, frameset, parameters, 
                                      RECIPE_ID, instrument));
     check( xsh_add_product_spectrum( fluxcal_1D_frame, frameset, parameters, 
                                      RECIPE_ID, instrument));
  }


  cleanup:

    xsh_end( RECIPE_ID, frameset, parameters );
    XSH_FREE(crh_clipping_par);
    XSH_FREE( backg_par ) ;
    XSH_FREE( crh_single_par ) ;
    XSH_FREE( rectify_par ) ;
    XSH_FREE( loc_obj_par ) ;
    XSH_FREE(merge_par);
    XSH_FREE(extract_par);
    XSH_FREE(combine_nod_param);
    xsh_instrument_free(&instrument );
    xsh_free_frameset(&rect2_frameset_tables);
    xsh_free_frameset(&raws);
    /* the following two frees are dangerous in case of failure (e.g. nsky=0)
    xsh_free_frameset( &raw_object);
    xsh_free_frameset( &raw_sky);
    */
    xsh_free_frameset(&calib);
    xsh_free_frameset(&sub_frameset);
    xsh_free_frameset(&clean_frameset);
    xsh_free_frameset(&rmbkg_frameset);
    xsh_free_frameset(&rect_frameset);

    xsh_free_frameset(&sky_bias);
    xsh_free_frameset(&sky_dark);
   

    xsh_free_frame(&loc_table_frame) ;
    xsh_free_frame(&clean_frame) ;
    xsh_free_frame(&res1D_frame) ;
    xsh_free_frame(&res2D_frame) ;
    xsh_free_frame( &comb_frame ) ;
    xsh_free_frame( &comb_frame_eso ) ;
    xsh_free_frame( &ext_frame ) ;

    xsh_free_frame( &avg_sky ) ;
    xsh_free_frame( &sky_divided ) ;
    xsh_free_frame( &rec_sky ) ;
    xsh_free_frame( &rec_sky_eso ) ;
    xsh_free_frame( &mer_sky ) ;

    xsh_free_frame( &frame_backg ) ;
    xsh_free_frame( &grid_back ) ;
    xsh_free_frame(&fluxcal_rect_1D_frame) ;
    xsh_free_frame(&fluxcal_rect_2D_frame) ;
    xsh_free_frame(&fluxcal_1D_frame) ;
    xsh_free_frame(&fluxcal_2D_frame) ;

    return;
}

/**@}*/
