/*                                                                          *
 *   This file is part of the ESO X-shooter Pipeline                        *
 *   Copyright (C) 2006 European Southern Observatory                       *
 *                                                                          *
 *   This library is free software; you can redistribute it and/or modify   *
 *   it under the terms of the GNU General Public License as published by   *
 *   the Free Software Foundation; either version 2 of the License, or      *
 *   (at your option) any later version.                                    *
 *                                                                          *
 *   This program is distributed in the hope that it will be useful,        *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for more details.                           *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to the Free Software            *
 *   Foundation, 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA   *
 *                                                                          */
 
/*
 * $Author: amodigli $
 * $Date: 2010/10/03 14:40:01 $
 * $Revision: 1.5 $
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

/*-------------------------------------------------------------------------*/
/**
  @defgroup test_xsh_spectrum_merge_3d  Test a cube
  @ingroup unit_tests 
*/
/*-------------------------------------------------------------------------*/
/**@{*/

/*--------------------------------------------------------------------------
  Includes
  --------------------------------------------------------------------------*/

#include <tests.h>
#include <xsh_data_pre.h>
#include <xsh_error.h>
#include <xsh_msg.h>
#include <xsh_data_instrument.h>
#include <xsh_data_rec.h>
#include <xsh_data_localization.h>
#include <xsh_drl.h>
#include <xsh_pfits.h>
#include <xsh_data_pre_3d.h>
#include <xsh_badpixelmap.h>

#include <cpl.h>
#include <math.h>

#include <getopt.h>

/*--------------------------------------------------------------------------
  Defines
  --------------------------------------------------------------------------*/

#define MODULE_ID "XSH_SPECTRUM_MERGE3D"

static const char * Options = "" ;

enum {
  ZSTEP_OPT, ZCHUNK_HSIZE_OPT, ZMETHOD_OPT
};

static struct option long_options[] = {
  {"z-step", required_argument, 0, ZSTEP_OPT},
  {"z-chunk-hsize", required_argument, 0, ZCHUNK_HSIZE_OPT},
  {"z-method", required_argument, 0, ZMETHOD_OPT},
  {0, 0, 0, 0}
};

static void Help( void )
{
  puts( "Unitary test of a cube");
  puts( "Usage: test_xsh_data_spectrum_merge_3d [options] <input_files>");

  puts( "Options" ) ;
  puts( "  --z-step=<n>");
  puts( "  --z-chunk-hsize=<n>");
  puts( "  --z-method : MAX=0 | GAUSSIAN=1 [1]");
  puts( "\nInput Files" ) ;
  puts( "The input files argument MUST be in this order:" ) ;
  puts( " 1. Cube" ) ;
  TEST_END();
}


static void HandleOptions( int argc, char **argv, 
  int *zstep, int *zchunk_hsize, int *zmethod)
{
  int opt ;
  int option_index = 0;
      
  while (( opt = getopt_long (argc, argv, Options,
                              long_options, &option_index)) != EOF )
    switch ( opt ) {
    case ZSTEP_OPT:
      *zstep = atoi( optarg);
    break;
    case ZCHUNK_HSIZE_OPT:
      *zchunk_hsize = atoi( optarg);                   
    break;
    case ZMETHOD_OPT:
      *zmethod = atoi( optarg);
    break;
    default: Help() ; exit( 0);
  }
}
/*--------------------------------------------------------------------------
  Implementation
  --------------------------------------------------------------------------*/

/**
  @brief
    Unit test of xsh_opt_extract
  @return
    0 if success

*/
int main( int argc, char **argv)
{
  int ret;
  /* Declarations */
  xsh_instrument* instrument = NULL;
  const char* sci_name = NULL;
  cpl_frame *sci_frame = NULL;
  
  xsh_pre_3d *cube = NULL;
  xsh_image_3d *cube_data_img = NULL;
  int cube_x, cube_y, cube_z;
  float *cube_data = NULL;
  double *median = NULL;
  double *coadd =NULL;
  cpl_vector *med_vect = NULL;
  cpl_vector *coadd_vect = NULL;
  cpl_vector *positions = NULL;
  int x,y,z;
  double med_z;  
  cpl_image *med_img = NULL;
  double *med_img_data = NULL;
  double *profil = NULL;
  cpl_vector *profil_vect = NULL;
    FILE *profil_file = NULL;
    char filename[256];
  int zmed;
  cpl_propertylist *header = NULL;
  double crval1, crval2, crval3;
  double crpix1, crpix2, crpix3;
  double cdelt1, cdelt2, cdelt3;

  /* Initialize libraries */
  TESTS_INIT( MODULE_ID);
  cpl_msg_set_level( CPL_MSG_DEBUG);
  xsh_debug_level_set( XSH_DEBUG_LEVEL_MEDIUM);

  /* Analyse parameters */
  if (argc >= 2 ) {
    sci_name = argv[1];
  }
  else{
    Help();
    exit(0);
  }

  TESTS_XSH_FRAME_CREATE( sci_frame, "CUBE_arm", sci_name);

  /* USE load function */
  xsh_msg("CUBE            : %s",
    cpl_frame_get_filename( sci_frame));
  
  xsh_msg(" Parameters ");

  check( cube = xsh_pre_3d_load( sci_frame));
  header = cube->data_header;

  check( cube_data_img = xsh_pre_3d_get_data( cube));
  check( cube_x = xsh_image_3d_get_size_x( cube_data_img));
  check( cube_y = xsh_image_3d_get_size_y( cube_data_img));
  check( cube_z = xsh_image_3d_get_size_z( cube_data_img));

  xsh_msg(" Cube SLITLET*SLIT*LAMBDAS (%dx%dx%d)", cube_x, cube_y, cube_z);
  
  check( crpix1 = xsh_pfits_get_crpix1( header));
  check( crval1 = xsh_pfits_get_crval1( header));
  check( cdelt1 = xsh_pfits_get_cdelt1( header));

  check( crpix2 = xsh_pfits_get_crpix2( header));
  check( crval2 = xsh_pfits_get_crval2( header));
  check( cdelt2 = xsh_pfits_get_cdelt2( header));

  check( crpix3 = xsh_pfits_get_crpix3( header));
  check( crval3 = xsh_pfits_get_crval3( header));
  check( cdelt3 = xsh_pfits_get_cdelt3( header));

  xsh_msg(" SLITLET pix %f ref %f with step %f", crpix1, crval1, cdelt1);
  xsh_msg(" SLIT    pix %f ref %f with step %f", crpix2, crval2, cdelt2);
  xsh_msg(" LAMBDAS pix %f ref %f with step %f", crpix3, crval3, cdelt3); 

  check( cube_data = (float*)xsh_image_3d_get_data( cube_data_img));
  
  XSH_MALLOC( median, double, cube_z);
  check( med_vect = cpl_vector_wrap( cube_z, median));
  check( med_img = cpl_image_new( cube_x, cube_y, CPL_TYPE_DOUBLE));
  check( med_img_data = cpl_image_get_data_double( med_img));
 
  for( y=0; y<cube_y; y++){
    for( x=0; x<cube_x; x++){
      for(z=0; z< cube_z; z++){
        median[z] = cube_data[cube_x*cube_y*z+cube_x*y+x];
      }
      check (med_z = cpl_vector_get_median( med_vect));
      med_img_data[y*cube_x+x] = med_z;
    }
  }
  check( cpl_image_save( med_img, "MEDIAN_CUBE.fits", CPL_BPP_IEEE_DOUBLE, 
    NULL, CPL_IO_CREATE));

  sprintf( filename, "profil_%.3f.dat", cdelt2);
  profil_file = fopen( filename, "w");
  fprintf( profil_file, "#slit UP CEN LO\n");
 
  for( y=0; y<cube_y; y++){
    fprintf( profil_file, "%f ", crval2+(y+0.5)*cdelt2);
    for( x=0; x<cube_x; x++){
      fprintf( profil_file, "%f ", med_img_data[y*cube_x+x]);
    }
    fprintf( profil_file, "\n");
  }
  fclose( profil_file);

  /* second test */
  XSH_MALLOC( coadd, double, cube_y);
  check( coadd_vect = cpl_vector_wrap( cube_y, coadd));
  check( positions = cpl_vector_new( cube_y));

  for( y=0; y<cube_y; y++){
    cpl_vector_set( positions, y, y);
  }

  profil_file = fopen( "cube.dat", "w");


  fprintf( profil_file, "# z lref up cen low\n");
  for(z=10; z< cube_z; z+=100){
    double cen[3];
    double x0= 0.0,sigma =0.0, area=0.0, offset=0.0;

    for( x=0; x<cube_x; x++){
      
      for( y=0; y<cube_y; y++){
        coadd[y]=0;
      }
      for(zmed=z-10; zmed< z+10; zmed++){
        int max_pos = 0;
        double max =0;

        for( y=0; y<cube_y; y++){        
          coadd[y] += cube_data[cube_x*cube_y*zmed+cube_x*y+x];
        }
      }
      cpl_vector_fit_gaussian( positions, NULL, coadd_vect, NULL,CPL_FIT_ALL,&x0,&sigma,&area,&offset,NULL,NULL,NULL);
      if (cpl_error_get_code() != CPL_ERROR_NONE){
        xsh_error_reset();
      for( y=0; y<cube_y; y++){
          xsh_msg("NO FIT %f %f\n", cpl_vector_get( positions, y), cpl_vector_get( coadd_vect, y));
        }
        cen[x]=-1;
      }
      else{
        cen[x] = x0;
      } 
    }
    fprintf( profil_file, "%d %f %f %f %f\n", z, crval3+z*cdelt3, cen[0], cen[1], cen[2]);
  }
  fclose( profil_file);

  cleanup:
   if (cpl_error_get_code() != CPL_ERROR_NONE) {
      xsh_error_dump(CPL_MSG_ERROR);
      ret = 1;
    }
    xsh_free_frame( &sci_frame);
    xsh_pre_3d_free( &cube);
    XSH_FREE( median);
    XSH_FREE( coadd);
    xsh_unwrap_vector( &coadd_vect);
    xsh_unwrap_vector( &med_vect);
    xsh_free_image( &med_img);
    TEST_END();
    return ret;
}

/**@}*/
