/*                                                                            *
 *   This file is part of the ESO X-shooter Pipeline                          *
 *   Copyright (C) 2006 European Southern Observatory                         *
 *                                                                            *
 *   This library is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU General Public License as published by     *
 *   the Free Software Foundation; either version 2 of the License, or        *
 *   (at your option) any later version.                                      *
 *                                                                            *
 *   This program is distributed in the hope that it will be useful,          *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *   GNU General Public License for more details.                             *
 *                                                                            *
 *   You should have received a copy of the GNU General Public License        *
 *   along with this program; if not, write to the Free Software              *
 *   Foundation, 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA     *
 *                                                                            */

/*
 * $Author: amodigli $
 * $Date: 2010/10/18 13:14:50 $
 * $Revision: 1.13 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*----------------------------------------------------------------------------*/
/**
 * @defgroup xsh_compute_slice_dist  Compute Slice Offsets (xsh_geom_ifu)
 * @ingroup drl_functions
 *
 * Function ...
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*-----------------------------------------------------------------------------
  Includes
  -----------------------------------------------------------------------------*/

#include <math.h>
#include <xsh_drl.h>

#include <xsh_utils_table.h>
#include <xsh_badpixelmap.h>
#include <xsh_data_pre.h>
#include <xsh_dfs.h>
#include <xsh_pfits.h>
#include <xsh_error.h>
#include <xsh_msg.h>
#include <xsh_fit.h>
#include <xsh_data_instrument.h>
#include <xsh_data_localization.h>
#include <xsh_data_spectrum.h>
#include <xsh_data_slice_offset.h>
#include <xsh_data_rec.h>
#include <xsh_ifu_defs.h>

#include <cpl.h>

/*-----------------------------------------------------------------------------
  Functions prototypes
  -----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
  Implementation
  -----------------------------------------------------------------------------*/

cpl_frame* xsh_compute_slice_dist( cpl_frameset *loc_frame_set,
  cpl_frame *order_tab_frame, cpl_frame *slitmap_frame, 
  cpl_frameset *rec_frameset,
  double slicedist_lambda,
  xsh_instrument *instrument)
{
  cpl_frame * result = NULL ;
  cpl_frame *loc_frame = NULL;
  cpl_frame *rec_frame = NULL;
  xsh_slice_offset *slice = NULL ;
  xsh_localization *loc_down = NULL, *loc_cen = NULL, *loc_up = NULL;
  char fname[256];
  double slit_down, slit_cen, slit_up;
  double sdown, sldown, slup, sup;
  double dist_up, dist_down;
  double lambda;
  //xsh_order_list *orderlist = NULL ;
  xsh_rec_list *reclist = NULL;

  XSH_ASSURE_NOT_NULL( loc_frame_set);
  XSH_ASSURE_NOT_NULL( rec_frameset);
  XSH_ASSURE_NOT_NULL( instrument);

  xsh_msg( "Get Localization table for slitlet %s",
	   SlitletName[LOWER_IFU_SLITLET] ) ;
  check( loc_frame = cpl_frameset_get_frame( loc_frame_set, 0));
  xsh_msg( "     '%s'", cpl_frame_get_filename( loc_frame ) ) ;
  check( loc_down = xsh_localization_load( loc_frame));

  xsh_msg( "Get Localization table for slitlet %s",
	   SlitletName[CENTER_IFU_SLITLET] ) ;
  check( loc_frame = cpl_frameset_get_frame( loc_frame_set, 1));
  xsh_msg( "     '%s'", cpl_frame_get_filename( loc_frame ) ) ;
  check( loc_cen = xsh_localization_load( loc_frame));

  xsh_msg( "Get Localization table for slitlet %s",
	   SlitletName[UPPER_IFU_SLITLET] ) ;
  check( loc_frame = cpl_frameset_get_frame( loc_frame_set, 2));
  xsh_msg( "     '%s'", cpl_frame_get_filename( loc_frame ));
  check( loc_up = xsh_localization_load( loc_frame));

  check( rec_frame = cpl_frameset_get_frame( rec_frameset, 0));
  check( reclist = xsh_rec_list_load( rec_frame, instrument));
 
  check( slice = xsh_slice_offset_create());

  if ( slicedist_lambda < 0){
    check( lambda = xsh_rec_list_get_lambda_max( reclist));
  }
  else{
    lambda = slicedist_lambda;
  }
  xsh_msg(" Evaluate at lambda %f", lambda);
  check( slit_down = cpl_polynomial_eval_1d(loc_down->cenpoly, 
    lambda, NULL));
  check( slit_cen = cpl_polynomial_eval_1d(loc_cen->cenpoly, 
    lambda, NULL));
  check( slit_up = cpl_polynomial_eval_1d(loc_up->cenpoly, 
    lambda, NULL));

  xsh_msg("SLIT center from localization [%f,%f,%f]", slit_down, slit_cen,
    slit_up);

  check( xsh_get_slit_edges(
    slitmap_frame, &sdown, &sldown, &slup, &sup, instrument));

  xsh_msg("EDGES form order tab [%f,[%f %f],%f]",sdown, sldown,
    slup, sup);

  dist_up = 2*slup-slit_cen-slit_up;
  dist_down = 2*sldown-slit_cen-slit_down;
  xsh_msg("DISTANCES edgU-->Up = %f edgU-->cen %f ==> %f", slup-slit_up,
    slup-slit_cen, dist_up);
  xsh_msg("DISTANCES edgD-->Down = %f edgD-->cen %f ==> %f", sldown-slit_down,
    sldown-slit_cen, dist_down);

  slice->cen_up = dist_up;
  slice->cen_down = dist_down;

  sprintf( fname, "SLICE_OFFSET_%s.fits", 
    xsh_instrument_arm_tostring( instrument));
  check( result = xsh_slice_offset_save( slice, fname, instrument));

  cleanup:
    xsh_localization_free( &loc_down);
    xsh_localization_free( &loc_cen);
    xsh_localization_free( &loc_up);
    xsh_slice_offset_free( &slice);
    xsh_rec_list_free( &reclist);

    if ( cpl_error_get_code() != CPL_ERROR_NONE ) {
      xsh_free_frame( &result);
    }
    return result;
}
/*--------------------------------------------------------------------------*/

/**@}*/
