/*                                                                           *
 *   This file is part of the ESO X-shooter Pipeline                         *
 *   Copyright (C) 2006 European Southern Observatory                        *
 *                                                                           *
 *   This library is free software; you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by    *
 *   the Free Software Foundation; either version 2 of the License, or       *
 *   (at your option) any later version.                                     *
 *                                                                           *
 *   This program is distributed in the hope that it will be useful,         *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *   GNU General Public License for more details.                            *
 *                                                                           *
 *   You should have received a copy of the GNU General Public License       *
 *   along with this program; if not, write to the Free Software             *
 *   Foundation, 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA    *
 *                                                                           */

/*
 * $Author: amodigli $
 * $Date: 2011/03/04 13:02:24 $
 * $Revision: 1.177 $
 * $Name: HEAD $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*---------------------------------------------------------------------------*/
/**
 * @defgroup xsh_dfs  DFS related functions
 * @ingroup xsh_tools
 *
 */
/*---------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------
  Includes
  ----------------------------------------------------------------------------*/

#include <xsh_dfs.h>
#include <xsh_pfits.h>
#include <xsh_data_pre.h>
#include <xsh_data_pre_3d.h>
#include <xsh_error.h>
#include <xsh_msg.h>
#include <cpl.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <xsh_paf_save.h>
#include <xsh_utils_table.h>
#include <xsh_utils.h>
#include <xsh_data_spectrum.h>
/*----------------------------------------------------------------------------
  Function prototypes
  ----------------------------------------------------------------------------*/

static cpl_frame *xsh_find_frame (cpl_frameset * frames,
				  const char *tags[]);


/**@{*/



 /**
    @brief    Check if an input parameter has been changed by the user
    @param    p input parameter
    @return  if has not changed 0, else 1;
 */
int 
xsh_parameter_get_default_flag(const cpl_parameter* p) {
   int flag_gasgano=0;
   int flag_norm=0;
   int flag=0;
   cpl_type type =0;

   flag_norm = (cpl_parameter_get_default_flag(p) == 0) ? 1 : 0; 
   type=cpl_parameter_get_type(p);

   switch(type) {
      case CPL_TYPE_BOOL:
         flag_gasgano = (cpl_parameter_get_default_bool(p) ==
                 cpl_parameter_get_bool(p)) ? 1:0;
         break;
      case CPL_TYPE_INT:
         flag_gasgano = (cpl_parameter_get_default_int(p) == 
                 cpl_parameter_get_int(p)) ? 1:0;
         break;
      case CPL_TYPE_DOUBLE:
         flag_gasgano = (cpl_parameter_get_default_double(p) == 
                 cpl_parameter_get_double(p)) ? 1:0;
         break;
      case CPL_TYPE_STRING:
         flag_gasgano = (cpl_parameter_get_default_string(p) == 
                 cpl_parameter_get_string(p)) ? 1:0;
         break;

      default:
         xsh_msg_error("type not supported");
   }

   flag = (flag_gasgano && flag_norm) ? 0 : 1;
 
   return flag;
}

 /**
    @brief    Validate model cfg
    @param    mod model cfg frame
    @param    set input frameset
    @return   a new allocated frameset filled by frame tables.
 */
cpl_error_code 
xsh_validate_model_cfg(cpl_frame* mod, cpl_frameset* set) {

  const char* mod_name=NULL;
  const char* raw_name=NULL;
  cpl_frame* raw_frame=0;
  cpl_propertylist* mod_header=NULL;
  cpl_propertylist* raw_header=NULL;
  double mod_mjd=0;
  double raw_mjd=0;

  raw_frame=cpl_frameset_get_frame(set,0);

  mod_name=cpl_frame_get_filename(mod);
  mod_header=cpl_propertylist_load(mod_name,0);
  mod_mjd=xsh_pfits_get_mjdobs(mod_header);

  raw_name=cpl_frame_get_filename(raw_frame);
  raw_header=cpl_propertylist_load(mod_name,0);
  check(raw_mjd=xsh_pfits_get_mjdobs(mod_header));

  if(raw_mjd < mod_mjd) {

    xsh_msg_warning("Raw frame %s has MJD-OBS  prior than model cfg frame %s",
		    raw_name,mod_name);
    xsh_msg_warning("The user should use a model cfg frame corresponding to a more recent period");
  }

 cleanup:
   return cpl_error_get_code();

 }

 /**
    @brief    Extract frameset sub set containing only table frames
    @param    set input frameset
    @return   a new allocated frameset filled by frame tables.
 */
 cpl_frameset* 
 xsh_frameset_ext_table_frames(cpl_frameset* set) {
   cpl_frameset* result=NULL;
   int nset=0;
   int i=0;
   cpl_frame* frm=NULL;
   cpl_propertylist* plist=NULL;
   const char* name=NULL;
   int naxis=0;
 
   check(nset=cpl_frameset_get_size(set));
   result=cpl_frameset_new();
   for(i=0;i<nset;i++) {
     check(frm=cpl_frameset_get_frame(set,i));
     check(name=cpl_frame_get_filename(frm));
     check(plist=cpl_propertylist_load(name,0));
     check(naxis=xsh_pfits_get_naxis(plist));
     if(naxis==0) {
       check(cpl_frameset_insert(result,cpl_frame_duplicate(frm)));
     }
     xsh_free_propertylist(&plist);
   }
 
  cleanup:
 
   return result;
 }
 
/*----------------------------------------------------------------------------*/
/**
 * @brief
 *   Check if all SOF files exist
 *
 * @param frameset The input set-of-frames
 *
 * @return 1 if not all files exist, 0 if they all exist.
 *
 */
/*----------------------------------------------------------------------------*/
int xsh_dfs_files_dont_exist(cpl_frameset *frameset)
{
    const char *func = "dfs_files_dont_exist";
    cpl_frame  *frame;


    if (frameset == NULL) {
        cpl_error_set(func, CPL_ERROR_NULL_INPUT);
        return 1;
    }

    if (cpl_frameset_is_empty(frameset)) {
        return 0;
    }

    frame = cpl_frameset_get_first(frameset);

    while (frame) {
        if (access(cpl_frame_get_filename(frame), F_OK)) {
            cpl_msg_error(func, "File %s (%s) was not found", 
                          cpl_frame_get_filename(frame), 
                          cpl_frame_get_tag(frame));
            cpl_error_set(func, CPL_ERROR_FILE_NOT_FOUND);
        }

        frame = cpl_frameset_get_next(frameset);
    }

    if (cpl_error_get_code())
        return 1;

    return 0;
}

/**
   @brief    Extract frameset sub set containing only table frames
   @param    set input frameset
   @return   a new allocated frameset filled by frame tables.
*/
cpl_frameset*
xsh_frameset_ext_image_frames(cpl_frameset* set) {
  cpl_frameset* result=NULL;
  int nset=0;
  int i=0;
  cpl_frame* frm=NULL;
  cpl_propertylist* plist=NULL;
  const char* name=NULL;
  int naxis=0;

  check(nset=cpl_frameset_get_size(set));
  result=cpl_frameset_new();
  for(i=0;i<nset;i++) {
    check(frm=cpl_frameset_get_frame(set,i));
    check(name=cpl_frame_get_filename(frm));
    check(plist=cpl_propertylist_load(name,0));
    check(naxis=xsh_pfits_get_naxis(plist));
    if(naxis==2) {
      check(cpl_frameset_insert(result,cpl_frame_duplicate(frm)));
    }
    xsh_free_propertylist(&plist);
  }

 cleanup:

  return result;
}

/**
 @brief extract DRL specific frames from frameset
 @param    set input frameset
 @return frameset containing DRL specific frames
 */
cpl_frameset*
xsh_frameset_drl_frames(cpl_frameset* set) {
  cpl_frameset* result=NULL;
  int nset=0;
  int i=0;
  cpl_frame* frm=NULL;
  cpl_propertylist* plist=NULL;
  const char* name=NULL;
  int naxis=0;

  check(nset=cpl_frameset_get_size(set));
  result=cpl_frameset_new();
  for(i=0;i<nset;i++) {
    check(frm=cpl_frameset_get_frame(set,i));
    check(name=cpl_frame_get_filename(frm));
    check(plist=cpl_propertylist_load(name,0));
    check(naxis=xsh_pfits_get_naxis(plist));
    if(naxis==0) {
      check(cpl_frameset_insert(result,cpl_frame_duplicate(frm)));
    }
    xsh_free_propertylist(&plist);
  }

 cleanup:

  return result;
}
/**
   @brief    Make sure input frames is an even number (eventually removes the last from the input list)
   @param    raws input value
   @param    instrument arm setting
   @return   cpl_error_code
*/
cpl_error_code
xsh_ensure_raws_input_offset_recipe_is_proper(cpl_frameset* raws, 
					      xsh_instrument* instrument)
{
  int nraws=0;
  cpl_frameset* obj=NULL;
  cpl_frameset* sky=NULL;
  const char * obj_tag=NULL;
  const char * sky_tag=NULL;
  int nobj=0;
  int nsky=0;
  int i=0;
  int ndif=0;
  cpl_frame* frm=NULL;

  check(nraws=cpl_frameset_get_size(raws));

  assure(nraws>=2, CPL_ERROR_ILLEGAL_INPUT, "Too few input frames. At least one OBJ and a SKY frame are required" );

  if(xsh_instrument_get_mode(instrument) == XSH_MODE_IFU) {
      obj_tag = XSH_GET_TAG_FROM_ARM( XSH_OBJECT_IFU_OFFSET,instrument);
  } else {
	  obj_tag = XSH_GET_TAG_FROM_ARM( XSH_OBJECT_SLIT_OFFSET,instrument);
  }

  check(obj=xsh_frameset_extract(raws,obj_tag));
  nobj=cpl_frameset_get_size(obj);
  if(nobj==0) {
     if(xsh_instrument_get_mode(instrument) == XSH_MODE_IFU) {
        obj_tag = XSH_GET_TAG_FROM_ARM( XSH_STD_FLUX_IFU_OFFSET,instrument);
     } else {
        obj_tag = XSH_GET_TAG_FROM_ARM( XSH_STD_FLUX_SLIT_OFFSET,instrument);
     }
     xsh_free_frameset(&obj);
     check(obj=xsh_frameset_extract(raws,obj_tag));
     nobj=cpl_frameset_get_size(obj);
  }

  if(xsh_instrument_get_mode(instrument) == XSH_MODE_IFU) {
      sky_tag = XSH_GET_TAG_FROM_ARM( XSH_SKY_IFU,instrument);
  } else {
	  sky_tag = XSH_GET_TAG_FROM_ARM( XSH_SKY_SLIT,instrument);
  }
  xsh_free_frameset(&sky);
  check(sky=xsh_frameset_extract(raws,sky_tag));


 
  nsky=cpl_frameset_get_size(sky);

  assure(nobj>0, CPL_ERROR_ILLEGAL_INPUT, "Too few input obj frames. At least a OBJ frame is required" );

  assure(nsky>0, CPL_ERROR_ILLEGAL_INPUT, "Too few input sky frames. At least a SKY frame is required" );

  if(nobj>nsky) {
    xsh_msg("case nobj> nsky");
     ndif=nobj-nsky;
     for(i=0;i<ndif;i++) {
        check(frm=cpl_frameset_get_frame(obj,nobj-i-1));
        check(cpl_frameset_erase_frame(obj,frm));
     }
  } else if(nsky>nobj) {
     xsh_msg("case nsky> nobj");
     ndif=nsky-nobj;
     for(i=0;i<ndif;i++) {
        check(frm=cpl_frameset_get_frame(sky,nsky-i-1));
        check(cpl_frameset_erase_frame(sky,frm));
     }
  } else {
     ndif=nsky-nobj;
  xsh_free_frameset(&obj);
  xsh_free_frameset(&sky);
  return CPL_ERROR_NONE;

  }
  xsh_free_frameset(&raws);
  check(raws=cpl_frameset_duplicate(obj));
  check(xsh_frameset_merge(raws,sky));

 cleanup:
  xsh_free_frameset(&obj);
  xsh_free_frameset(&sky);

  return cpl_error_get_code();

}


/**
   @brief    Computes offsety
   @param    plist propertylist
   @return   offsety
*/

static double
xsh_get_offsety(cpl_propertylist* plist)
{

   double offsety=0;
   double ra_off=0;
   double dec_off=0;
   double posang=0;

   if(cpl_propertylist_has(plist,XSH_NOD_CUMULATIVE_OFFSETY)) {
      offsety=xsh_pfits_get_cumoffsety(plist);
   } else {
      ra_off=xsh_pfits_get_ra_cumoffset(plist);
      dec_off=xsh_pfits_get_dec_cumoffset(plist);
      posang = xsh_pfits_get_posang(plist);
      posang = posang/180.0* M_PI;
/*
  shift_pix = xsh_round_double((cos(-posang)*dec_off+
  sin(-posang)*ra_off)/bin_space);
*/
      offsety=(cos(-posang)*dec_off+
               sin(-posang)*ra_off);
      xsh_msg("ra_off=%f,dec_off=%f,offsety=%f,posang=%f",
              ra_off,dec_off,offsety,posang);
   }

   return offsety;

}

/**
   @brief    Extract frames that has same cumoffy as offset 
   @param    raws input value
   @param    offset offset value
   @return   frameset with matching frames
*/

cpl_frameset*
xsh_frameset_extract_offsety_matches(cpl_frameset* raws,const double offset)
{
   cpl_frame* frm=NULL;
   cpl_frameset* result=NULL;
   cpl_propertylist* plist=NULL;
   double offsety=0;
   const char* name=NULL;
   int nraw=0;
   int i=0;
 
   XSH_ASSURE_NOT_NULL_MSG(raws,"null input frameset");
   nraw=cpl_frameset_get_size(raws);
   result=cpl_frameset_new();
   for(i=0;i<nraw;i++) {
      frm=cpl_frameset_get_frame(raws,i);
      name=cpl_frame_get_filename(frm);
      plist=cpl_propertylist_load(name,0);
      offsety=xsh_get_offsety(plist);
       
      if(fabs(offsety-offset)<1.E-8) {
         cpl_frameset_insert(result,cpl_frame_duplicate(frm));
      }
      xsh_free_propertylist(&plist);
   }

 cleanup:
      xsh_free_propertylist(&plist);

  return result;

}



/**
   @brief    Extract frames that has different cumoffy than offset 
   @param    raws input value
   @param    offset offset value
   @return   frameset with not matching frames
*/

cpl_frameset*
xsh_frameset_extract_offsety_mismatches(cpl_frameset* raws,const double offset)
{
   cpl_frame* frm=NULL;
   cpl_frameset* result=NULL;
   cpl_propertylist* plist=NULL;
   double offsety=0;
   const char* name=NULL;
   int nraw=0;
   int i=0;

   double ra_off=0;
   double dec_off=0;
   double posang=0;
 
   XSH_ASSURE_NOT_NULL_MSG(raws,"null input frameset");
   nraw=cpl_frameset_get_size(raws);
   result=cpl_frameset_new();
   for(i=0;i<nraw;i++) {
      frm=cpl_frameset_get_frame(raws,i);
      name=cpl_frame_get_filename(frm);
      plist=cpl_propertylist_load(name,0);
      if(cpl_propertylist_has(plist,XSH_NOD_CUMULATIVE_OFFSETY)) {
         offsety=xsh_pfits_get_cumoffsety(plist);
      } else {
         ra_off=xsh_pfits_get_ra_cumoffset(plist);
         dec_off=xsh_pfits_get_dec_cumoffset(plist);
         posang = xsh_pfits_get_posang(plist);
         posang = posang/180.0* M_PI;
/*
         shift_pix = xsh_round_double((cos(-posang)*dec_off+
                                       sin(-posang)*ra_off)/bin_space);
*/
         offsety=(cos(-posang)*dec_off+
                  sin(-posang)*ra_off);
         xsh_msg("ra_off=%f,dec_off=%f,offsety=%f,posang=%f",
                 ra_off,dec_off,offsety,posang);
      }
      if(offsety!=offset) {
         cpl_frameset_insert(result,cpl_frame_duplicate(frm));
      }
      xsh_free_propertylist(&plist);
   }

 cleanup:
      xsh_free_propertylist(&plist);

  return result;

}

/**
   @brief    Make sure input frames is composed by nod a-b sequence 
   @param    raws input value
   @param    offset vector with different offset values
   @return   frameset with a-b sequence frames
*/
cpl_frameset*
xsh_extract_nod_pairs(cpl_frameset* raws, 
                      cpl_vector* offset)
{
  int nraws=0;
  cpl_frameset* set_a=NULL;
  cpl_frameset* set_b=NULL;
  cpl_frameset* result=NULL;
  cpl_propertylist* plist=NULL;
  double offsety_a=0;


  int na=0;
  int nb=0;
  int i=0;
  int j=0;
  int k=0;
  cpl_frame* frm_a=NULL;
  cpl_frame* frm_b=NULL;
  cpl_frame* frm_tmp=NULL;
  double mjd_obs_a=0;
  double mjd_obs_b=0;
  double mjd_obs_tmp=0;
  const char* name=NULL;
  int noff=0;
  double mjd_obs_ref=-999999.;
   XSH_ASSURE_NOT_NULL_MSG(raws,"null input frameset");

  check(nraws=cpl_frameset_get_size(raws));
  xsh_msg("nraws=%d",nraws);
  assure(nraws>=2, CPL_ERROR_ILLEGAL_INPUT, "Too few input frames. At least two NOD positions are required" );

  noff=cpl_vector_get_size(offset);
  result=cpl_frameset_new();
  for(i=0;i<noff;i++) {
     offsety_a=cpl_vector_get(offset,i);
     xsh_msg("offsety_a=%f",offsety_a);
     xsh_free_frameset(&set_a);
     xsh_free_frameset(&set_b);
     check(set_a=xsh_frameset_extract_offsety_matches(raws,offsety_a));
     check(set_b=xsh_frameset_extract_offsety_mismatches(raws,offsety_a));
  
     na=cpl_frameset_get_size(set_a);
     nb=cpl_frameset_get_size(set_b);
     xsh_msg("na=%d nb=%d",na,nb);
     for(k=0;k<na;k++) {
    
        check(frm_a=cpl_frameset_get_frame(set_a,k));
        check(name=cpl_frame_get_filename(frm_a));
        check(plist=cpl_propertylist_load(name,0));
        check(mjd_obs_a=xsh_pfits_get_mjdobs(plist));
        xsh_free_propertylist(&plist);


       /* get the closest in mjd-obs */
        mjd_obs_b=mjd_obs_ref;
        for(j=0;j<nb;j++) {

           check(frm_tmp=cpl_frameset_get_frame(set_b,j));
           check(name=cpl_frame_get_filename(frm_tmp));
           check(plist=cpl_propertylist_load(name,0));
           check(mjd_obs_tmp=xsh_pfits_get_mjdobs(plist));

           if(fabs(mjd_obs_tmp-mjd_obs_a) < fabs(mjd_obs_b-mjd_obs_a)) {
              mjd_obs_b=mjd_obs_tmp;
              frm_b=frm_tmp;
           }
        }


        cpl_frameset_erase_frame(raws,frm_a);

/*
        xsh_msg("before");
        xsh_frameset_dump(raws);
        xsh_msg("after");
        xsh_frameset_dump(raws);
        */
        cpl_frameset_erase_frame(raws,frm_b);
        cpl_frameset_insert(result,cpl_frame_duplicate(frm_a));
        cpl_frameset_insert(result,cpl_frame_duplicate(frm_b));
        xsh_free_frameset(&set_a);
        xsh_free_frameset(&set_b);
        check(set_a=xsh_frameset_extract_offsety_matches(raws,offsety_a));
        check(set_b=xsh_frameset_extract_offsety_mismatches(raws,offsety_a));
        na=cpl_frameset_get_size(set_a);
        nb=cpl_frameset_get_size(set_b);

        xsh_msg("check again na=%d nb=%d",na,nb);

     }
     xsh_msg("ok0 i=%d",i);

  }

 cleanup:
  xsh_free_frameset(&set_a);
  xsh_free_frameset(&set_b);
  xsh_free_propertylist(&plist);


  return result;

}




/**
   @brief    Make sure input frames is composed by obj-sky pairs
   @param    raws input value
   @param    instrument arm
   @return   cpl_error_code
*/
cpl_frameset*
xsh_extract_obj_and_sky_pairs(cpl_frameset* raws, xsh_instrument* instrument)
{
  int nraws=0;
  cpl_frameset* obj=NULL;
  cpl_frameset* sky=NULL;
  cpl_frameset* result=NULL;
  cpl_propertylist* plist=NULL;

  const char * obj_tag=NULL;
  const char * sky_tag=NULL;
  int nobj=0;
  int nsky=0;
  int i=0;
  int j=0;
  cpl_frame* frm_obj=NULL;
  cpl_frame* frm_sky=NULL;
  cpl_frame* frm_tmp=NULL;
  double mjdobs_obj=0;
  double mjdobs_sky=0;
  double mjdobs_tmp=0;
  const char* name=NULL;

  check(nraws=cpl_frameset_get_size(raws));
  xsh_msg("nraws=%d",nraws);
  assure(nraws>=2, CPL_ERROR_ILLEGAL_INPUT, "Too few input frames. At least one OBJ and a SKY frame are required" );
  if(xsh_instrument_get_mode(instrument) == XSH_MODE_IFU) {
      obj_tag = XSH_GET_TAG_FROM_ARM( XSH_OBJECT_IFU_OFFSET,instrument);
      sky_tag = XSH_GET_TAG_FROM_ARM( XSH_SKY_IFU,instrument);
  } else {
	  obj_tag = XSH_GET_TAG_FROM_ARM( XSH_OBJECT_SLIT_OFFSET,instrument);
	  sky_tag = XSH_GET_TAG_FROM_ARM( XSH_SKY_SLIT,instrument);
  }

  check(obj=xsh_frameset_extract(raws,obj_tag));
  check(sky=xsh_frameset_extract(raws,sky_tag));

  nobj=cpl_frameset_get_size(obj);
  nsky=cpl_frameset_get_size(sky);

  assure(nobj>0, CPL_ERROR_ILLEGAL_INPUT, "Too few input obj frames. At least a OBJ frame is required" );

  assure(nsky>0, CPL_ERROR_ILLEGAL_INPUT, "Too few input sky frames. At least a SKY frame is required" );

  result=cpl_frameset_new();
  for(i=0;i<nobj;i++) {
    check(frm_obj=cpl_frameset_get_frame(obj,i));
    check(name=cpl_frame_get_filename(frm_obj));
    check(plist=cpl_propertylist_load(name,0));
    check(mjdobs_obj=xsh_pfits_get_mjdobs(plist));
    xsh_free_propertylist(&plist);


    check(frm_sky=cpl_frameset_get_frame(sky,0));
    check(name=cpl_frame_get_filename(frm_sky));
    check(plist=cpl_propertylist_load(name,0));
    check(mjdobs_sky=xsh_pfits_get_mjdobs(plist));
    xsh_free_propertylist(&plist);
    cpl_frameset_insert(result,cpl_frame_duplicate(frm_obj));

    for(j=1;j<nsky;j++) {
      check(frm_tmp=cpl_frameset_get_frame(sky,j));
      check(name=cpl_frame_get_filename(frm_tmp));
      check(plist=cpl_propertylist_load(name,0));
      check(mjdobs_tmp=xsh_pfits_get_mjdobs(plist));

      if(fabs(mjdobs_tmp-mjdobs_obj) < fabs(mjdobs_sky-mjdobs_obj)) {
        mjdobs_sky=mjdobs_tmp;
        frm_sky=frm_tmp;
      }
    }
    cpl_frameset_insert(result,cpl_frame_duplicate(frm_sky));
  }


 cleanup:
  xsh_free_frameset(&obj);
  xsh_free_frameset(&sky);
  xsh_free_propertylist(&plist);

  return result;

}

/**
   @brief    Make sure input frames is an even number (eventually removes the last from the input list)
   @param    raws input value
   @return   cpl_error_code
*/
cpl_error_code
xsh_ensure_raws_number_is_even(cpl_frameset* raws)
{
  int nraws=0;
  cpl_frame* frame1=NULL;  
  cpl_frame* frame2=NULL;
  cpl_frame* frameN=NULL;
  cpl_propertylist* plist1=NULL;
  cpl_propertylist* plist2=NULL;
  cpl_propertylist* plistN=NULL;
  const char* name1=NULL;
  const char* name2=NULL;
  const char* nameN=NULL;
  double yshift1=0;
  double yshift2=0;
  double yshiftN=0;



  check(nraws=cpl_frameset_get_size(raws));
  if ( (nraws % 2) != 0 ) {
     xsh_msg_warning("odd number of frames: nraws=%d",nraws);
     check(frame1=cpl_frameset_get_frame(raws,0));
     check(frame2=cpl_frameset_get_frame(raws,1));
     check(frameN=cpl_frameset_get_frame(raws,nraws-1));
     check(name1=cpl_frame_get_filename(frame1));
     check(name2=cpl_frame_get_filename(frame2));
     check(nameN=cpl_frame_get_filename(frameN));
     check(plist1=cpl_propertylist_load(name1,0));
     check(plist2=cpl_propertylist_load(name2,0));
     check(plistN=cpl_propertylist_load(nameN,0));
     check(yshift1=xsh_get_offsety(plist1));
     check(yshift2=xsh_get_offsety(plist2));
     check(yshiftN=xsh_get_offsety(plistN));
    if(yshift1 == yshiftN) {
       xsh_msg_warning("yshift(Frame1)==yshift(FrameN)");
       if ( fabs(yshift1-yshift2) <= DBL_MIN ) {
          xsh_msg_warning("yshift(Frame1)!=yshift(Frame2)");
          xsh_msg_warning("Remove last from list. Frame %s",
                          cpl_frame_get_filename(frameN));
          cpl_frameset_erase_frame(raws,frameN);
       } else {
          xsh_msg_warning("yshift(Frame1)==yshift(Frame2)");
          xsh_msg_warning("Remove last from list. Frame %s",
                          cpl_frame_get_filename(frameN));
          cpl_frameset_erase_frame(raws,frameN);
       }
    } else {
       xsh_msg_warning("yshift(Frame1)!=yshift(FrameN)");
       if (fabs(yshift1 - yshift2) <= DBL_MIN) {
         xsh_msg_warning("yshift(Frame1)!=yshift(Frame2)");
         xsh_msg_warning("Remove last from list. Frame %s",
                          cpl_frame_get_filename(frameN));
          cpl_frameset_erase_frame(raws,frameN);
      } else {
         xsh_msg_warning("yshift(Frame1)==yshift(Frame2)");
         xsh_msg_warning("Remove first from list. Frame %s",
                          cpl_frame_get_filename(frame1));
          cpl_frameset_erase_frame(raws,frame1);
      }
    }
  }

 cleanup:
  xsh_free_propertylist(&plist1);
  xsh_free_propertylist(&plist2);
  xsh_free_propertylist(&plistN);

  return cpl_error_get_code();

}
/**
   @brief    Check if an error has happened and returns error kind and location
   @param    val input value
   @return   0 if no error is detected,-1 else
*/
int 
xsh_print_rec_status(const int val) {
   if(cpl_error_get_code() != CPL_ERROR_NONE) {
      xsh_msg_error("Recipe status at %d",val);
      xsh_msg_error("%s",(const char*) cpl_error_get_message());
      xsh_msg_error("%s",(const char*) cpl_error_get_where());
      return -1;
   }
   return 0;
}


/**
   @brief    Define a frame characteristics
   @param    fname frame filename
   @param    tag   frame tag
   @param    type  frame type
   @param    group frame group
   @param    level frame level
   @param    frame output frame
   @return   0 if no error is detected,-1 else
*/

void
xsh_frame_config(const char* fname, 
                 const char* tag, 
                 cpl_frame_type type, 
                 cpl_frame_group group, 
                 cpl_frame_level level, 
                 cpl_frame** frame)
{

  check(cpl_frame_set_filename(*frame,fname));
  check(cpl_frame_set_tag(*frame,tag));
  check(cpl_frame_set_type(*frame,type));
  check(cpl_frame_set_group(*frame,group));
  check(cpl_frame_set_level(*frame,level));

  cleanup:

  return;

}
/**
   @brief    Creates a frame with given characteristics
   @param    fname frame filename
   @param    tag   frame tag
   @param    type  frame type
   @param    group frame group
   @param    level frame level

   @return   0 if no error is detected,-1 else
*/
cpl_frame* 
xsh_frame_product(const char* fname, const char* tag, cpl_frame_type type, 
                  cpl_frame_group group,cpl_frame_level level)
{
  cpl_frame* frame=NULL;
  check(frame=cpl_frame_new());

  check(xsh_frame_config(fname,tag,type,group,level,&frame));

  return frame;

 cleanup:
    xsh_free_frame(&frame);

    return NULL;
}
/*-------------------------------------------------------------------------*/
/**
  @name     xsh_file_exists
  @memo     Find if a given file name corresponds to an existing file.
  @param    filename    Name of the file to look up.
  @return   int 1 if file exists, 0 if not
  @doc

  Find out if the given character string corresponds to a file that
  can be stat()'ed.
 */
/*--------------------------------------------------------------------------*/


int xsh_file_exists(const char * filename)
{
  int exists=0;
  FILE* fo=NULL;
  if ((fo=fopen(filename,"r"))==NULL) {
     exists=0;
   } else {
     exists=1;
   }
  fclose(fo);
   return exists;
}

/*---------------------------------------------------------------------------*/
/** 
 * @brief  Extract frames with given tag from frameset
 * @param  frames      frame set
 * @param  tag         to search for
 * @return newly allocated, possibly empty, frameset, or NULL on error
 */
/*---------------------------------------------------------------------------*/
cpl_frameset *
xsh_frameset_extract(const cpl_frameset *frames,
                       const char *tag)
{
   cpl_frameset *subset = NULL;
   const cpl_frame *f;



   assure( frames != NULL, CPL_ERROR_ILLEGAL_INPUT, "Null frameset" );
   assure( tag    != NULL, CPL_ERROR_ILLEGAL_INPUT, "Null tag" );
    
   subset = cpl_frameset_new();

   for (f = cpl_frameset_find_const(frames, tag);
        f != NULL;
        f = cpl_frameset_find_const(frames, NULL)) {

      cpl_frameset_insert(subset, cpl_frame_duplicate(f));
   }
  cleanup:
   return subset;
}

/**
   @brief    Extracts raw frames 
   @param    set1     The input frameset that is extended
   @param    set2     The input frameset that is merged to set1
   @return   CPL_ERROR_NONE iff ok
*/
/*----------------------------------------------------------------------------*/
cpl_error_code
xsh_frameset_merge(cpl_frameset * set1, cpl_frameset* set2)
{

   cpl_frame* frm_tmp=NULL;
   cpl_frame* frm_dup=NULL;

   passure(set1 != NULL, "Wrong input set");
  
   check(frm_tmp = cpl_frameset_get_first(set2));
   while (frm_tmp != NULL)
   {
      frm_dup=cpl_frame_duplicate(frm_tmp);
      cpl_frameset_insert(set1,frm_dup);
      frm_tmp = cpl_frameset_get_next(set2);
   }

  cleanup:
   return cpl_error_get_code();
}

/*---------------------------------------------------------------------------*/
/**
   @brief    Set the group as RAW or CALIB in a frameset and return the 
   instrument detected
   @param    set     the input frameset
   @return an instrument structure which contains arm, mode and lamp
*/
/*---------------------------------------------------------------------------*/
xsh_instrument* xsh_dfs_set_groups (cpl_frameset * set)
{
  cpl_frame * cur_frame = NULL;
  xsh_instrument* instrument = NULL;

  /* Check entries */
  assure (set != NULL, CPL_ERROR_NULL_INPUT, "Null frame set");
  assure (!cpl_frameset_is_empty(set),CPL_ERROR_ILLEGAL_INPUT,
	  "Empty frame set");
		  
  /* allocate instrument structure */
  instrument = xsh_instrument_new();
  
  /* Loop on frames */
  for (cur_frame = cpl_frameset_get_first (set);
       cur_frame != NULL; 
       cur_frame = cpl_frameset_get_next (set)) {
    const char *tag = NULL;

    check(tag = cpl_frame_get_tag (cur_frame));

    /* SET GROUP */
    /* Initialize group to none */
    check(cpl_frame_set_group (cur_frame, CPL_FRAME_GROUP_NONE));

    /* RAW frames */
    if (XSH_IS_RAW (tag)) {
      check(cpl_frame_set_group (cur_frame, CPL_FRAME_GROUP_RAW));
    }
    /* CALIB frames */
    else if (XSH_IS_CALIB (tag)) {
      check(cpl_frame_set_group (cur_frame, CPL_FRAME_GROUP_CALIB));
    }
    /* UNRECOGNIZED frames */
    else {
      xsh_msg_error("Unrecognized frame tag '%s'",tag);
      check(cpl_error_set(cpl_func,CPL_ERROR_ILLEGAL_INPUT ));
      //assure(0,CPL_ERROR_ILLEGAL_INPUT, 
      //     "Unrecognized frame tag: %s", tag);
    }
    /* we have validate the tag frame like an XSH tag */
    
    /* CHECK INSTRUMENT */
   check_msg(xsh_instrument_parse_tag(instrument,tag),
	      "invalid tag %s in file %s",
	      tag, cpl_frame_get_filename (cur_frame));
  }
  /* validate arm */
  assure(xsh_instrument_get_arm(instrument)!=XSH_ARM_UNDEFINED,
         CPL_ERROR_TYPE_MISMATCH,"Arm is undefined");
  xsh_instrument_get_config( instrument ) ;
  xsh_msg( "Arm %s, Mode %s Nb Orders %d, min: %d, max: %d",
	   xsh_instrument_arm_tostring(instrument),
           xsh_instrument_mode_tostring( instrument),
	   instrument->config->orders, instrument->config->order_min,
	   instrument->config->order_max ) ;
 cleanup:
  if (cpl_error_get_code () != CPL_ERROR_NONE){
    xsh_instrument_free(&instrument);
    instrument = NULL;
  }
  return instrument;
}
/*---------------------------------------------------------------------------*/
/**
   @brief    split input sof in groups: raw and calib
   @param    input the input set of files
   @param    raws the raws file of the input set of files
   @return   cpl error code
*/
/*---------------------------------------------------------------------------*/
cpl_error_code
xsh_dfs_extract_raw_frames (cpl_frameset * input, cpl_frameset * raws)
{
  cpl_frame *cur_frame = NULL;

  /* Loop on frames */
  for (cur_frame = cpl_frameset_get_first (input);
       cur_frame != NULL; cur_frame = cpl_frameset_get_next (input)) {
    /* RAW frames */
    if (cpl_frame_get_group (cur_frame) == CPL_FRAME_GROUP_RAW) {
      cpl_frameset_insert (raws, cpl_frame_duplicate(cur_frame));
    }
  }

return cpl_error_get_code();

}
/*---------------------------------------------------------------------------*/
/**
   @brief    split input sof in groups: raw and calib
   @param    input the input set of files
   @param    calib the calib files of the input set of files
   @return   cpl error code
*/
/*---------------------------------------------------------------------------*/
cpl_error_code
xsh_dfs_extract_calib_frames (cpl_frameset * input, cpl_frameset * calib)
{
  cpl_frame *cur_frame = NULL;

  /* Loop on frames */
  for (cur_frame = cpl_frameset_get_first (input);
       cur_frame != NULL; cur_frame = cpl_frameset_get_next (input)) {
    /* RAW frames */
    if (cpl_frame_get_group (cur_frame) == CPL_FRAME_GROUP_CALIB) {
      cpl_frameset_insert (calib, cpl_frame_duplicate(cur_frame));
    }
  }

return cpl_error_get_code();

}

/*---------------------------------------------------------------------------*/
/**
   @brief    split input sof in groups: raw and calib
   @param    input the input set of files
   @param    pros  the product files of the input set of files
   @return   cpl error code
*/
/*---------------------------------------------------------------------------*/
cpl_error_code
xsh_dfs_extract_pro_frames (cpl_frameset * input, cpl_frameset * pros)
{
  cpl_frame *cur_frame = NULL;
  XSH_ASSURE_NOT_NULL_MSG(pros,"Null pros frameset. Alllocated it outside!");
  /* Loop on frames */
  for (cur_frame = cpl_frameset_get_first (input);
       cur_frame != NULL; cur_frame = cpl_frameset_get_next (input)) {
    /* PRO frames */
    if (cpl_frame_get_group (cur_frame) == CPL_FRAME_GROUP_PRODUCT) {
      cpl_frameset_insert (pros, cpl_frame_duplicate(cur_frame));
    }
  }
  cleanup:
  return cpl_error_get_code();

}

/*---------------------------------------------------------------------------*/
/**
   @brief    split input sof in groups: raw and calib
   @param    input the input set of files
   @param    raws the raws file of the input set of files
   @param    calib the calib files of the input set of files
*/
/*---------------------------------------------------------------------------*/
void
xsh_dfs_split_in_group (cpl_frameset * input, cpl_frameset * raws,
			cpl_frameset * calib)
{
  cpl_frame *cur_frame = NULL;

  /* Loop on frames */
  for (cur_frame = cpl_frameset_get_first (input);
       cur_frame != NULL; cur_frame = cpl_frameset_get_next (input)) {
    /* RAW frames */
    if (cpl_frame_get_group (cur_frame) == CPL_FRAME_GROUP_RAW) {
      //xsh_dfs_check_nan(cur_frame);
      cpl_frameset_insert (raws, cpl_frame_duplicate(cur_frame));
    }
    /* CALIB frames */
    else if (cpl_frame_get_group (cur_frame) == CPL_FRAME_GROUP_CALIB) {
      cpl_frameset_insert (calib, cpl_frame_duplicate(cur_frame));
    }
    /* UNRECOGNIZED frames */
    else {
      xsh_msg_error ("Unrecognized group for frame %s",
		     cpl_frame_get_filename (cur_frame));
    }
  }
}


/*---------------------------------------------------------------------------*/
/**
   @brief Extracts QTH and D2 frames from input frameset
   @param input the input set of files
   @param qth frameset with QTH frames
   @param d2 frameset with D2 frames
   @return void
*/
/*---------------------------------------------------------------------------*/
void 
xsh_dfs_split_qth_d2(cpl_frameset * input, 
		     cpl_frameset** qth,
		     cpl_frameset** d2)
{
  cpl_frame* cur_frame = NULL;
  cpl_frame* temp =NULL;

  /* check parameters input */
  XSH_ASSURE_NOT_NULL( input);
  XSH_ASSURE_NOT_NULL( qth);
  XSH_ASSURE_NOT_NULL( d2);

  /* create frameset */
  XSH_NEW_FRAMESET( *qth);
  XSH_NEW_FRAMESET( *d2);

  /* Loop on frames */
  for (cur_frame = cpl_frameset_get_first (input);
       cur_frame != NULL; cur_frame = cpl_frameset_get_next (input)) {
    const char* tag = NULL;

    check(tag = cpl_frame_get_tag(cur_frame));
    /* ON frames */
    if ( strstr(tag,"D2") != NULL) {
      check(temp = cpl_frame_duplicate(cur_frame)); 
      check(cpl_frameset_insert(*d2, temp));
    }
    /* OFF frames */
    else if ( strstr(tag, "QTH") != NULL) {
      check(temp = cpl_frame_duplicate(cur_frame));
      check(cpl_frameset_insert(*qth, temp));
    }
    /* UNRECOGNIZED frames */
    else {
      xsh_msg_error ("Invalid tag %s for frame %s",tag,
        cpl_frame_get_filename (cur_frame));
    }
  }

  cleanup:
    if (cpl_error_get_code() != CPL_ERROR_NONE){
      xsh_free_frame(&temp);
      xsh_free_frameset(d2);
      xsh_free_frameset(qth);
    }
    return;
}
/*---------------------------------------------------------------------------*/
/**
   @brief split input RAW NIR sof in ON and OFF
   @param[in] input the input RAW NIR set of files
   @param[out] on   the on files of the input set of files
   @param[out] off  the off files of the input set of files
*/
/*---------------------------------------------------------------------------*/
void xsh_dfs_split_nir(cpl_frameset * input, cpl_frameset** on,
  cpl_frameset** off)
{
  cpl_frame* cur_frame = NULL;
  cpl_frame* temp =NULL;

  /* check parameters input */
  XSH_ASSURE_NOT_NULL(input);
  XSH_ASSURE_NOT_NULL(on);
  XSH_ASSURE_NOT_NULL(off);

  /* create frameset */
  XSH_NEW_FRAMESET(*on);
  XSH_NEW_FRAMESET(*off);
 
  /* Loop on frames */
  for (cur_frame = cpl_frameset_get_first (input);
       cur_frame != NULL; cur_frame = cpl_frameset_get_next (input)) {
    const char* tag = NULL;

    check(tag = cpl_frame_get_tag(cur_frame));
    /* ON frames */
    if ( strstr(tag,"ON") != NULL) {
      check(temp = cpl_frame_duplicate(cur_frame)); 
      check(cpl_frameset_insert(*on, temp));
    }
    /* OFF frames */
    else if ( strstr(tag, "OFF") != NULL) {
      check(temp = cpl_frame_duplicate(cur_frame));
      check(cpl_frameset_insert(*off, temp));
    }
    /* UNRECOGNIZED frames */
    else {
      xsh_msg_error ("Invalid tag %s for frame %s",tag,
        cpl_frame_get_filename (cur_frame));
    }
  }

  cleanup:
    if (cpl_error_get_code() != CPL_ERROR_NONE){
      xsh_free_frame(&temp);
      xsh_free_frameset(on);
      xsh_free_frameset(off);
    }
    return;
}

/*---------------------------------------------------------------------------*/
/**
   @brief split input RAW offset sof in OBJECT_... and SKY_... framesets
   @param[in] input input frameset
   @param[in] object_tag  The OBJECT Tag
   @param[in] sky_tag The SKY Tag
   @param[out] object the object files of the input set of files
   @param[out] sky the sky files of the input set of files

   @return Number of OBJECT frames (equal nb of SKY frames)
*/
/*---------------------------------------------------------------------------*/
int 
xsh_dfs_split_offset(cpl_frameset * input, 
		     const char * object_tag,
		     const char * sky_tag,
		     cpl_frameset ** object,
		     cpl_frameset ** sky)
{
  cpl_frame* cur_frame = NULL;
  cpl_frame* temp =NULL;
  int nobj = 0, nsky = 0 ;

  /* check parameters input */
  XSH_ASSURE_NOT_NULL(input);
  XSH_ASSURE_NOT_NULL(object);
  XSH_ASSURE_NOT_NULL(sky);

  /* create frameset */
  *object=cpl_frameset_new();
  *sky=cpl_frameset_new();

  /* Loop on frames */
  for (cur_frame = cpl_frameset_get_first (input);
       cur_frame != NULL; cur_frame = cpl_frameset_get_next (input)) {
    const char* tag = NULL;

    check(tag = cpl_frame_get_tag(cur_frame));
    /* OBJECT frames */
    xsh_msg_dbg_medium( "Tag: %s", tag ) ;

    if ( strstr(tag, object_tag) != NULL) {
      check(temp = cpl_frame_duplicate(cur_frame)); 
      check(cpl_frameset_insert(*object, temp));
      nobj++ ;
    }
    /* OFF frames */
    else if ( strstr(tag, sky_tag) != NULL) {
      check(temp = cpl_frame_duplicate(cur_frame));
      check(cpl_frameset_insert(*sky, temp));
      nsky++ ;
    }
    /* UNRECOGNIZED frames */
    else {
      xsh_msg_error ("Invalid tag %s for frame %s",tag,
        cpl_frame_get_filename (cur_frame));
    }

  }

  xsh_msg_dbg_medium( "Nobj: %d, Nsky: %d", nobj, nsky ) ;
  XSH_ASSURE_NOT_ILLEGAL( nobj == nsky ) ;

  cleanup:
    if (cpl_error_get_code() != CPL_ERROR_NONE){
       xsh_print_rec_status(0);
      xsh_free_frame(&temp);
      return 0 ;
    }
    return nobj ;
}



/*---------------------------------------------------------------------------*/
/**
  @brief
    check if the tag is in the tag list
  @param tag
    the tag to check
  @param tag_list
    the list of tags
  @param size size of list
  @return
    TRUE if the tag is find else false
*/
/*---------------------------------------------------------------------------*/
static bool 
xsh_dfs_tag_check( const char* tag, const char* tag_list[], int size)
{
  int i;

  for( i=0; i< size; i++){
    if ( strstr( tag, tag_list[i])!=NULL){
      return true;
    }
  }
  return false;  
}

/*---------------------------------------------------------------------------*/
/**
  @brief 
    filter the frameset with the given tag list
  @param set [IN/OUT] 
    The input set of frames
  @param tags 
    the filtered tag list
  @param size size of list
  @return void
*/
/*---------------------------------------------------------------------------*/
void 
xsh_dfs_filter( cpl_frameset * set, const char* tags[], int size){
  cpl_frame *cur_frame = NULL;

  XSH_ASSURE_NOT_NULL( set);
  XSH_ASSURE_NOT_NULL( tags);
  XSH_ASSURE_NOT_ILLEGAL( size > 0);

  /* Loop on frames */
  check( cur_frame = cpl_frameset_get_first (set));

  while(cur_frame != NULL){
    const char* ftag = NULL;

    check( ftag = cpl_frame_get_tag( cur_frame));
    if (!xsh_dfs_tag_check( ftag, tags, size)){
      cpl_frame *del_frame = NULL;
      
      del_frame = cur_frame;
      check( cur_frame = cpl_frameset_get_next( set));
      check( cpl_frameset_erase_frame( set, del_frame));
    }
    else{
      check(cur_frame = cpl_frameset_get_next (set));
    }  
  }

  cleanup:
    return;
}   

/*---------------------------------------------------------------------------*/
/**
   @brief    Extracts pre frames from a frameset 
   @param    frameset the set of files
   @param    prefix the pre frame prefix 

   @return   the frameset with pre frames
*/
/*---------------------------------------------------------------------------*/
cpl_frameset *
xsh_frameset_extract_pre(cpl_frameset * frameset, const char* prefix)
{
   cpl_frameset* set=NULL;
   cpl_frame* current=NULL;
   cpl_frame* found=NULL;
   int sz=0;
   int i=0;
   char tag[80];
   char name[80];

   check(sz=cpl_frameset_get_size(frameset));
 
   check(set=cpl_frameset_new());
   for(i=0;i<sz;i++) {
    check( current = cpl_frameset_get_frame(frameset, i));
    sprintf(name, "%s_PRE_%d.fits", prefix, i);
    sprintf(tag, "%s_PRE_%d", prefix, i);

    if(xsh_file_exists(name) == 1) {
       check(found=cpl_frame_duplicate(current));
       check( cpl_frame_set_filename(found, name) );
       cpl_frameset_insert(set,found);
    }

   }
  cleanup:

   return set;

}

/*---------------------------------------------------------------------------*/
/**
   @brief    find the bad pixel map in a set of files
   @param    set the set of files
   @return   the bad pixel map or NULL if not found
*/
/*---------------------------------------------------------------------------*/
cpl_frame *
xsh_find_bpmap (cpl_frameset * set)
{
  cpl_frame *cur_frame = NULL;

  /* Loop on frames */
  for (cur_frame = cpl_frameset_get_first (set);
       cur_frame != NULL; cur_frame = cpl_frameset_get_next (set)) {
    if (strstr (cpl_frame_get_tag (cur_frame), "BADPIXEL_MAP")) {
      return cur_frame;
    }
  }
  return NULL;
}


/*---------------------------------------------------------------------------*/
/**
 *    @brief    find the master bad pixel map in a set of files
 *    @param    set the set of files
 *    @return   the bad pixel map or NULL if not found
 */
/*---------------------------------------------------------------------------*/
cpl_frame *
xsh_find_master_bpmap (cpl_frameset * set)
{
  cpl_frame *cur_frame = NULL;

  /* Loop on frames */
  for (cur_frame = cpl_frameset_get_first (set);
       cur_frame != NULL; cur_frame = cpl_frameset_get_next (set)) {
    if (strstr (cpl_frame_get_tag (cur_frame), XSH_MASTER_BP_MAP)) {
      return cur_frame;
    }
  }
  return NULL;
}

/*---------------------------------------------------------------------------*/
/**
   @brief    find the off frame in set of files
   @param    set the set of files
   @return   the off frame or NULL if not found
*/
/*---------------------------------------------------------------------------*/
cpl_frame * xsh_find_off(cpl_frameset* set)
{
  cpl_frame *cur_frame = NULL;

  /* Loop on frames */
  for (cur_frame = cpl_frameset_get_first (set);
       cur_frame != NULL; cur_frame = cpl_frameset_get_next (set)) {
    if (strstr (cpl_frame_get_tag (cur_frame), "OFF")) {
      return cur_frame;
    }
  }
  return NULL;
}

/*---------------------------------------------------------------------------*/
/**
   @brief    find the on frame in set of files
   @param    set the set of files
   @return   the on frame or NULL if not found
*/
/*---------------------------------------------------------------------------*/
cpl_frame * xsh_find_on(cpl_frameset* set)
{
  cpl_frame *cur_frame = NULL;

  /* Loop on frames */
  for (cur_frame = cpl_frameset_get_first (set);
       cur_frame != NULL; cur_frame = cpl_frameset_get_next (set)) {
    if (strstr (cpl_frame_get_tag (cur_frame), "ON")) {
      return cur_frame;
    }
  }
  return NULL;
}
/*---------------------------------------------------------------------------*/
/**
   @brief find the flat set of files in the input SOF and detect the mode of
   instrument. This function assure that all flat files have the same tag
   @param raws set of file where we want to do the search
   @param flat frameset contains only flat
*/
/*---------------------------------------------------------------------------*/
void
xsh_dfs_find_flat(cpl_frameset* raws,cpl_frameset* flat)
{
  cpl_frame *cur_frame = NULL;
  const char* tag = NULL;
  /* check input */
  assure (raws != NULL, CPL_ERROR_NULL_INPUT, "Null frameset");
  assure (flat != NULL, CPL_ERROR_NULL_INPUT, "Null frameset");
  
  /* init */
  tag = cpl_frame_get_tag(cpl_frameset_get_first(raws)); 
  
  /* Loop on frames */
  for (cur_frame = cpl_frameset_get_first (raws);
       cur_frame != NULL; cur_frame = cpl_frameset_get_next (raws))
    {
      tag = cpl_frame_get_tag (cur_frame);
      if (strstr(tag,"FLAT")!=NULL)
	cpl_frameset_insert (flat, cur_frame);
    }
  
 cleanup:
  return;
  
}

/*---------------------------------------------------------------------------*/
/**
   @brief    Get unique product filename
   @param    context                 The current context, or NULL
   @param    caller_id               Name of calling module
   @param    pro_catg                The PROduct CATeGory
   @return   Newly allocated unique filename.

   This function can be used to avoid overwriting existing files. It computes
   the filename depending on the current context (position in the call-tree).

   If this turns out not to be a problem, one could simply use pro_catg.fits
   as the filename.

*/
/*---------------------------------------------------------------------------*/
char *
xsh_unique_filename (const char *context, const char *caller_id,
		     const char *pro_catg)
{
  if (context != NULL) {
    //fixme: if context contains '.' replace it with '-'
    return xsh_stringcat_6 (context, "-", caller_id, "-", pro_catg, ".fits");
  }
  else {
    return xsh_stringcat_4 (caller_id, "-", pro_catg, ".fits");
  }
}


/** 
 * Save product
 * 
 * @param frame Product Frame
 * @param frameset Input Frameset
 * @param parameters Parameters list
 * @param recipe_id Recipe Identification String
 * @param instrument Instrument structure
 * @param final_prefix Final file name
 * @param type of frame (image/imagelist/table)
 */
void xsh_add_product(cpl_frame *frame, 
                     cpl_frameset * frameset,
		     const cpl_parameterlist * parameters,
		     const char *recipe_id,
		     xsh_instrument* instrument,
		     const char *final_prefix ,
		     const char* type)
{
  cpl_propertylist *primary_header = NULL;
  cpl_propertylist *tbl_header = NULL;
  int extension=0;
  const char *pro_catg = NULL;
  cpl_frame* product_frame = NULL;

  cpl_table *table = NULL ;
  cpl_image *image = NULL ;
  cpl_imagelist *imagelist = NULL ;


  const char *fname = NULL;
  char *final_name = NULL ;
  time_t now ;
  char* date = NULL ;
  /* char *paf_name = NULL ; */

  /* Check input parameters */
  XSH_ASSURE_NOT_NULL(frame);
  XSH_ASSURE_NOT_NULL(frameset);
  XSH_ASSURE_NOT_NULL(parameters);
  XSH_ASSURE_NOT_NULL(recipe_id);
  XSH_ASSURE_NOT_NULL(instrument);
  XSH_ASSURE_NOT_NULL(final_prefix);
 
  check(pro_catg = cpl_frame_get_tag (frame));
  XSH_ASSURE_NOT_NULL(pro_catg);
  
  /* Load header and imagelist of bpmap frame */
  check(fname = cpl_frame_get_filename( frame ));
  check(primary_header = cpl_propertylist_load( fname, 0 ));

  if(strcmp(type,"imagelist") == 0) {
    check(imagelist = cpl_imagelist_load( fname, CPL_TYPE_FLOAT, 0 ));
  } else if(strcmp(type,"image") == 0) {
    check(image = cpl_image_load( fname, CPL_TYPE_FLOAT, 0,0 ));
  } else if(strcmp(type,"table") == 0) {
    check(table = cpl_table_load( fname, 1,0 ));
  } else {
    xsh_msg_error("Frame type %s not supported",type);
  }

  /* update header */
  check( cpl_frame_set_group( frame, CPL_FRAME_GROUP_PRODUCT)); 
  check_msg (cpl_dfs_setup_product_header ( primary_header,
					    frame,frameset,parameters,
					    recipe_id,
					    instrument->pipeline_id,
					    instrument->dictionary,NULL),
	     "Problem in the product DFS-compliance");

  /* Create final name */
  if ( xsh_time_stamp_get() ) {
    time( &now ) ;
    date = xsh_sdate_utc( &now ) ;
    final_name = xsh_stringcat_any( XSH_PRODUCT_PREFIX,
      final_prefix, "_", date,
				    ".fits", "" ) ;
  }
  else
    final_name = xsh_stringcat_any( XSH_PRODUCT_PREFIX,
      final_prefix, ".fits", "" ) ;

  xsh_msg_dbg_low( "Final product name: %s", final_name ) ;



  /* Now save file with updated header */
  if(strcmp(type,"imagelist") == 0) {
    cpl_imagelist_save(imagelist, final_name, CPL_BPP_IEEE_FLOAT, 
		       primary_header,CPL_IO_DEFAULT ) ;
  } else if(strcmp(type,"image") == 0) {
    cpl_image_save(image, final_name, CPL_BPP_IEEE_FLOAT, 
		   primary_header,CPL_IO_DEFAULT ) ;
  } else if(strcmp(type,"table") == 0) {
    check(cpl_table_save( table, primary_header, tbl_header,
			  final_name, extension ));
  }




  /*
    Save PAF
  
  paf_name = xsh_stringcat_any( final_prefix, ".paf", NULL ) ;
  XSH_ASSURE_NOT_NULL( paf_name ) ;
  check_msg( xsh_paf_save( instrument, recipe_id,
			   primary_header, paf_name, pro_catg ),
	     "Cant save PAF file" ) ;

  */


  /* Create product frame */
  if(strcmp(type,"imagelist") == 0) {
  check(product_frame=xsh_frame_product(final_name,pro_catg,
					CPL_FRAME_TYPE_IMAGE,
					CPL_FRAME_GROUP_PRODUCT,
					CPL_FRAME_LEVEL_FINAL));
  } else if(strcmp(type,"image") == 0) {
  check(product_frame=xsh_frame_product(final_name,pro_catg,
					CPL_FRAME_TYPE_IMAGE,
					CPL_FRAME_GROUP_PRODUCT,
					CPL_FRAME_LEVEL_FINAL));
  } else if(strcmp(type,"table") == 0) {
  check(product_frame=xsh_frame_product(final_name,pro_catg,
					CPL_FRAME_TYPE_TABLE,
					CPL_FRAME_GROUP_PRODUCT,
					CPL_FRAME_LEVEL_FINAL));

  }

  check(cpl_frameset_insert(frameset, product_frame));
  xsh_add_product_file(final_name);
 cleanup:
  XSH_FREE(date);  
  xsh_free_propertylist(&primary_header);
  xsh_free_image( &image ) ;
  xsh_free_imagelist( &imagelist ) ;
  xsh_free_table( &table ) ;
  XSH_FREE( final_name ) ;
  /* XSH_FREE( paf_name ) ; */
  return;
}

/** 
 * Save spectrum product
 * 
 * @param frame Product Frame
 * @param frameset Input Frameset
 * @param parameters Parameters list
 * @param recipe_id Recipe Identification String
 * @param instr Instrument structure
 */
void 
xsh_add_product_spectrum( cpl_frame *frame, 
			  cpl_frameset *frameset,
			  const cpl_parameterlist *parameters,
			  const char *recipe_id,
			  xsh_instrument* instr)
{

  const char *pro_catg = "";
  xsh_spectrum *spectrum = NULL;
  cpl_frame *product_frame = NULL;
  char product_id[256];
  char* product_name = NULL;
  /* char* paf_name = NULL; */
  time_t now ;
  char* date = NULL;

  /* Check entries */
  XSH_ASSURE_NOT_NULL( frame);
  XSH_ASSURE_NOT_NULL( frameset);
  XSH_ASSURE_NOT_NULL( parameters);
  XSH_ASSURE_NOT_NULL( recipe_id);
  XSH_ASSURE_NOT_NULL( instr);
 
  check( pro_catg = cpl_frame_get_tag (frame));

  assure( pro_catg != NULL, CPL_ERROR_NULL_INPUT,
    "Frame tag has not been set");

  /* Load PRE */
  check( spectrum = xsh_spectrum_load( frame, instr));

  /* update header */
  cpl_frame_set_group( frame, CPL_FRAME_GROUP_PRODUCT);
  check( cpl_dfs_setup_product_header( spectrum->flux_header,
    frame, frameset, parameters, recipe_id, instr->pipeline_id,
				       instr->dictionary, NULL));
  /* Save product */
  if (xsh_time_stamp_get() ){
    time( &now );
    date = xsh_sdate_utc(&now);
    sprintf( product_id, "%s%s_%s", XSH_PRODUCT_PREFIX,
      pro_catg, date);
  }
  else {
    sprintf( product_id, "%s%s", XSH_PRODUCT_PREFIX, 
      pro_catg);
  }
  product_name = xsh_stringcat_any( product_id, ".fits", NULL);
  check( product_frame = xsh_spectrum_save(spectrum, product_name,pro_catg));

  /* Save PAF 
  paf_name = xsh_stringcat_any( product_id, ".paf", NULL);
  XSH_ASSURE_NOT_NULL( paf_name);
  check_msg( xsh_paf_save( instr, recipe_id,
			   spectrum->flux_header, paf_name,
			   pro_catg ),
    "Cant save PAF file" ) ;
  */

  /* Add the new frame into the frameset */
  cpl_frame_set_type( product_frame, CPL_FRAME_TYPE_IMAGE);
  cpl_frame_set_group( product_frame, CPL_FRAME_GROUP_PRODUCT);
  cpl_frame_set_level( product_frame, CPL_FRAME_LEVEL_FINAL);
  cpl_frame_set_tag( product_frame, pro_catg);
  cpl_frameset_insert( frameset, product_frame);
  xsh_add_product_file(product_name);

  cleanup:
    if (cpl_error_get_code () != CPL_ERROR_NONE) {
      xsh_free_frame( &product_frame);
      product_frame = NULL;
    }
    XSH_FREE( date);
    xsh_spectrum_free( &spectrum);
    XSH_FREE( product_name);
    /* XSH_FREE( paf_name); */
    return;
}

/*---------------------------------------------------------------------------*/
/**
  @brief 
    Add DFS keywords to a product frame
  @param frame The product frame
  @param frameset The recipe frameset.
  @param parameters The recipe parameter list
  @param recipe_id The recipe name
  @param instr The instrument settings  

   Mandatory DFS keywords are added by 
   calling @c cpl_dfs_setup_product_header()

   The function performs the following steps
   - Load primary header, infer product type,
   - Load product,
   - Modify primary header by adding DFS keywords,
   - Save product using the new header,
   - Insert the new frame into the frameset.

  @note
    A new frame is created and inserted into the frameset.
    The provided @em product_frame must still be deallocated.

*/
/*---------------------------------------------------------------------------*/
void 
xsh_add_product_pre (cpl_frame * frame, 
		     cpl_frameset * frameset,
		     const cpl_parameterlist *parameters, 
		     const char *recipe_id,
		     xsh_instrument* instr)
{
  const char *pro_catg = "";
  xsh_pre* product_pre = NULL;
  cpl_frame* product_frame = NULL;
  char product_id[256];
  char* product_name = NULL;
  /* char* paf_name = NULL; */
  time_t now ;
  char* date = NULL;
  cpl_propertylist* plist=NULL;
  char arcfile[80];
  bool arcfile_found=false;

  /* Check entries */
  XSH_ASSURE_NOT_NULL( frame);
  XSH_ASSURE_NOT_NULL( frameset);
  XSH_ASSURE_NOT_NULL( parameters);
  XSH_ASSURE_NOT_NULL( recipe_id);
  XSH_ASSURE_NOT_NULL( instr);
 
  check( pro_catg = cpl_frame_get_tag (frame));

  assure( pro_catg != NULL, CPL_ERROR_NULL_INPUT,
    "Frame tag has not been set");

  /* Load PRE */
  check( product_pre = xsh_pre_load( frame,instr));
  if(cpl_propertylist_has(product_pre->data_header,"ARCFILE")) {
  sprintf(arcfile,"%s",xsh_pfits_get_arcfile(product_pre->data_header));
  arcfile_found=true;
  }
  /* update header */
  cpl_frame_set_group( frame, CPL_FRAME_GROUP_PRODUCT);
  check( cpl_dfs_setup_product_header( product_pre->data_header,
    frame, frameset, parameters, recipe_id, instr->pipeline_id,
				       instr->dictionary,NULL));

  /* Save product */
  if (xsh_time_stamp_get() ){
    time( &now );
    date = xsh_sdate_utc(&now);
    sprintf( product_id, XSH_PRODUCT_PREFIX, "%s%s_%s", 
      pro_catg, product_pre->binx, product_pre->biny, date);
  }
  else {
    sprintf( product_id, "%s%s", XSH_PRODUCT_PREFIX, 
      pro_catg, product_pre->binx, product_pre->biny);
  }
  product_name = xsh_stringcat_any( product_id, ".fits", NULL);
  check( product_frame = xsh_pre_save( product_pre, product_name, pro_catg,0));

  /* Save PAF 
  paf_name = xsh_stringcat_any( product_id, ".paf", NULL);
  XSH_ASSURE_NOT_NULL( paf_name);
  */
  check(plist=cpl_propertylist_duplicate(product_pre->data_header));
  if(arcfile_found) {

  check(cpl_propertylist_append_string(plist,"ARCFILE",arcfile));
  check(cpl_propertylist_set_comment(plist,"ARCFILE","Archive File Name"));
  /*
  check_msg( xsh_paf_save( instr, recipe_id,plist, paf_name,pro_catg ),
    "Can't save PAF file" ) ;
  */
  } else {
  /*
  check_msg( xsh_paf_save( instr, recipe_id,product_pre->data_header,
    paf_name,pro_catg ),
    "Can't save PAF file" ) ;
  */
  }

  /* Add the new frame into the frameset */
  cpl_frame_set_type( product_frame, CPL_FRAME_TYPE_IMAGE);
  cpl_frame_set_group( product_frame, CPL_FRAME_GROUP_PRODUCT);
  cpl_frame_set_level( product_frame, CPL_FRAME_LEVEL_FINAL);
  cpl_frame_set_tag( product_frame, pro_catg);
  cpl_frameset_insert( frameset, product_frame);
  xsh_add_product_file(product_name);

  cleanup:
    if (cpl_error_get_code () != CPL_ERROR_NONE) {
      xsh_free_frame( &product_frame);
      product_frame = NULL;
    }
    XSH_FREE( date);
    xsh_pre_free( &product_pre);
    xsh_free_propertylist(&plist);
    XSH_FREE( product_name);
    /* XSH_FREE( paf_name); */
    return;
}

/*---------------------------------------------------------------------------*/
/**
  @brief 
    Add 3d product frame
  @param frame The product frame
  @param frameset The recipe frameset.
  @param parameters The recipe parameter list
  @param recipe_id The recipe name
  @param instr The instrument settings  

*/

/*---------------------------------------------------------------------------*/
void 
xsh_add_product_pre_3d( cpl_frame * frame, 
			cpl_frameset * frameset,
			const cpl_parameterlist *parameters,
			const char *recipe_id,
			xsh_instrument* instr) 
{
  const char * pro_catg = "";
  xsh_pre_3d * product_pre = NULL;
  cpl_frame * product_frame = NULL;
  char product_id[256];
  char * product_name = NULL;
  /* char * paf_name = NULL; */
  time_t now ;
  char * date = NULL;

  /* Check entries */
  XSH_ASSURE_NOT_NULL( frame);
  XSH_ASSURE_NOT_NULL( frameset);
  XSH_ASSURE_NOT_NULL( parameters);
  XSH_ASSURE_NOT_NULL( recipe_id);
  XSH_ASSURE_NOT_NULL( instr);
 
  check( pro_catg = cpl_frame_get_tag (frame));

  assure( pro_catg != NULL, CPL_ERROR_NULL_INPUT,
	  "Frame tag has not been set");

  /* Load PRE */
  check( product_pre = xsh_pre_3d_load( frame));

  /* update header */
  cpl_frame_set_group( frame, CPL_FRAME_GROUP_PRODUCT);
  check( cpl_dfs_setup_product_header( product_pre->data_header,
				       frame, frameset, parameters,
				       recipe_id, instr->pipeline_id,
				       instr->dictionary,NULL));

  /* Save product */
  if (xsh_time_stamp_get() ){
    time( &now );
    date = xsh_sdate_utc(&now);
    sprintf( product_id, XSH_PRODUCT_PREFIX, "%s%s_%s", 
      pro_catg, date);
  }
  else {
    sprintf( product_id, "%s%s", XSH_PRODUCT_PREFIX, pro_catg );
  }
  product_name = xsh_stringcat_any( product_id, ".fits", NULL);
  check( product_frame = xsh_pre_3d_save( product_pre, product_name, 0));

  /* Save PAF 
  paf_name = xsh_stringcat_any( product_id, ".paf", NULL);
  XSH_ASSURE_NOT_NULL( paf_name);
  check_msg( xsh_paf_save( instr, recipe_id,
			   product_pre->data_header, paf_name,
			   pro_catg ),
	     "Cant save PAF file" ) ;
  */
  /* Add the new frame into the frameset */
  cpl_frame_set_type( product_frame, CPL_FRAME_TYPE_IMAGE);
  cpl_frame_set_group( product_frame, CPL_FRAME_GROUP_PRODUCT);
  cpl_frame_set_level( product_frame, CPL_FRAME_LEVEL_FINAL);
  cpl_frame_set_tag( product_frame, pro_catg);
  cpl_frameset_insert( frameset, product_frame);
  xsh_add_product_file(product_name);

  cleanup:
    if (cpl_error_get_code () != CPL_ERROR_NONE) {
      xsh_free_frame( &product_frame);
      product_frame = NULL;
    }
    XSH_FREE( date);
    xsh_pre_3d_free( &product_pre);
    XSH_FREE( product_name);
    /* XSH_FREE( paf_name); */
    return;
}

/** 
 * Save BadPixelMap product
 * 
 * @param frame Product Frame
 * @param frameset Input Frameset
 * @param parameters Parameters list
 * @param recipe_id Recipe Identification String
 * @param instrument Instrument structure
 * @param final_prefix Final file name
 */
void 
xsh_add_product_bpmap( cpl_frame *frame, 
		       cpl_frameset * frameset,
		       const cpl_parameterlist * parameters,
		       const char *recipe_id,
		       xsh_instrument* instrument,
		       const char *final_prefix ) 
{
  cpl_propertylist *primary_header = NULL;
  const char *pro_catg = "";
  cpl_frame* product_frame = NULL;
  cpl_image *image = NULL ;
  const char *fname = NULL;
  char *final_name = NULL ;
  time_t now ;
  char* date = NULL ;
  /* char *paf_name = NULL ; */

  /* Check input parameters */
  XSH_ASSURE_NOT_NULL(frame);
  XSH_ASSURE_NOT_NULL(frameset);
  XSH_ASSURE_NOT_NULL(parameters);
  XSH_ASSURE_NOT_NULL(recipe_id);
  XSH_ASSURE_NOT_NULL(instrument);
  XSH_ASSURE_NOT_NULL(final_prefix);
 
  check(pro_catg = cpl_frame_get_tag (frame));
  XSH_ASSURE_NOT_NULL(pro_catg);
  
  /* Load header and image of bpmap frame */
  check(fname = cpl_frame_get_filename( frame ));
  check(primary_header = cpl_propertylist_load( fname, 0 ));
  check(image = cpl_image_load( fname, CPL_TYPE_INT, 0, 0 ));

  /* update header */
  check( cpl_frame_set_group( frame, CPL_FRAME_GROUP_PRODUCT));
  check_msg (cpl_dfs_setup_product_header ( primary_header,
					    frame,frameset,parameters,
					    recipe_id,
					    instrument->pipeline_id,
					    instrument->dictionary,NULL),
	     "Problem in the product DFS-compliance");

  /* Create final name */
  if ( xsh_time_stamp_get() == TRUE ) {
    time( &now ) ;
    date = xsh_sdate_utc( &now ) ;
    final_name = xsh_stringcat_any( XSH_PRODUCT_PREFIX,
      final_prefix, "_", date, ".fits", "" ) ;
  }
  else
    final_name = xsh_stringcat_any( XSH_PRODUCT_PREFIX,
      final_prefix, ".fits", "" ) ;

  xsh_msg_dbg_low( "Final product name: %s", final_name ) ;

  /* Now save file with updated header */
  cpl_image_save( image, final_name, CPL_BPP_32_SIGNED, primary_header,
		  CPL_IO_DEFAULT ) ;

  /*
    Save PAF
  
  paf_name = xsh_stringcat_any( final_prefix, ".paf", NULL ) ;
  XSH_ASSURE_NOT_NULL( paf_name ) ;
  check_msg( xsh_paf_save( instrument, recipe_id,
			   primary_header, paf_name, pro_catg ),
	     "Cant save PAF file" ) ;
  */
  /* Create product frame */
  check(product_frame=xsh_frame_product(final_name,pro_catg,
					CPL_FRAME_TYPE_IMAGE,
					CPL_FRAME_GROUP_PRODUCT,
					CPL_FRAME_LEVEL_FINAL));

  /*
  XSH_NEW_FRAME(product_frame);

  check(cpl_frame_set_filename (product_frame, final_name));
  check(cpl_frame_set_type (product_frame, CPL_FRAME_TYPE_IMAGE));

  // Add the new frame into the frameset 
  check(cpl_frame_set_group (product_frame, CPL_FRAME_GROUP_PRODUCT));
  check(cpl_frame_set_level (product_frame, CPL_FRAME_LEVEL_FINAL));
  check(cpl_frame_set_tag (product_frame, pro_catg));
  */

  check(cpl_frameset_insert(frameset, product_frame));
  xsh_add_product_file(final_name);

 cleanup:
  XSH_FREE(date);  
  xsh_free_propertylist(&primary_header);
  xsh_free_image( &image ) ;
  XSH_FREE( final_name ) ;
  /* XSH_FREE( paf_name ) ; */
  return;
}

/*****************************************************************************/
/** 
 * Save Image product
 * 
 * @param frame Product Frame
 * @param frameset Input Frameset
 * @param parameters Parameters list
 * @param recipe_id Recipe Identification String
 * @param instrument Instrument structure
 * @param final_prefix Final file name
 */
/*****************************************************************************/
void 
xsh_add_product_imagelist( cpl_frame *frame, 
			   cpl_frameset *frameset,
			   const cpl_parameterlist *parameters, 
			   const char *recipe_id,
			   xsh_instrument *instrument, 
			   const char *final_prefix)
{
  cpl_propertylist *primary_header = NULL;
  const char *pro_catg = NULL;
  cpl_frame* product_frame = NULL;
  cpl_imagelist *imagelist = NULL ;
  const char *fname = NULL;
  char *final_name = NULL ;
  time_t now ;
  char* date = NULL ;
  /* char *paf_name = NULL ; */

  /* Check input parameters */
  XSH_ASSURE_NOT_NULL(frame);
  XSH_ASSURE_NOT_NULL(frameset);
  XSH_ASSURE_NOT_NULL(parameters);
  XSH_ASSURE_NOT_NULL(recipe_id);
  XSH_ASSURE_NOT_NULL(instrument);
  XSH_ASSURE_NOT_NULL(final_prefix);
 
  check(pro_catg = cpl_frame_get_tag (frame));
  XSH_ASSURE_NOT_NULL(pro_catg);
  
  /* Load header and imagelist of bpmap frame */
  check(fname = cpl_frame_get_filename( frame ));
  check(primary_header = cpl_propertylist_load( fname, 0 ));
  check(imagelist = cpl_imagelist_load( fname, CPL_TYPE_FLOAT, 0 ));

  /* update header */
  check( cpl_frame_set_group( frame, CPL_FRAME_GROUP_PRODUCT));
  check_msg ( cpl_dfs_setup_product_header ( primary_header,
					    frame,frameset,parameters,
					    recipe_id,
					    instrument->pipeline_id,
					    instrument->dictionary,NULL),
	     "Problem in the product DFS-compliance");

  /* Create final name */
  if ( xsh_time_stamp_get()){
    time( &now);
    date = xsh_sdate_utc( &now);
    final_name = xsh_stringcat_any( XSH_PRODUCT_PREFIX,
      final_prefix, "_", date, ".fits", "" );
  }
  else
    final_name = xsh_stringcat_any( XSH_PRODUCT_PREFIX,
      final_prefix, ".fits", "");

  xsh_msg_dbg_low( "Final product name: %s", final_name ) ;

  /* Now save file with updated header */
  cpl_imagelist_save(imagelist, final_name, CPL_BPP_IEEE_FLOAT, primary_header,
		  CPL_IO_DEFAULT);

  /*
    Save PAF
  
  paf_name = xsh_stringcat_any( final_prefix, ".paf", NULL ) ;
  XSH_ASSURE_NOT_NULL( paf_name ) ;
  check_msg( xsh_paf_save( instrument, recipe_id,
			   primary_header, paf_name, pro_catg ),
	     "Cant save PAF file" ) ;
  */
  /* Create product frame */
  check(product_frame=xsh_frame_product(final_name,pro_catg,
					CPL_FRAME_TYPE_IMAGE,
					CPL_FRAME_GROUP_PRODUCT,
					CPL_FRAME_LEVEL_FINAL));

  /* old version
  XSH_NEW_FRAME(product_frame);

  check(cpl_frame_set_filename (product_frame, final_name));
  check(cpl_frame_set_type (product_frame, CPL_FRAME_TYPE_IMAGE));

  // Add the new frame into the frameset 
  check(cpl_frame_set_group (product_frame, CPL_FRAME_GROUP_PRODUCT));
  check(cpl_frame_set_level (product_frame, CPL_FRAME_LEVEL_FINAL));
  check(cpl_frame_set_tag (product_frame, pro_catg));
  */

  check(cpl_frameset_insert(frameset, product_frame));
  xsh_add_product_file(final_name);

 cleanup:
  XSH_FREE(date);  
  xsh_free_propertylist(&primary_header);
  xsh_free_imagelist( &imagelist ) ;
  XSH_FREE( final_name ) ;
  /* XSH_FREE( paf_name ) ; */
  return;
}
/*****************************************************************************/
/*****************************************************************************/
/** 
 * Save Image product
 * 
 * @param frame Product Frame
 * @param frameset Input Frameset
 * @param parameters Parameters list
 * @param recipe_id Recipe Identification String
 * @param instrument Instrument structure
 * @param final_prefix Final file name
 */
/*****************************************************************************/
void 
xsh_add_product_image( cpl_frame *frame, 
		       cpl_frameset *frameset,
		       const cpl_parameterlist *parameters,
		       const char *recipe_id,
		       xsh_instrument *instrument,
		       const char *final_prefix)
{
  cpl_propertylist *primary_header = NULL;
  const char *pro_catg = NULL;
  cpl_frame* product_frame = NULL;
  cpl_image *image = NULL ;
  const char *fname = NULL;
  char *final_name = NULL ;
  time_t now ;
  char* date = NULL ;
  /* char *paf_name = NULL ; */
  int i=0;
  int nbext=0;
  char arcfile[80];
  bool arcfile_found=false;

  /* Check input parameters */
  XSH_ASSURE_NOT_NULL(frame);
  XSH_ASSURE_NOT_NULL(frameset);
  XSH_ASSURE_NOT_NULL(parameters);
  XSH_ASSURE_NOT_NULL(recipe_id);
  XSH_ASSURE_NOT_NULL(instrument);
  XSH_ASSURE_NOT_NULL(final_prefix);
 
  check( pro_catg = cpl_frame_get_tag (frame));
 
  XSH_ASSURE_NOT_NULL(pro_catg);
  
  /* Load header and image of bpmap frame */
  check( fname = cpl_frame_get_filename( frame));

  check( primary_header = cpl_propertylist_load( fname, 0));

  check( image = cpl_image_load( fname, CPL_TYPE_FLOAT, 0, 0));
  if(cpl_propertylist_has(primary_header,"ARCFILE")) {
    sprintf(arcfile,"%s",xsh_pfits_get_arcfile(primary_header));
    arcfile_found=true;
  }
  /* update header */
  check( cpl_frame_set_group( frame, CPL_FRAME_GROUP_PRODUCT));
  
  check_msg (cpl_dfs_setup_product_header ( primary_header,
					    frame,frameset,parameters,
					    recipe_id,
					    instrument->pipeline_id,
					    instrument->dictionary,NULL),
	     "Problem in the product DFS-compliance");

  /* Create final name */
  if ( xsh_time_stamp_get() ) {
    time( &now ) ;
    date = xsh_sdate_utc( &now ) ;
    final_name = xsh_stringcat_any( XSH_PRODUCT_PREFIX,
      final_prefix, "_", date, ".fits", "" ) ;
  }
  else if ( strstr( final_prefix, ".fits" ) == NULL )
    final_name = xsh_stringcat_any( XSH_PRODUCT_PREFIX,
      final_prefix, ".fits", "");
  else final_name = xsh_stringcat_any( XSH_PRODUCT_PREFIX,
				       final_prefix, "" ) ;

  xsh_msg( "Final product name: %s", final_name ) ;

  /* Now save file with updated header */
  check(cpl_image_save( image, final_name, CPL_BPP_IEEE_FLOAT, primary_header,
			CPL_IO_DEFAULT )) ;
 

  nbext = cpl_frame_get_nextensions( frame);


  for( i = 1 ; i<=nbext ; i++ ) {
    int extension ;
    cpl_image* ext_img=NULL;
    cpl_propertylist* ext_header=NULL;

    check( ext_img = cpl_image_load( fname, CPL_TYPE_FLOAT,0, i));
    check( ext_header = cpl_propertylist_load( fname ,i));

    if ( i == 0 ) extension = CPL_IO_DEFAULT ;
    else extension = CPL_IO_EXTEND ;
    check(cpl_image_save( ext_img, final_name, CPL_BPP_IEEE_FLOAT,ext_header,
			  extension));
    xsh_free_image( &ext_img) ;
    xsh_free_propertylist( &ext_header) ;
  }

  
  /* Save PAF
  paf_name = xsh_stringcat_any( final_prefix, ".paf", NULL);
  XSH_ASSURE_NOT_NULL( paf_name);
  */
  if(arcfile_found) {
    check(cpl_propertylist_append_string(primary_header,"ARCFILE",arcfile));
    check(cpl_propertylist_set_comment(primary_header,"ARCFILE",
				     "Archive File Name"));
  }
  /*
  check_msg( xsh_paf_save( instrument, recipe_id,
			   primary_header, paf_name, pro_catg ),
	     "Cant save PAF file" ) ;
  */
  /* Create product frame */
  check( product_frame = xsh_frame_product( final_name,pro_catg,
					CPL_FRAME_TYPE_IMAGE,
					CPL_FRAME_GROUP_PRODUCT,
					CPL_FRAME_LEVEL_FINAL));
  
  check(cpl_frameset_insert(frameset, product_frame));
  xsh_add_product_file(final_name);
 
  cleanup:
    XSH_FREE(date);  
    xsh_free_propertylist( &primary_header);
    xsh_free_image( &image);
    XSH_FREE( final_name);
    /* XSH_FREE( paf_name); */
    return;
}
/*****************************************************************************/
/*****************************************************************************/
/** 
  @brief
    Save Table product (input frame has several extensions, 1 table per extension)
  @param frame Product Frame
  @param frameset Input Frameset
  @param parameters Parameters list
  @param recipe_id Recipe Identification String
  @param instrument Instrument structure
 */
/*****************************************************************************/
void 
xsh_add_product_table_multi( cpl_frame *frame, 
			     cpl_frameset *frameset,
			     const cpl_parameterlist * parameters,
			     const char *recipe_id,
			     xsh_instrument* instrument)
{
  cpl_propertylist *primary_header = NULL;
  cpl_propertylist *tbl_header = NULL;
  const char *pro_catg = "";
  cpl_frame* product_frame = NULL;
  cpl_table * table = NULL ;
  const char *fname = NULL;
  char *product_id = NULL;
  char *product_name = NULL ;
  char *tmp_name = NULL ;

  time_t now ;
  char* date = NULL ;
  /* char *paf_name = NULL ; */
  int nbext, i ;

  /* Check input parameters */
  XSH_ASSURE_NOT_NULL(frame);
  XSH_ASSURE_NOT_NULL(frameset);
  XSH_ASSURE_NOT_NULL(parameters);
  XSH_ASSURE_NOT_NULL(recipe_id);
  XSH_ASSURE_NOT_NULL(instrument);

  /* set the procatg */
  check(pro_catg = cpl_frame_get_tag (frame));

  XSH_ASSURE_NOT_NULL(pro_catg); 

  /* Load header and image of table frame */
  check( fname = cpl_frame_get_filename( frame));
  check( primary_header = cpl_propertylist_load( fname, 0));
  XSH_ASSURE_NOT_NULL( primary_header);
  check( tbl_header = cpl_propertylist_load( fname, 1));
  XSH_ASSURE_NOT_NULL( tbl_header);
  nbext = cpl_frame_get_nextensions( frame);

  /* update header */
  check( cpl_frame_set_group (frame, CPL_FRAME_GROUP_PRODUCT));
  check( cpl_dfs_setup_product_header ( primary_header,
    frame, frameset, parameters, recipe_id, instrument->pipeline_id,
					instrument->dictionary,NULL));

  /* Create final name */
  if ( xsh_time_stamp_get() ) {
    time( &now ) ;
    date = xsh_sdate_utc( &now ) ;
    product_id = xsh_stringcat_any( XSH_PRODUCT_PREFIX, 
      pro_catg, "_", date, "" ) ;
  }
  else
    product_id = xsh_stringcat_any( XSH_PRODUCT_PREFIX,
       pro_catg, "" ) ;

  product_name = xsh_stringcat_any( product_id, ".fits", NULL );

  /* Now save file with updated header */
  /* Save tables */
  tmp_name=xsh_stringcat_any( "tmp_",product_name, NULL );
  for( i = 0 ; i<nbext ; i++ ) {
    int extension ;
    cpl_table *tbl_ext = NULL;
    cpl_propertylist *tbl_ext_header = NULL;
    /* Load table
       save table
    */
    check( tbl_ext = cpl_table_load( fname, i+1, 0));
    check( tbl_ext_header = cpl_propertylist_load( fname, i+1));

    if ( i == 0 ) extension = CPL_IO_DEFAULT ;
    else extension = CPL_IO_EXTEND ;
    check(cpl_table_save( tbl_ext, primary_header, tbl_ext_header,
			  tmp_name, extension));
    xsh_free_table( &tbl_ext);
    xsh_free_propertylist( &tbl_ext_header);
  }

  xsh_fileutils_move( tmp_name,product_name );
 
  /* Save PAF  

  paf_name = xsh_stringcat_any( product_id, ".paf", NULL ) ;
  XSH_ASSURE_NOT_NULL( paf_name ) ;
 

  check_msg( xsh_paf_save( instrument, recipe_id,
			   primary_header, paf_name, pro_catg ),
	     "Cant save PAF file" ) ;
  */
  /* Create product frame */
  check(product_frame=xsh_frame_product(product_name,pro_catg,
					CPL_FRAME_TYPE_IMAGE,
					CPL_FRAME_GROUP_PRODUCT,
					CPL_FRAME_LEVEL_FINAL));

  cpl_frameset_insert(frameset, product_frame);
  xsh_add_product_file(product_name);

  cleanup:
    xsh_free_propertylist(&primary_header);
    xsh_free_propertylist(&tbl_header);
    XSH_TABLE_FREE( table) ;
    XSH_FREE(product_id);
    XSH_FREE(tmp_name);
    XSH_FREE(product_name);
    XSH_FREE(date);
    /* XSH_FREE(paf_name ) ; */
   return;
}
/*****************************************************************************/

/** 
  @brief Save Table product
  @param frame Product Frame
  @param frameset Input Frameset
  @param parameters Parameters list
  @param recipe_id Recipe Identification String
  @param instrument Instrument structure
 */
void 
xsh_add_product_table( cpl_frame *frame, 
		       cpl_frameset * frameset,
		       const cpl_parameterlist * parameters,
		       const char *recipe_id,
		       xsh_instrument* instrument) 
{
  cpl_propertylist *primary_header = NULL;
  cpl_propertylist *tbl_header = NULL;
  const char *pro_catg = NULL;
  cpl_frame* product_frame = NULL;
  cpl_table * table = NULL ;
  const char *fname = NULL;
  char *product_id = NULL;
  char *product_name = NULL;
  time_t now ;
  char* date = NULL ;
  /* char *paf_name = NULL ; */

  /* Check input parameters */
  XSH_ASSURE_NOT_NULL(frame);
  XSH_ASSURE_NOT_NULL(frameset);
  XSH_ASSURE_NOT_NULL(parameters);
  XSH_ASSURE_NOT_NULL(recipe_id);
  XSH_ASSURE_NOT_NULL(instrument);

  /* set the procatg */
  check( pro_catg = cpl_frame_get_tag (frame));

  XSH_ASSURE_NOT_NULL(pro_catg); 

  /* Load header and image of table frame */
  check( fname = cpl_frame_get_filename( frame ));
  check( primary_header = cpl_propertylist_load( fname, 0 ));
  XSH_ASSURE_NOT_NULL( primary_header);
  check( tbl_header = cpl_propertylist_load( fname, 1 ));
  XSH_ASSURE_NOT_NULL( tbl_header);
  XSH_TABLE_LOAD( table, fname);
  XSH_ASSURE_NOT_NULL( table);

  /* update header */
  check( cpl_frame_set_group (frame, CPL_FRAME_GROUP_PRODUCT));
  check( cpl_dfs_setup_product_header( primary_header,
				       frame, frameset, parameters,
				       recipe_id, instrument->pipeline_id,
				       instrument->dictionary,NULL));

  /* Create final name */
  if ( xsh_time_stamp_get() ) {
    time( &now ) ;
    date = xsh_sdate_utc( &now ) ;
    product_id = xsh_stringcat_any( XSH_PRODUCT_PREFIX, 
      pro_catg, date, "" ) ;
  }
  else
    product_id = xsh_stringcat_any( XSH_PRODUCT_PREFIX,
      pro_catg, "");

  /* Now save file with updated header */
  product_name = xsh_stringcat_any( product_id,".fits", NULL);
  /* Save table */
  check( cpl_table_save( table, primary_header, tbl_header,
    product_name, CPL_IO_DEFAULT ));

  /* Save PAF   
  paf_name = xsh_stringcat_any( product_id, ".paf", NULL ) ;
  XSH_ASSURE_NOT_NULL( paf_name ) ;

  check_msg( xsh_paf_save( instrument, recipe_id,
			   primary_header, paf_name, pro_catg ),
	     "Cant save PAF file" ) ;
  */
  /* Create product frame */
  check(product_frame=xsh_frame_product(product_name,pro_catg,
					CPL_FRAME_TYPE_IMAGE,
					CPL_FRAME_GROUP_PRODUCT,
					CPL_FRAME_LEVEL_FINAL));


  cpl_frameset_insert(frameset, product_frame);
  xsh_add_product_file(product_name);

 cleanup:
    xsh_free_propertylist(&primary_header);
    xsh_free_propertylist(&tbl_header);
    XSH_TABLE_FREE( table) ;
    XSH_FREE( product_id);
    XSH_FREE(product_name);
    XSH_FREE(date);
    /* XSH_FREE(paf_name ) ; */
   return;
}

/*---------------------------------------------------------------------------*/
/**
   @brief Find standard star frame
   @param frames The input frameset
   @return Pointer to first standard star frame, or NULL on error.

   The function sets an error if no standard star frame is present.

*/
/*---------------------------------------------------------------------------*/

cpl_frame *
xsh_find_std_flux (cpl_frameset * frames)
{
  const char *tags[] = { 
                         XSH_STD_FLUX_SLIT (XSH_ARM_UVB),
			 XSH_STD_FLUX_SLIT (XSH_ARM_VIS),
			 XSH_STD_FLUX_SLIT (XSH_ARM_NIR),
                         XSH_STD_TELL_SLIT (XSH_ARM_UVB),
			 XSH_STD_TELL_SLIT (XSH_ARM_VIS),
			 XSH_STD_TELL_SLIT (XSH_ARM_NIR),
			 XSH_STD_FLUX_OFFSET (XSH_ARM_UVB),
			 XSH_STD_FLUX_OFFSET (XSH_ARM_VIS),
			 XSH_STD_FLUX_OFFSET (XSH_ARM_NIR),
			 XSH_STD_TELL_OFFSET (XSH_ARM_UVB),
			 XSH_STD_TELL_OFFSET (XSH_ARM_VIS),
			 XSH_STD_TELL_OFFSET (XSH_ARM_NIR),
			 NULL
  };

  return xsh_find_frame (frames, tags);
}
/*---------------------------------------------------------------------------*/
/**
   @brief Find standard standard star flux frame
   @param frames The input frameset
   @return Pointer to first standard star frame, or NULL on error.

   The function sets an error if no standard star frame is present.

*/
/*---------------------------------------------------------------------------*/

cpl_frame *
xsh_find_std_star_flux (cpl_frameset * frames)
{
  const char *tags[] = { XSH_STD_STAR_FLUX (XSH_ARM_UVB),
			 XSH_STD_STAR_FLUX (XSH_ARM_VIS),
			 XSH_STD_STAR_FLUX (XSH_ARM_NIR),
			 NULL
  };

  return xsh_find_frame (frames, tags);
}

/*---------------------------------------------------------------------------*/
/**
   @brief Find master bias frame
   @param frames The input frameset
   @param instr the settings of instrument

   @return Pointer to first standard star frame, or NULL on error.

   The function sets an error if no standard star frame is present.

*/
/*---------------------------------------------------------------------------*/

cpl_frame* 
xsh_find_master_bias (cpl_frameset * frames, xsh_instrument* instr)
{
  const char *tags[2] ={NULL,NULL};
  cpl_frame* result = NULL;

  tags[0] = XSH_GET_TAG_FROM_ARM( XSH_MASTER_BIAS, instr);

  check(result = xsh_find_frame (frames,(const char**) tags));

 cleanup:
  return result;
}

/*---------------------------------------------------------------------------*/
/**
   @brief Find frame with a given tag
   @param frames The input frameset
   @param tag    The frame tag
   @param instr the settings of instrument

   @return Pointer to first standard star frame, or NULL on error.
    
   The function sets an error if no standard star frame is present.
  
*/
/*---------------------------------------------------------------------------*/
  
cpl_frame* 
xsh_find_frame_with_tag (cpl_frameset * frames, 
			 const char* tag,  
			 xsh_instrument* instr)
{
  char *tags[2];
  cpl_frame* result = NULL;

  tags[0] = xsh_get_tag_from_arm( tag,  instr);
  tags[1] = NULL;
  xsh_msg_dbg_high("search for tag %s",tags[0]);
  result = xsh_find_frame (frames,(const char**) tags);
  cpl_free( tags[0]);
  return result;
}


/*---------------------------------------------------------------------------*/
/**
   @brief Find arm specific tag from base and instrument setting 
   @param tag    The frame tag
   @param instr the settings of instrument

   @return string composed tag
   
*/
/*---------------------------------------------------------------------------*/

char* xsh_get_tag_from_arm(const char* tag, xsh_instrument* instr){
   const char* arm=xsh_instrument_arm_tostring(instr);
   char* composed_tag=NULL;

   int len = strlen(tag);

   if(tag[len-1]=='_') {
      composed_tag=cpl_sprintf("%s%s",tag,arm);
   } else {
      composed_tag=cpl_sprintf("%s%s%s",tag,"_",arm);
   }
   xsh_msg_dbg_high("composed tag='%s'",composed_tag);

   return composed_tag;
}


/*---------------------------------------------------------------------------*/
/**
   @brief Find master dark frame
   @param frames The input frameset
   @param instr the settings of instrument

   @return Pointer to first standard star frame, or NULL on error.
    
   The function sets an error if no standard star frame is present.
  
*/
/*---------------------------------------------------------------------------*/
  
cpl_frame* xsh_find_master_dark (cpl_frameset * frames, xsh_instrument* instr)
{
  const char *tags[2];
  
  tags[0] = XSH_GET_TAG_FROM_ARM( XSH_MASTER_DARK,  instr);
  tags[1] = NULL;

  return xsh_find_frame (frames, tags);
}

/*---------------------------------------------------------------------------*/
/**
   @brief Find master flat frame
   @param frames The input frameset
   @param instr the settings of instrument

   @return Pointer to first standard star frame, or NULL on error.

   The function sets an error if no standard star frame is present.

*/
/*---------------------------------------------------------------------------*/
cpl_frame* xsh_find_master_flat (cpl_frameset * frames, xsh_instrument* instr)
{
  const char *tags[2];

  tags[0] = XSH_GET_TAG_FROM_MODE ( XSH_MASTER_FLAT, instr);
  tags[1] = NULL;

  return xsh_find_frame (frames, tags);
}

/*---------------------------------------------------------------------------*/
/**
  @brief Find a theoretical tab signle pinhole in SOF 
  @param frames The input frameset where we do the search
  @param instr The instrument
  
  @return Pointer to the search frame or exception if don't find
*/
/*---------------------------------------------------------------------------*/
cpl_frame *xsh_find_theo_tab_sing( cpl_frameset *frames, xsh_instrument* instr)
{
  const char* tags[2] = {NULL, NULL};
  cpl_frame* result = NULL;

  XSH_ASSURE_NOT_NULL( frames);
  XSH_ASSURE_NOT_NULL( instr);
  tags[0] = XSH_GET_TAG_FROM_ARM( XSH_THEO_TAB_SING, instr);
  check(result = xsh_find_frame (frames, tags));
  
  cleanup:
    return result;
}

/*---------------------------------------------------------------------------*/
/**
  @brief Find a theoretical tab multi pinhole in SOF 
  @param frames The input frameset where we do the search
  @param instr The instrument
  
  @return Pointer to the search frame or exception if don't find
*/
/*---------------------------------------------------------------------------*/
cpl_frame *xsh_find_theo_tab_mult( cpl_frameset *frames, xsh_instrument* instr)
{
  const char *tags[2] = {NULL, NULL};
  cpl_frame* result = NULL;

  XSH_ASSURE_NOT_NULL( frames);
  XSH_ASSURE_NOT_NULL( instr);
  tags[0] = XSH_GET_TAG_FROM_ARM( XSH_THEO_TAB_MULT, instr);
  check(result = xsh_find_frame (frames, tags));

  cleanup:
    return result;
}
/*---------------------------------------------------------------------------*/
/**
  @brief Find a theoretical tab IFU or SINg depending of MODE in SOF 
  @param frames The input frameset where we do the search
  @param instr The instrument
  
  @return Pointer to the search frame or exception if don't find
*/
/*---------------------------------------------------------------------------*/
cpl_frame *xsh_find_theo_tab_mode( cpl_frameset *frames, xsh_instrument* instr)
{
  const char *tags[2] = {NULL, NULL};
  cpl_frame* result = NULL;

  XSH_ASSURE_NOT_NULL( frames);
  XSH_ASSURE_NOT_NULL( instr);

  if ( instr->mode == XSH_MODE_IFU){
    tags[0] = XSH_GET_TAG_FROM_ARM( XSH_THEO_TAB_IFU, instr);
  }
  else{
    tags[0] = XSH_GET_TAG_FROM_ARM( XSH_THEO_TAB_SING, instr);
  }
  check(result = xsh_find_frame (frames, tags));

  cleanup:
    return result;
}

/*---------------------------------------------------------------------------*/
/**
  @brief Find an order tab RECOV
  @param frames The input frameset where we do the search
  @param instr The instrument
  
  @return Pointer to the search frame or exception if don't find
*/
/*---------------------------------------------------------------------------*/
cpl_frame *
xsh_find_order_tab_recov( cpl_frameset *frames, xsh_instrument* instr)
{
  const char *tags[2] = {NULL, NULL};
  cpl_frame* result = NULL;

  XSH_ASSURE_NOT_NULL( frames);
  XSH_ASSURE_NOT_NULL( instr);

  tags[0] = XSH_GET_TAG_FROM_ARM( XSH_ORDER_TAB_RECOV, instr);
  check( result = xsh_find_frame (frames, tags));

  cleanup:
    return result;
}

/*---------------------------------------------------------------------------*/
/**
  @brief Find an order tab GUESS 
  @param frames The input frameset where we do the search
  @param instr The instrument
  
  @return Pointer to the search frame or exception if don't find
*/
/*---------------------------------------------------------------------------*/
cpl_frame *
xsh_find_order_tab_guess( cpl_frameset *frames, xsh_instrument* instr)
{
  const char *tags[2] = {NULL, NULL};
  cpl_frame* result = NULL;

  XSH_ASSURE_NOT_NULL( frames);
  XSH_ASSURE_NOT_NULL( instr);

  tags[0] = XSH_GET_TAG_FROM_ARM( XSH_ORDER_TAB_GUESS, instr);
  check(result = xsh_find_frame (frames, tags));

  cleanup:
    return result;
}

/*---------------------------------------------------------------------------*/
/**
  @brief Find an order tab CENTR
  @param frames The input frameset where we do the search
  @param instr The instrument
  
  @return Pointer to the search frame or exception if don't find
*/
/*---------------------------------------------------------------------------*/
cpl_frame *
xsh_find_order_tab_centr( cpl_frameset *frames, xsh_instrument* instr)
{
  const char *tags[2] = {NULL, NULL};
  cpl_frame* result = NULL;

  XSH_ASSURE_NOT_NULL( frames);
  XSH_ASSURE_NOT_NULL( instr);

  tags[0] = XSH_GET_TAG_FROM_ARM( XSH_ORDER_TAB_CENTR, instr);
  check(result = xsh_find_frame (frames, tags));

  cleanup:
    return result;
}

/*---------------------------------------------------------------------------*/
/**
  @brief Find an order tab EDGES
  @param frames The input frameset where we do the search
  @param instr The instrument
  
  @return Pointer to the search frame or exception if don't find
*/
/*---------------------------------------------------------------------------*/
cpl_frame *
xsh_find_order_tab_edges( cpl_frameset *frames, xsh_instrument* instr)
{
  const char *tags[3] = {NULL, NULL};
  cpl_frame* result = NULL;

  XSH_ASSURE_NOT_NULL( frames);
  XSH_ASSURE_NOT_NULL( instr);

  tags[0] = XSH_GET_TAG_FROM_MODE( XSH_ORDER_TAB_AFC, instr);
  tags[1] = XSH_GET_TAG_FROM_LAMP( XSH_ORDER_TAB_EDGES, instr);

  check(result = xsh_find_frame (frames, tags));

  xsh_msg_dbg_medium( "ORDER_TAB => %s", cpl_frame_get_filename( result));

  cleanup:
    return result;
}

/*---------------------------------------------------------------------------*/
/**
  @brief Find a wave tab guess
  @param frames The input frameset where we do the search
  @param instr The instrument
  
  @return  Pointer to the search frame or exception if don't find
*/
/*---------------------------------------------------------------------------*/
cpl_frame *
xsh_find_wave_tab_guess( cpl_frameset *frames, xsh_instrument* instr)
{
  const char *tags[2] = {NULL, NULL};
  cpl_frame* result = NULL;

  XSH_ASSURE_NOT_NULL( frames);
  XSH_ASSURE_NOT_NULL( instr);

  tags[0] = XSH_GET_TAG_FROM_ARM( XSH_WAVE_TAB_GUESS, instr);
  check(result = xsh_find_frame (frames, tags));

  cleanup:
    return result;
}

/*---------------------------------------------------------------------------*/
/**
  @brief Find a wave tab 2D
  @param frames The input frameset where we do the search
  @param instr The instrument
  
  @return  Pointer to the search frame or exception if don't find
*/
/*---------------------------------------------------------------------------*/
cpl_frame *
xsh_find_wave_tab_2d( cpl_frameset *frames,  xsh_instrument* instr)
{
  const char *tags[2] = {NULL, NULL};
  cpl_frame* result = NULL;

  XSH_ASSURE_NOT_NULL( frames);
  XSH_ASSURE_NOT_NULL( instr);

  tags[0] = XSH_GET_TAG_FROM_ARM( XSH_WAVE_TAB_2D, instr);
  check(result = xsh_find_frame (frames, tags));

  cleanup:
    return result;
}

/*---------------------------------------------------------------------------*/
/**
  @brief Find a slit map 
  @param frames The input frameset where we do the search
  @param instr The instrument
  
  @return  Pointer to the search frame or exception if don't find
*/
cpl_frame *xsh_find_slitmap( cpl_frameset *frames,
			     xsh_instrument* instr)
{
  const char *tags[3] = {NULL, NULL, NULL};
  cpl_frame* result = NULL;

  XSH_ASSURE_NOT_NULL( frames);
  XSH_ASSURE_NOT_NULL( instr);

  tags[0] = XSH_GET_TAG_FROM_ARM( XSH_SLIT_MAP, instr);
  tags[1] = XSH_GET_TAG_FROM_ARM( XSH_IFU_MAP, instr);
  check(result = xsh_find_frame (frames, tags));

  cleanup:
    return result;
}

/*---------------------------------------------------------------------------*/
/**
  @brief Find a wave tab 2D or a wave tab ARC
  @param frames The input frameset where we do the search
  @param instr The instrument
  
  @return  Pointer to the search frame or exception if don't find
*/
/*---------------------------------------------------------------------------*/
cpl_frame *xsh_find_wave_tab( cpl_frameset *frames,
  xsh_instrument* instr)
{
  const char *tags[4] = {NULL, NULL};
  cpl_frame* result = NULL;

  XSH_ASSURE_NOT_NULL( frames);
  XSH_ASSURE_NOT_NULL( instr);

  tags[0] = XSH_GET_TAG_FROM_ARM( XSH_WAVE_TAB_AFC, instr);
  tags[1] = XSH_GET_TAG_FROM_MODE( XSH_WAVE_TAB_ARC, instr);
  tags[2] = XSH_GET_TAG_FROM_ARM( XSH_WAVE_TAB_2D, instr);
  check(result = xsh_find_frame (frames, tags));
 
  cleanup:
    return result;
}


/*---------------------------------------------------------------------------*/
/**
  @brief Find a model config
  @param frames The input frameset where we do the search
  @param instr The instrument
  
  @return  Pointer to the search frame or exception if don't find
*/
/*---------------------------------------------------------------------------*/
cpl_frame *
xsh_find_model_config( cpl_frameset *frames,  xsh_instrument* instr)
{
  const char *tags[4] = {NULL, NULL};
  cpl_frame* result = NULL;

  XSH_ASSURE_NOT_NULL( frames);
  XSH_ASSURE_NOT_NULL( instr);

  tags[0] = XSH_GET_TAG_FROM_ARM( XSH_MOD_CFG_OPT_AFC, instr);
  tags[1] = XSH_GET_TAG_FROM_ARM( XSH_MOD_CFG_OPT_2D, instr);
  tags[2] = XSH_GET_TAG_FROM_ARM( XSH_MOD_CFG_TAB, instr);
  check(result = xsh_find_frame (frames, tags));

  cleanup:
    return result;
}

/*---------------------------------------------------------------------------*/
/**
  @brief Find wave tab ARC (for IFU 3 frames)
  @param frames The input frameset where we do the search
  @param instrument The instrument
  
  @return  Pointer to the frame set containing the 3 wave tab arc frames
    or exception if not founf
*/
/*---------------------------------------------------------------------------*/
cpl_frameset * 
xsh_find_wave_tab_ifu( cpl_frameset *frames, xsh_instrument* instrument)
{
  const char * tags[3] = { NULL, NULL, NULL} ;
  cpl_frameset * result = NULL;
  cpl_frame * resframe = NULL ;

  XSH_ASSURE_NOT_NULL( frames);
  XSH_ASSURE_NOT_NULL( instrument);

  /* Create frameset */
  check( result = cpl_frameset_new() ) ;

  /*
    If only one wave solution "XSH_WAVE_TAB_ARC", duplicat the frame 3
     times and set the correct tags for each of them
     else fail if the 3 wave solutions are not avaiable
  */
  tags[0] = XSH_GET_TAG_FROM_ARM( XSH_WAVE_TAB_AFC, instrument);
  tags[1] = XSH_GET_TAG_FROM_ARM( XSH_WAVE_TAB_2D, instrument);
  resframe = xsh_find_frame (frames, tags) ;
  if (resframe != NULL ) {
    cpl_frame * down = NULL, * cen = NULL , * up = NULL ;

    xsh_msg_warning( "Only one Wave Solution Frame %s, use it for the 3 slitlets",
      cpl_frame_get_filename( resframe)) ;
    check( down = cpl_frame_duplicate( resframe ) ) ;
    tags[0] = XSH_GET_TAG_FROM_ARM( XSH_WAVE_TAB_ARC_DOWN_IFU, instrument ) ;
    check( cpl_frame_set_tag( down, tags[0] ) ) ;
    check( cpl_frameset_insert( result, down ) ) ;

    check( cen = cpl_frame_duplicate( resframe ) ) ;
    tags[0] = XSH_GET_TAG_FROM_ARM( XSH_WAVE_TAB_ARC_CEN_IFU, instrument ) ;
    check( cpl_frame_set_tag( cen, tags[0] ) ) ;
    check( cpl_frameset_insert( result, cen ) ) ;

    check( up = cpl_frame_duplicate( resframe ) ) ;
    tags[0] = XSH_GET_TAG_FROM_ARM( XSH_WAVE_TAB_ARC_UP_IFU, instrument ) ;
    check( cpl_frame_set_tag( up, tags[0] ) ) ;
    check( cpl_frameset_insert( result, up ) ) ;
  }
  else {
    xsh_msg( "Three wave solution found" ) ;

    tags[0] = XSH_GET_TAG_FROM_ARM( XSH_WAVE_TAB_ARC_DOWN_IFU, instrument ) ;
    check(resframe = xsh_find_frame (frames, tags));
    check( cpl_frameset_insert( result, resframe ) ) ;

    tags[0] = XSH_GET_TAG_FROM_ARM( XSH_WAVE_TAB_ARC_CEN_IFU, instrument ) ;
    check(resframe = xsh_find_frame (frames, tags));
    check( cpl_frameset_insert( result, resframe ) ) ;

    tags[0] = XSH_GET_TAG_FROM_ARM( XSH_WAVE_TAB_ARC_UP_IFU, instrument ) ;
    check(resframe = xsh_find_frame (frames, tags));
    check( cpl_frameset_insert( result, resframe ) ) ;
  }

  cleanup:
  if ( cpl_error_get_code() != CPL_ERROR_NONE ) {
    cpl_frameset_delete( result ) ;
    return NULL ;
  }
  else return result;
}

/*---------------------------------------------------------------------------*/
/**
  @brief Find an arc lines list clean frame
  @param frames The input frameset where we do the search
  @param instr The instrument
  
  @return  Pointer to the search frame or exception if don't find
*/
/*---------------------------------------------------------------------------*/
cpl_frame *
xsh_find_arc_line_list_clean( cpl_frameset *frames,  xsh_instrument* instr)
{
  const char *tags[2] = {NULL, NULL};
  cpl_frame* result = NULL;

  XSH_ASSURE_NOT_NULL( frames);
  XSH_ASSURE_NOT_NULL( instr);

  tags[0] = XSH_GET_TAG_FROM_ARM( XSH_ARC_LINE_LIST, instr);
  check(result = xsh_find_frame (frames, tags));

  cleanup:
    return result;
}


/*---------------------------------------------------------------------------*/
/**
  @brief Find an arc line list frame
  @param frames The input frameset where we do the search
  @param instr The instrument
  
  @return  Pointer to the search frame or exception if don't find
*/
/*---------------------------------------------------------------------------*/
cpl_frame *
xsh_find_arc_line_list( cpl_frameset *frames,  xsh_instrument* instr)
{
  const char *tags[2] = {NULL, NULL};
  cpl_frame* result = NULL;

  XSH_ASSURE_NOT_NULL( frames);
  XSH_ASSURE_NOT_NULL( instr);

  tags[0] = XSH_GET_TAG_FROM_ARM( XSH_ARC_LINE_LIST, instr);
  check(result = xsh_find_frame (frames, tags));

  cleanup:
    return result;
}
/*---------------------------------------------------------------------------*/
/**
  @brief Find a user guess line positions table frame
  @param frames The input frameset where we do the search
  @param instr The instrument
  
  @return  Pointer to the search frame or exception if don't find
*/
/*---------------------------------------------------------------------------*/
cpl_frame *
xsh_find_usr_lines_guess_tab( cpl_frameset *frames,  xsh_instrument* instr)
{
  const char *tags[2] = {NULL, NULL};
  cpl_frame* result = NULL;

  XSH_ASSURE_NOT_NULL( frames);
  XSH_ASSURE_NOT_NULL( instr);

  tags[0] = XSH_GET_TAG_FROM_ARM( XSH_GUESS_LINES_POS, instr);
  check(result = xsh_find_frame (frames, tags));

  cleanup:
    return result;
}

/*---------------------------------------------------------------------------*/
/**
  @brief Find a model configuration table frame
  @param frames The input frameset where we do the search
  @param instr The instrument
  
  @return  Pointer to the search frame or exception if don't find
*/
/*---------------------------------------------------------------------------*/
cpl_frame *
xsh_find_model_config_tab( cpl_frameset *frames,  xsh_instrument* instr)
{
  const char *tags[2] = {NULL, NULL};
  cpl_frame* result = NULL;

  XSH_ASSURE_NOT_NULL( frames);
  XSH_ASSURE_NOT_NULL( instr);

  tags[0] = XSH_GET_TAG_FROM_ARM( XSH_MOD_CFG_TAB, instr);
  check(result = xsh_find_frame (frames, tags));

  cleanup:
    return result;
}

/**
   @brief Find Wave Map frame. The frame returned should not be free by
   the caller.

   @param frames The input frameset
   @param instr the instrument

   @return Pointer to wavemap frame

*/
/*---------------------------------------------------------------------------*/
cpl_frame * xsh_find_wavemap(cpl_frameset *frames,
			    xsh_instrument* instr)
{
  const char *tags[2] ={NULL,NULL};
  cpl_frame* result = NULL;

  tags[0] = XSH_GET_TAG_FROM_ARM( XSH_WAVE_MAP, instr);
  check(result = xsh_find_frame (frames,(const char**) tags));
  
  cleanup:
    return result;
}

/**
   @brief Find orderdef raw frame for UVB or VIS
   @param frames The input frameset
    
   @return Pointer to order table frame
    
*/
/*---------------------------------------------------------------------------*/

cpl_frame * xsh_find_raw_orderdef_vis_uvb ( cpl_frameset * frames )
{
  char *tags[4] ={NULL,NULL, NULL,NULL};
  cpl_frame* result = NULL;

  check(tags[0] = xsh_stringcat_any( XSH_ORDERDEF_D2_UVB, NULL )) ;
  check(tags[1] = xsh_stringcat_any( XSH_ORDERDEF_QTH_UVB, NULL )) ;
  check(tags[2] = xsh_stringcat_any( XSH_ORDERDEF_VIS, NULL )) ;

  check(result = xsh_find_frame (frames,(const char**) tags));

 cleanup:
  cpl_free( tags[0]);
  cpl_free( tags[1]);
  cpl_free( tags[2]);
  return result;
}

/**
   @brief Find orderdef raw frame for NIR
   @param frames The input frameset
    
   @return Pointer to order table frame
    
*/
/*---------------------------------------------------------------------------*/

cpl_frame * xsh_find_raw_orderdef_nir( cpl_frameset * frames )
{
  char *tags[] ={NULL,NULL};
  cpl_frame* result = NULL;

  check(tags[0] = xsh_stringcat_any( XSH_ORDERDEF_NIR, NULL ) );
  check(result = xsh_find_frame (frames,(const char**) tags));

 cleanup:
  cpl_free(tags[0]);
  return result;
}

/*---------------------------------------------------------------------------*/
/**
 *  @brief
 *    Find Dispersol tab frame. The frame returned should not be free
 *    by the caller.
 *  @param frames The input frameset
 *  @param instr  The instrument
 *  @return Pointer to wavemap frame
 */
/*---------------------------------------------------------------------------*/
cpl_frame* xsh_find_disp_tab( cpl_frameset *frames,
			      xsh_instrument* instr)
{
  const char *tags[3] ={NULL,NULL};
  cpl_frame* result = NULL;

  tags[0] = XSH_GET_TAG_FROM_ARM( XSH_DISP_TAB_AFC, instr);
  tags[1] = XSH_GET_TAG_FROM_ARM( XSH_DISP_TAB, instr);

  check( result = xsh_find_frame( frames,(const char**) tags));

  if (result == NULL){
    xsh_msg( "No DISP TAB frame found !!");
  }
  else{
    xsh_msg("DISP_TAB => %s", cpl_frame_get_filename( result));
  }

  cleanup:
    return result;
}
/*---------------------------------------------------------------------------*/


/**
   @brief Find orderdef raw frame for NIR OFF
   @param frames The input frameset
    
   @return Pointer to order table frame
    
*/
/*---------------------------------------------------------------------------*/

cpl_frame * xsh_find_raw_orderdef_nir_off( cpl_frameset * frames )
{
  char *tags[] ={NULL,NULL};
  cpl_frame* result = NULL;

  check(tags[0] = xsh_stringcat_any( XSH_ORDERDEF_OFF, NULL ) ) ;
  check(result = xsh_find_frame (frames,(const char**) tags));

 cleanup:
  cpl_free(tags[0]);
  return result;
}

/**
   @brief Find raw frame for Slit UVB/VIS
   @param frames The input frameset
   @param arm The ARM (UVB or VIS)
   @return Pointer to raw frame
    
*/
cpl_frame * xsh_find_raw_arc_slit_uvb_vis( cpl_frameset * frames,
					   XSH_ARM arm )
{
  char *tags[] ={NULL,NULL};
  cpl_frame* result = NULL;

  if ( arm == XSH_ARM_UVB )
    check( tags[0] = xsh_stringcat_any( XSH_ARC_SLIT_UVB, NULL ) ) ;
  else if ( arm == XSH_ARM_VIS )
    check( tags[0] = xsh_stringcat_any( XSH_ARC_SLIT_VIS, NULL ) ) ;
  else goto cleanup ;

  check(result = xsh_find_frame (frames,(const char**) tags));

 cleanup:
  cpl_free(tags[0]);
  return result;
}

/**
   @brief Find raw frame for Slit NIR ON
   @param frames The input frameset
    
   @return Pointer to raw frame
    
*/
cpl_frame * xsh_find_raw_arc_slit_nir_on( cpl_frameset * frames )
{
  char *tags[] ={NULL,NULL};
  cpl_frame* result = NULL;

  check(tags[0] = xsh_stringcat_any( XSH_ARC_SLIT_NIR, NULL ) ) ;
  check(result = xsh_find_frame (frames,(const char**) tags));

 cleanup:
  cpl_free(tags[0]);
  return result;
}

/**
   @brief Find raw frame for Slit NIR OFF
   @param frames The input frameset
    
   @return Pointer to raw frame
    
*/
cpl_frame * xsh_find_raw_arc_slit_nir_off( cpl_frameset * frames )
{
  char *tags[] ={NULL,NULL};
  cpl_frame* result = NULL;

  check(tags[0] = xsh_stringcat_any( XSH_ARC_SLIT_OFF, NULL ) ) ;
  check(result = xsh_find_frame (frames,(const char**) tags));

 cleanup:
  cpl_free(tags[0]);
  return result;
}


/**
   @brief Find model config frame 
   @param frames The input frameset
   @param instr the instrument arm setting 
   @return Pointer to raw frame
    
*/

cpl_frame * xsh_find_calpro_model_config( cpl_frameset * frames,
				      xsh_instrument * instr )
{
  char *tags[] ={NULL,NULL};
  cpl_frame* result = NULL;

  check(tags[0] = xsh_stringcat_any( XSH_MOD_CFG,
				     xsh_instrument_arm_tostring(instr),
				     NULL ) ) ;
  xsh_msg_debug("tag=%s",tags[0]);
  check(result = xsh_find_frame (frames,(const char**) tags));

 cleanup:
  cpl_free(tags[0]);
  return result;
}


/**
   @brief Find model config frame 
   @param frames The input frameset
   @param instr the instrument arm setting 
    
   @return Pointer to raw frame
    
*/

cpl_frame * xsh_find_model_config_open( cpl_frameset * frames,
				      xsh_instrument * instr )
{
  char *tags[] ={NULL,NULL};
  cpl_frame* result = NULL;

  check(tags[0] = xsh_stringcat_any( XSH_MOD_CFG_OPEN,
				     xsh_instrument_arm_tostring(instr),
				     NULL ) ) ;
  xsh_msg_debug("tag=%s",tags[0]);
  check(result = xsh_find_frame (frames,(const char**) tags));

 cleanup:
  cpl_free(tags[0]);
  return result;
}


/**
   @brief Find model config frame 
   @param frames The input frameset
   @param instr the instrument arm setting 
    
   @return Pointer to raw frame
    
*/

cpl_frame * xsh_find_calpro_model_meas_coord( cpl_frameset * frames,
				      xsh_instrument * instr )
{
  char *tags[] ={NULL,NULL};
  cpl_frame* result = NULL;

  check(tags[0] = xsh_stringcat_any( XSH_MEASCOORD,
				     xsh_instrument_arm_tostring(instr),
				     NULL ) ) ;
  check(result = xsh_find_frame (frames,(const char**) tags));

 cleanup:
  cpl_free(tags[0]);
  return result;
}

/**
   @brief Find model wavelength list frame 
   @param frames The input frameset
   @param instr the instrument arm setting 
    
   @return Pointer to raw frame
    
*/

cpl_frame * xsh_find_model_wavelist( cpl_frameset * frames,
				      xsh_instrument * instr )
{
  char *tags[] ={NULL,NULL};
  cpl_frame* result = NULL;

  check(tags[0] = xsh_stringcat_any( XSH_MODEL_WAVE_LIST,
				     xsh_instrument_arm_tostring(instr),
				     NULL ) ) ;
  check(result = xsh_find_frame (frames,(const char**) tags));

 cleanup:
  cpl_free(tags[0]);
  return result;
}



/**
   @brief Find model test parameter list 
   @param frames The input frameset
   @param instr the instrument arm setting 
    
   @return Pointer to raw frame
    
*/

cpl_frame * xsh_find_model_testpar( cpl_frameset * frames,
				      xsh_instrument * instr )
{
  char *tags[] ={NULL,NULL};
  cpl_frame* result = NULL;

  check(tags[0] = xsh_stringcat_any( XSH_MODEL_TEST_PAR,
				     xsh_instrument_arm_tostring(instr),
				     NULL ) ) ;
  check(result = xsh_find_frame (frames,(const char**) tags));

 cleanup:
  cpl_free(tags[0]);
  return result;
}


/**
   @brief Find spectral format frame 
   @param frames The input frameset
   @param instr the instrument arm setting 
    
   @return Pointer to raw frame
    
*/

cpl_frame * xsh_find_spectral_format( cpl_frameset * frames,
				      xsh_instrument * instr )
{
  char *tags[] ={NULL,NULL};
  cpl_frame* result = NULL;

  check(tags[0] = xsh_stringcat_any( XSH_SPECTRAL_FORMAT,
				     xsh_instrument_arm_tostring(instr),
				     NULL ) ) ;
  check(result = xsh_find_frame (frames,(const char**) tags));

 cleanup:
  cpl_free(tags[0]);
  return result;
}

/*---------------------------------------------------------------------------*/
/**
  @brief Find a frame
  @param frames The input frameset
  @param tags Null-terminated array of tags to search for
  
  @return Pointer to first standard star frame, or NULL on error.

   The function sets an error if no standard star frame is present.
*/
/*---------------------------------------------------------------------------*/
static cpl_frame * 
xsh_find_frame( cpl_frameset *frames, const char *tags[])
{
  cpl_frame *frame = NULL;
  bool found = false;
  char *tags_string = NULL;	/* Used for error message, only */
  char *temp = NULL;
  int i;

  assure (frames != NULL, CPL_ERROR_NULL_INPUT, "Null frameset");
  assure (tags != NULL, CPL_ERROR_NULL_INPUT, "Null tags");
 
  tags_string = xsh_stringdup ("");	/* Message used for empty frame set */

  for (i = 0; !found && tags[i] != NULL; i++) {
    /* Concatenate name to 'tags_string' */
    cpl_free (temp);
    temp = NULL;
    temp = xsh_stringdup (tags_string);

    if (i == 0) {
      cpl_free (tags_string);
      tags_string = NULL;
      check (tags_string = xsh_stringdup (tags[i]));
    }
    else {
      cpl_free (tags_string);
      tags_string = NULL;
      check (tags_string = xsh_stringcat_3 (temp, ", ", tags[i]));
    }

    /* Search for frame */
    frame = cpl_frameset_find (frames, tags[i]);

    if (frame != NULL) {
      found = true;
    }
  }
  if(!found) {
    xsh_msg_warning("%d, Frameset does not contain any %s frame(s)", 
           CPL_ERROR_DATA_NOT_FOUND,tags_string);
  }
 cleanup:
  cpl_free (tags_string);
  cpl_free (temp);

  return frame;
}

/**@}*/
