/*                                                                            *
 *   This file is part of the ESO X-shooter Pipeline                          *
 *   Copyright (C) 2006 European Southern Observatory                         *
 *                                                                            *
 *   This library is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU General Public License as published by     *
 *   the Free Software Foundation; either version 2 of the License, or        *
 *   (at your option) any later version.                                      *
 *                                                                            *
 *   This program is distributed in the hope that it will be useful,          *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *   GNU General Public License for more details.                             *
 *                                                                            *
 *   You should have received a copy of the GNU General Public License        *
 *   along with this program; if not, write to the Free Software              *
 *   Foundation, 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA     *
 *                                                                            */

/*
 * $Author: amodigli $

 * $Date: 2011/01/17 14:53:15 $
 * $Revision: 1.51 $
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

/*----------------------------------------------------------------------------*/
/**
 * @defgroup xsh_remove_crh_single  Remove Cosmic Rays single (xsh_remove_crh_single)
 * @ingroup drl_functions
 *
 * Suppress Cosmic Rays by analysing on files
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*-----------------------------------------------------------------------------
  Includes
-----------------------------------------------------------------------------*/

#include <math.h>

#include <xsh_drl.h>
#include <xsh_data_pre.h>
#include <xsh_dfs.h>
#include <xsh_pfits.h>
#include <xsh_error.h>
#include <xsh_msg.h>
#include <cpl.h>
#include <xsh_badpixelmap.h>
#include <xsh_parameters.h>
#include <xsh_utils_wrappers.h>

/*-----------------------------------------------------------------------------
  Functions prototypes
 -----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
  Implementation
 -----------------------------------------------------------------------------*/

/* This should be defined in a more clever way, a parameter for example */
#define MAX_ITERATIONS 6
#define REGDEBUG_FULL 0




/** 
 * Calculates and set the QC parameters relevant for Cosmic Rays.
 * 
 * @param pre Pointer to XSH_PRE Structure
 * @param nbcrh Total umber of cosmics found
 * @param nframes Number of frames used to remove cosmics
 * @param instrument Pointer to indstrument description structure
 */
static void add_qc_crh (xsh_pre* pre, int nbcrh, int nframes,
			xsh_instrument * instrument )
{
  double qc_crrate = 0.0;
  int nbcr_avg ;		/**< Average of cosmics per frame */

  XSH_ASSURE_NOT_NULL(pre);
  XSH_ASSURE_NOT_ILLEGAL(pre->pszx >0. && pre->pszy > 0); 
  XSH_ASSURE_NOT_ILLEGAL(pre->exptime > 0); 
  /*
    QC.CRRATE = avg_nb_of_crh_per_frame/(exposure_time*size_of_detector_in_cm2)
    Nb of Cosmics per cm2 per second.
    The whole size of the image in cm2 is given by pszx and pszy. These
    values are in MICRONS.
  */
  xsh_msg_dbg_medium( "add_qc_crh - Exptime = %f", pre->exptime ) ;
  qc_crrate =
    (double) nbcrh / (((double) nframes) * (pre->exptime * 
    (pre->pszx / 10000.0) * (pre->pszy /10000.0) * pre->nx * pre->ny)) ;
  nbcr_avg = nbcrh/nframes ;

  /*
    Now set QC KW
  */
  check( xsh_pfits_set_qc_crrate( pre->data_header,qc_crrate) ) ;
  check( xsh_pfits_set_qc_ncrh( pre->data_header,nbcrh) ) ;
  check( xsh_pfits_set_qc_ncrh_mean( pre->data_header,nbcr_avg) ) ;
  /* Same thing for the qual header */
  check( xsh_pfits_set_qc_crrate( pre->qual_header,qc_crrate) ) ;
  check( xsh_pfits_set_qc_ncrh( pre->qual_header,nbcrh) ) ;
  check( xsh_pfits_set_qc_ncrh_mean( pre->qual_header,nbcr_avg) ) ;
  cleanup:
    return ;
}





cpl_frame * xsh_abs_remove_crh_single( cpl_frame *sci_frame,
                                   cpl_frame* wavemap_frame,
                                   xsh_instrument * instrument,
                                   xsh_remove_crh_single_param * single_par,
                                   const char * res_tag)
{
  cpl_frame *result_frame = NULL;
  cpl_frame *abs_frame = NULL;
  cpl_frame *sign_frame = NULL;
  cpl_frame *rmcrh_frame = NULL;
  const char* name = NULL;
  cpl_propertylist *header = NULL;
  int qc_ncrh =0;

  XSH_ASSURE_NOT_NULL( sci_frame);

  check( name = cpl_frame_get_filename( sci_frame));
  check( header = cpl_propertylist_load( name, 0));
  if(cpl_propertylist_has(header,XSH_QC_CRH_NUMBER)) {
     check( qc_ncrh = xsh_pfits_get_qc_ncrh( header));
  } else {
     qc_ncrh =0;
  }
  /* cosmics detection already done */
  if (qc_ncrh > 0){
    xsh_msg( "Not use remove crh single");
    check( result_frame = cpl_frame_duplicate( sci_frame));
  }
  else{
    check( abs_frame = xsh_frame_abs( sci_frame, instrument, &sign_frame));
    check( rmcrh_frame = xsh_remove_crh_single( abs_frame, wavemap_frame, instrument,
      single_par, res_tag));
    check( result_frame = xsh_frame_mult(  rmcrh_frame, instrument, sign_frame));
  }

  cleanup:
    if (cpl_error_get_code() != CPL_ERROR_NONE){
      xsh_free_frame( &result_frame);
    }
    xsh_free_propertylist( &header);
    xsh_free_frame( &abs_frame);
    xsh_free_frame( &sign_frame);
    xsh_free_frame( &rmcrh_frame);
    return result_frame;
}

/*****************************************************************************/
/** 
  @brief
    Remove cosmic rays from a single frame.
 @param[in] in_sci_frame 
   The science Frame
 @param[in] wavemap 
   The wave map Frame
 @param[in] instrument 
   Pointer to instrument description
 @param[in] single_par 
   Parameters for remove crh single
 @param[in] res_tag 
   Parameters for result tag 
 @return 
   The science frame after removal of Cosmics
*/
/*****************************************************************************/
cpl_frame * xsh_remove_crh_single( cpl_frame * in_sci_frame,
                                   cpl_frame* wavemap,
				   xsh_instrument * instrument, 
				   xsh_remove_crh_single_param * single_par,
				   const char * res_tag )
{
  int i,j,k,l,m;
  double sigma_lim, f_lim, frac_max, frac = 0.;
  /* Only pointers */
  cpl_frame* res_frame = NULL ;
  int *qual = NULL;
  char* res_name=NULL;
  /* Need to be free */
  cpl_image* laplacian_image = NULL;
  cpl_image* laplacian_redu_image = NULL;
  cpl_image* two_sub_sample = NULL;
  cpl_image* sci_median5_image = NULL;
  cpl_image* noise_image = NULL;
  cpl_image* s_image = NULL;
  cpl_image* s_median_image = NULL;
  cpl_image* s2_image = NULL;
  cpl_image* sci_median3_image = NULL;
  cpl_image* sci_median3_7_image = NULL;
  cpl_image* f_image = NULL;
  cpl_image* r_image = NULL;
  cpl_image* c_image = NULL;
  int two_sub_sample_nx = 0;
  int two_sub_sample_ny = 0;
  xsh_pre* in_sci_pre = NULL;	/**< Input PRE */
  xsh_pre* sci_pre = NULL;	/**< Result PRE */
  cpl_image* sci_image = NULL;	/**< Result image */
  /* Only pointers */
  float* sci_data = NULL;
  float* two_sub_sample_data = NULL;
  float* laplacian_data = NULL;
  float* laplacian_redu_data = NULL;
  float* sci_median5_data = NULL;
  float* sci_median3_data = NULL;
  float* sci_median3_7_data = NULL;
  float* noise_data = NULL;
  float* s_data = NULL;
  float* s_median_data = NULL;
  float* s2_data = NULL;
  float* f_data = NULL;
  float* r_data = NULL;
  int* c_data = NULL;
  /* Need to be free */
 
  cpl_matrix* laplacian_kernel = NULL;
  cpl_matrix* median3_kernel = NULL;
  cpl_matrix* median5_kernel = NULL;
  cpl_matrix* median7_kernel = NULL;
  int new_crh =1, nb_crh = 0;
  int nbiter = 1, max_iter = MAX_ITERATIONS ;
  cpl_vector* median = NULL;
  cpl_image* w_image=NULL;
  double* w_data=NULL;
  const char* w_name=NULL;
  int fctx=1;
  int fcty=1;
  int wmap_xsize_diff=0;
  double ron, conad;

  /* Check parameters */
  XSH_ASSURE_NOT_NULL_MSG( in_sci_frame,"Null input science frame" ) ;
  XSH_ASSURE_NOT_NULL_MSG( single_par,"Null input parameters" ) ;
  XSH_ASSURE_NOT_NULL_MSG( instrument,"Null instrument setting" ) ;

  sigma_lim = single_par->sigma_lim;
  f_lim = single_par->f_lim;
  frac_max = single_par->crh_frac_max ;
  max_iter = single_par->nb_iter ;
  xsh_msg_dbg_high( "  Params: frac_max %.1f, sigma_lim %.2f f_lim %.2f, iter %d",
	   single_par->crh_frac_max, sigma_lim, f_lim, max_iter);


  check( in_sci_pre = xsh_pre_load( in_sci_frame, instrument));
  ron = in_sci_pre->ron;
  conad = in_sci_pre->conad;
  check( sci_pre = xsh_pre_duplicate( in_sci_pre ) ) ;
  xsh_pre_free(&in_sci_pre);
  check( sci_image = xsh_pre_get_data( sci_pre));
  check ( qual = cpl_image_get_data_int( sci_pre->qual));

  if(wavemap!=NULL) {
     check( w_name = cpl_frame_get_filename( wavemap));
     check( w_image = cpl_image_load( w_name, CPL_TYPE_DOUBLE,0, 0));
  } else {
     /* not to fail the corresponding unit test we use a dummy wavemap
        here used only to mask out mpixels */
     check( w_image = cpl_image_new(sci_pre->nx,sci_pre->ny,CPL_TYPE_DOUBLE));
     cpl_image_add_scalar(w_image,1.);
  }

  fctx=cpl_image_get_size_x(w_image)/sci_pre->nx;
  fcty=cpl_image_get_size_y(w_image)/sci_pre->ny;
  if( xsh_instrument_get_arm(instrument) == XSH_ARM_NIR){
     fctx = (fctx<1) ? 1 : fctx;
     fcty = (fcty<1) ? 1 : fcty;
     wmap_xsize_diff=sci_pre->nx-cpl_image_get_size_x(w_image);
  }

  xsh_msg_dbg_low("fctx=%d fcty=%d",fctx,fcty);
  xsh_msg_dbg_low("sci nx=%d ny=%d",sci_pre->nx,sci_pre->ny);

  check(w_data=cpl_image_get_data_double(w_image));

  /* Preparing different kernels */

  xsh_msg_dbg_low("--remove-crh-single : Preparing kernels");
  /* Laplacian */
  check( laplacian_kernel = cpl_matrix_new(3,3));
  cpl_matrix_set( laplacian_kernel,0,0,0.0);
  cpl_matrix_set( laplacian_kernel,0,1,-1.0);
  cpl_matrix_set( laplacian_kernel,0,2,0.0);
  cpl_matrix_set( laplacian_kernel,1,0,-1.0);
  cpl_matrix_set( laplacian_kernel,1,1,4.0);
  cpl_matrix_set( laplacian_kernel,1,2,-1.0);
  cpl_matrix_set( laplacian_kernel,2,0,0.0);
  cpl_matrix_set( laplacian_kernel,2,1,-1.0);
  cpl_matrix_set( laplacian_kernel,2,2,0.0);
  cpl_matrix_divide_scalar( laplacian_kernel, 4.0); 	
/*
  cpl_matrix_set( laplacian_kernel,0,0,-1.0);
  cpl_matrix_set( laplacian_kernel,0,1,-1.0);
  cpl_matrix_set( laplacian_kernel,0,2,-1.0);
  cpl_matrix_set( laplacian_kernel,1,0,-1.0);
  cpl_matrix_set( laplacian_kernel,1,1,8.0);
  cpl_matrix_set( laplacian_kernel,1,2,-1.0);
  cpl_matrix_set( laplacian_kernel,2,0,-1.0);
  cpl_matrix_set( laplacian_kernel,2,1,-1.0);
  cpl_matrix_set( laplacian_kernel,2,2,-1.0);
  cpl_matrix_divide_scalar( laplacian_kernel, 8.0);
*/
  /* Median 3x3*/
  check( median3_kernel = cpl_matrix_new(3,3));
  for(j=0; j< 3; j++){
    for(i=0; i< 3; i++){
      cpl_matrix_set( median3_kernel, i,j,1.0);
    }
  }

  /* Median 5x5 */
  check( median5_kernel = cpl_matrix_new(5,5));
  for(j=0; j< 5; j++){
    for(i=0; i< 5; i++){
      cpl_matrix_set( median5_kernel, i,j,1.0);
    }
  }

  /* Median 7x7 */
  check( median7_kernel = cpl_matrix_new(7,7));
  for(j=0; j< 7; j++){
    for(i=0; i< 7; i++){
      cpl_matrix_set( median7_kernel, i,j,1.0);
    }
  }


  /* Allocate images and pointers */
  check (sci_data = cpl_image_get_data_float( sci_image));

  two_sub_sample_nx = sci_pre->nx*2;
  two_sub_sample_ny = sci_pre->ny*2;
  check( two_sub_sample = cpl_image_new( two_sub_sample_nx,
    two_sub_sample_ny, CPL_TYPE_FLOAT));
  check (two_sub_sample_data = cpl_image_get_data_float( two_sub_sample));
  check( laplacian_redu_image = cpl_image_new( sci_pre->nx,
    sci_pre->ny, CPL_TYPE_FLOAT));
  check (laplacian_redu_data = cpl_image_get_data_float(
    laplacian_redu_image));
  check( noise_image = cpl_image_new( sci_pre->nx,
    sci_pre->ny, CPL_TYPE_FLOAT));
  check ( noise_data = cpl_image_get_data_float( noise_image));
  check( s_image = cpl_image_new( sci_pre->nx,
    sci_pre->ny, CPL_TYPE_FLOAT));
  check ( s_data = cpl_image_get_data_float( s_image));
  check( s2_image = cpl_image_new( sci_pre->nx,
    sci_pre->ny, CPL_TYPE_FLOAT));
  check ( s2_data = cpl_image_get_data_float( s2_image));
  check( f_image = cpl_image_new( sci_pre->nx,
    sci_pre->ny, CPL_TYPE_FLOAT));
  check( f_data = cpl_image_get_data_float( f_image));
  check( r_image = cpl_image_new( sci_pre->nx,
    sci_pre->ny, CPL_TYPE_FLOAT));

  check( r_data = cpl_image_get_data_float( r_image));
  check( c_image = cpl_image_new( sci_pre->nx,
    sci_pre->ny, CPL_TYPE_INT));
  check( c_data = cpl_image_get_data_int( c_image));


  /* LGG - Added limit on frac_max AND limit on nb iterations */
  while( new_crh > 0 && frac < frac_max && nbiter <= max_iter ){
    xsh_msg_dbg_low("Iteration %d",nbiter);
    /* Create a 2n x 2n images like this 
        | 1 | 2 |  =>  | 1 | 1 | 2 | 2 |
        | 3 | 4 |      | 1 | 1 | 2 | 2 |
                       | 3 | 3 | 4 | 4 |
                       | 3 | 3 | 4 | 4 | */
    xsh_msg_dbg_medium("Create a 2n images");
    for( j=0; j< sci_pre->ny; j++){
      for( i=0; i< sci_pre->nx; i++){
        float val = sci_data[i+j*sci_pre->nx];

	if ( val < 0. ) val = 0. ;
        two_sub_sample_data[i*2+j*2*two_sub_sample_nx] = val;
        two_sub_sample_data[i*2+1+j*2*two_sub_sample_nx] = val;
        two_sub_sample_data[i*2+(j*2+1)*two_sub_sample_nx] = val;
        two_sub_sample_data[i*2+1+(j*2+1)*two_sub_sample_nx] = val;
      }
    }
    xsh_msg_dbg_medium("Doing laplacian convolution");
    /* Doing the laplacian convolution 
        0  -1   0
       -1   4  -1 
        0  -1   0 */
    check(laplacian_image = xsh_image_filter_linear( two_sub_sample, 
						     laplacian_kernel));

    /* multiply by two to normalize correctly the laplacian [RD5] 
       and filter negative values */
     xsh_msg_dbg_medium("Normalize laplacian");
    check (laplacian_data = cpl_image_get_data_float( laplacian_image));
    for ( i=0; i< two_sub_sample_nx*two_sub_sample_ny; i++){
      if (laplacian_data[i] > 0.0){
        laplacian_data[i] = 2.0 * laplacian_data[i]; 
      }
      else{
        laplacian_data[i] = 0.0;
      }
    }
#if REGDEBUG_FULL
    xsh_msg_dbg_medium("Save Lpositive");
    cpl_image_save(laplacian_image, "Lpositive.fits", CPL_BPP_IEEE_FLOAT, NULL,
      CPL_IO_DEFAULT);
#endif
    /* resample to the original size 
       | 1 | 1 | 2 | 2 |    | 1 | 2 | 
       | 1 | 1 | 2 | 2 |    | 3 | 4 |
       | 3 | 3 | 4 | 4 | => 
       | 3 | 3 | 4 | 4 |               */

    xsh_msg_dbg_medium("Resample to the original size");

    for( j=0; j< sci_pre->ny; j++){
      for( i=0; i< sci_pre->nx; i++){
        laplacian_redu_data[i+j*sci_pre->nx] = 
          (laplacian_data[i*2+j*2*two_sub_sample_nx]+
           laplacian_data[i*2+1+j*2*two_sub_sample_nx]+
           laplacian_data[i*2+(j*2+1)*two_sub_sample_nx]+
           laplacian_data[i*2+1+(j*2+1)*two_sub_sample_nx])/4.0;
      }
    }
#if REGDEBUG_FULL
    cpl_image_save(laplacian_redu_image, "Lplus.fits", CPL_BPP_IEEE_FLOAT, 
      NULL, CPL_IO_DEFAULT);
#endif
    xsh_msg_dbg_medium("Apply median filter");
    /* Apply 5x5 median filter on data */
    check( sci_median5_image = xsh_image_filter_median( sci_image, 
      median5_kernel)); 
    check (sci_median5_data = cpl_image_get_data_float( sci_median5_image));

    xsh_msg_dbg_medium("Compute noise");
    /* computes the noise image */
    for( i=0; i< sci_pre->nx*sci_pre->ny; i++){
      noise_data[i] = sqrt(sci_median5_data[i]*conad+ron*ron)/ conad;
    }

    xsh_msg_dbg_medium("Compute S");
    /* compute S image */
    for( i=0; i< sci_pre->nx*sci_pre->ny; i++){
      s_data[i] = laplacian_redu_data[i] / (2.0*noise_data[i]);
    }

    xsh_msg_dbg_medium("Compute S median");
    /* compute S median image */
    check( s_median_image = xsh_image_filter_median( s_image,
      median5_kernel));

    check( s_median_data = cpl_image_get_data_float( s_median_image));

    xsh_msg_dbg_medium("Compute s2");
    /* compute s2 */
    for( i=0; i< sci_pre->nx*sci_pre->ny; i++){
      s2_data[i] = s_data[i] -s_median_data[i];
    }
#if REGDEBUG_FULL
    cpl_image_save( s2_image, "S2.fits", CPL_BPP_IEEE_FLOAT, NULL,
	 CPL_IO_DEFAULT);
#endif
    xsh_msg_dbg_medium("Apply 3x3 filter");
    /* Apply 3x3 median filter on data */
    check( sci_median3_image = xsh_image_filter_median( sci_image,
      median3_kernel));

    xsh_msg_dbg_medium("Apply 7x7 filter");
    /* Apply 7x7 median filter */ 
    check( sci_median3_7_image = xsh_image_filter_median( sci_median3_image,
      median7_kernel));
    xsh_msg_dbg_medium("Apply 7x7 filter ok");

    check ( sci_median3_data = cpl_image_get_data_float( sci_median3_image));
    check ( sci_median3_7_data = cpl_image_get_data_float( 
      sci_median3_7_image));

    xsh_msg_dbg_medium("Compute F");
    /* compute F */
    for( i=0; i< sci_pre->nx*sci_pre->ny; i++){
      f_data[i] = sci_median3_data[i] -sci_median3_7_data[i];
      if (f_data[i] < 0.01){
        f_data[i] = 0.01;
      }
    }
#if REGDEBUG_FULL
    cpl_image_save( f_image, "F.fits", CPL_BPP_IEEE_FLOAT, NULL,
      CPL_IO_DEFAULT);
#endif
    xsh_msg_dbg_medium("Compute R");
    /* compute R */
    for( i=0; i< sci_pre->nx*sci_pre->ny; i++){
      r_data[i] = laplacian_redu_data[i]/f_data[i];
    }
#if REGDEBUG_FULL
    cpl_image_save( r_image, "R.fits", CPL_BPP_IEEE_FLOAT, NULL,
      CPL_IO_DEFAULT);
#endif
    /* Search for cosmics */
    xsh_msg_dbg_medium("Search for cosmic");
    new_crh = 0;
    median = cpl_vector_new(24);

    for( j=1; j< sci_pre->ny-1; j++){
       double *data = NULL;
       cpl_vector* med_vect = NULL;


       for( i=1; i< sci_pre->nx-1-wmap_xsize_diff; i++){
          if(w_data[i*fctx+j*fcty*(sci_pre->nx-wmap_xsize_diff)] != 0) {
             if ( (s2_data[i+j*sci_pre->nx] >= sigma_lim) && 
                  (r_data[i+j*sci_pre->nx] >= f_lim)){
                int li,lj,ui,uj;
                c_data[i+j*sci_pre->nx] = 1;
                new_crh++;
                li = i-2;
                lj = j-2;
                ui = i+2;
                uj = j+2;
                m = 0;
                if (li < 0) li = 0;
                if (ui >= sci_pre->nx) ui = sci_pre->nx-1;
                if (lj < 0) lj = 0;
                if (uj >= sci_pre->ny) uj = sci_pre->ny-1;
                for( k=lj; k <= uj; k++){
                   for( l=li; l <= ui; l++){
                      if ( k < j){
                         cpl_vector_set(median, m, sci_data[l+k*sci_pre->nx]);
                         m++;
                      }
                      else if ( (k == j) && ( l < i)){
                         cpl_vector_set(median, m, sci_data[l+k*sci_pre->nx]);
                         m++;
                      }
                      else if ( l!=i && k!=j && (s2_data[l+k*sci_pre->nx] < sigma_lim)
                                && (r_data[l+k*sci_pre->nx] < f_lim)){
                         cpl_vector_set(median, m, sci_data[l+k*sci_pre->nx]);
                         m++;
                      }
                   }
                }
                check( data = cpl_vector_get_data( median));
                //XSH_REGDEBUG("i %d j %d m %d", i, j ,m);
                check( med_vect = cpl_vector_wrap( m, data));
                check( sci_data[i+j*sci_pre->nx] = cpl_vector_get_median( med_vect));
                cpl_vector_unwrap( med_vect);
             } /* end check if image data point is above given threshols (a CRH) */
          } /* end check if points are on orders */
       } /* end loop over image columns */
    } /* end loop over image raws */
    xsh_free_vector( &median ) ;
    nb_crh += new_crh;
    frac = (double)nb_crh/(double)(sci_pre->nx*sci_pre->ny) ;
    xsh_msg_dbg_low("   new cosmics %d, total %d, frac %.4f [%d pixels]",new_crh,nb_crh,
	    frac, sci_pre->nx*sci_pre->ny);
    nbiter++;
    xsh_free_image( &laplacian_image);
    xsh_free_image( &sci_median3_7_image ) ;
    xsh_free_image( &sci_median3_image ) ;
    xsh_free_image( &s_median_image ) ;
    xsh_free_image( &sci_median5_image ) ;
  }
#if REGDEBUG_FULL
  cpl_image_save( c_image, "CRH_SINGLE.fits", CPL_BPP_32_SIGNED, NULL,
      CPL_IO_DEFAULT);
#endif

  if ( xsh_debug_level_get() >= XSH_DEBUG_LEVEL_MEDIUM){
    FILE *debug = NULL;

    debug = fopen("cosmic.log","w");
    
    for( j=0; j< sci_pre->ny; j++){
      for( i=0; i< sci_pre->nx; i++){
        if ( c_data[i+j*sci_pre->nx] == 1){
          fprintf(debug,"%.1f %.1f\n",i+1.0,j+1.0);
        }
      }
    }
    fclose(debug);
  }

  /* flag cosmic in the BP_MAP */
  for( j=0; j< sci_pre->ny; j++){
    for( i=0; i< sci_pre->nx; i++){
      if ( c_data[i+j*sci_pre->nx] == 1){
        qual[i+j*sci_pre->nx] |= QFLAG_COSMIC_RAY_REMOVED;
      }
    }
  } 

  //check( res_frame = cpl_frame_duplicate( sci_frame ) ) ;
  res_name=xsh_stringcat_any(res_tag,".fits", NULL) ;
  xsh_msg_dbg_high( "Saving Result Frame '%s'", res_name ) ;
  check( add_qc_crh( sci_pre, nb_crh, 1, instrument ) ) ;
  check( res_frame = xsh_pre_save( sci_pre, res_name, res_tag,1 ) ) ;
  check( cpl_frame_set_tag( res_frame, res_tag ) ) ;
  XSH_FREE( res_name);

  cleanup:
    xsh_pre_free( &sci_pre);

    /* free kernel */
    xsh_free_matrix( &laplacian_kernel);
    xsh_free_matrix( &median3_kernel);
    xsh_free_matrix( &median5_kernel);
    xsh_free_matrix( &median7_kernel);
    /* free images */
    xsh_free_image( &laplacian_image);
    xsh_free_image( &laplacian_redu_image);
    xsh_free_image( &two_sub_sample);
    xsh_free_image( &sci_median5_image);
    xsh_free_image( &noise_image);
    xsh_free_image( &s_image);
    xsh_free_image( &s_median_image); 
    xsh_free_image( &s2_image);
    xsh_free_image( &sci_median3_image);
    xsh_free_image( &sci_median3_7_image);
    xsh_free_image( &f_image);
    xsh_free_image( &r_image);
    xsh_free_image( &c_image);
    xsh_free_image( &w_image);
    /* free vector */
    xsh_free_vector( &median);
    /* free tab */
     return res_frame;
}
/**@}*/

