/*                                                                           *
 *   This file is part of the ESO X-shooter Pipeline                         *
 *   Copyright (C) 2006 European Southern Observatory                        *
 *                                                                           *
 *   This library is free software; you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by    *
 *   the Free Software Foundation; either version 2 of the License, or       *
 *   (at your option) any later version.                                     *
 *                                                                           *
 *   This program is distributed in the hope that it will be useful,         *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *   GNU General Public License for more details.                            *
 *                                                                           *
 *   You should have received a copy of the GNU General Public License       *
 *   along with this program; if not, write to the Free Software             *
 *   Foundation, 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA    *
 *                                                                           */

/*
 * $Author: amodigli $
 * $Date: 2011/01/10 09:11:01 $
 * $Revision: 1.108 $
 * $Name: HEAD $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*---------------------------------------------------------------------------*/
/**
 * @defgroup xsh_subtract Subtraction of Frames (xsh_subtract)
 * @ingroup drl_functions
 *
 * Contains functions to subtract various frames. 
 */
/*---------------------------------------------------------------------------*/
/**@{*/

/*---------------------------------------------------------------------------
                                Includes
 ----------------------------------------------------------------------------*/
#include <math.h>
#include <xsh_utils.h>
#include <xsh_drl.h>
#include <xsh_pfits.h>
#include <xsh_error.h>
#include <xsh_msg.h>
#include <xsh_badpixelmap.h>
#include <xsh_utils_image.h>
#include <xsh_data_order.h>
#include <xsh_data_grid.h>

/*---------------------------------------------------------------------------
                            Typedefs
  ---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------
                            Functions prototypes
  ---------------------------------------------------------------------------*/
static cpl_error_code
subtract_background(cpl_image *image, cpl_image *background_im, 
		    const polynomial *background_pol);

static cpl_image*
xsh_background_poly(cpl_image* image, xsh_grid* grid);

/*---------------------------------------------------------------------------
                              Implementation
  ---------------------------------------------------------------------------*/


/** 
  @brief
    Subtract the master bias frame from PRE frame
 @param frame
   The frame from we subtracted the master bias
 @param bias 
   The master bias frame
 @param instr
   The xsh instrument
 @param type  
   frame type used to define bias subtracted frame filename
 @param pre_overscan_corr 
   switch on/off pre-overscan correction
 @return the subtracted frame
*/

cpl_frame* 
xsh_subtract_bias( cpl_frame *frame, 
                   cpl_frame *bias,
		   xsh_instrument* instr, 
                   const char * type, 
                   const int pre_overscan_corr)
{
  /* MUST BE DEALLOCATED in caller */
  cpl_frame * result = NULL;
  /* ALLOCATED locally */
  xsh_pre* xframe = NULL;
  xsh_pre* xbias = NULL;
  /* Others */
  char resultname[256];
  int binx=0;
  int biny=0;
  char tag[80];
  double avg=0;

  /* check input */  
  XSH_ASSURE_NOT_NULL( frame);
  XSH_ASSURE_NOT_NULL( bias);
  XSH_ASSURE_NOT_NULL( instr);

  /* load the frames */
  check( xframe = xsh_pre_load( frame, instr));
  check( xbias = xsh_pre_load( bias, instr)); 

  check(binx=xsh_pfits_get_binx(xbias->data_header));
  check(biny=xsh_pfits_get_biny(xbias->data_header));
  /* do the subtract operation */
  if(pre_overscan_corr==0) {
    check( xsh_pre_subtract( xframe, xbias));
  } else {
    check(avg=cpl_image_get_mean(xbias->data));
    check(xsh_pre_subtract_scalar(xbias,avg));
    check( xsh_pre_subtract( xframe, xbias));
  }
  /* save result */
  sprintf(tag,"%sON_%s",type,xsh_instrument_arm_tostring (instr));

  /* Sabine does not like binning in file name
  sprintf(resultname,"%s_%dx%d.fits",tag,binx,biny);
  */
  sprintf(resultname,"%s.fits",tag);

  check( xsh_pfits_set_pcatg( xframe->data_header, tag));
  check( result = xsh_pre_save ( xframe, resultname, tag,1));
  check( cpl_frame_set_tag( result,tag));

  cleanup:
    if ( cpl_error_get_code() != CPL_ERROR_NONE) {
      xsh_free_frame( &result);
    }
    xsh_pre_free( &xframe);
    xsh_pre_free( &xbias);
    return result;
}



/*---------------------------------------------------------------------------*/
/**
 * @brief (NIR only) subtract the OFF set of files from the On set of files 
 * @param on the ON set of files
 * @param off the OFF set of files
 * @param instr instrument containing the arm , mode and lamp in use
 * @return the ON subtract OFF set of files
 */

/*---------------------------------------------------------------------------*/
cpl_frameset* xsh_subtract_nir_on_off(cpl_frameset* on, cpl_frameset* off, 
  xsh_instrument* instr){
  /* MUST BE DEALLOCATED in caller */
  cpl_frameset * result = NULL;
  /* ALLOCATED locally */
  cpl_frame* res = NULL;
  /* Others */
  char resultname[256];
  int i = 0, size_on = 0, size_off = 0;

  /* check input parameters */
  XSH_ASSURE_NOT_NULL(on);
  XSH_ASSURE_NOT_NULL(off);
  XSH_ASSURE_NOT_NULL(instr);

  check(size_on = cpl_frameset_get_size(on));
  check(size_off = cpl_frameset_get_size(off));

  XSH_ASSURE_NOT_ILLEGAL(size_on == size_off); 

  XSH_NEW_FRAMESET(result);

  /* Loop on frames */
  for ( i=0; i< size_on; i++) { 
    cpl_frame* on_f = NULL;
    cpl_frame* off_f = NULL;

    check( on_f = cpl_frameset_get_frame(on,i));
    check( off_f = cpl_frameset_get_frame(off,i));
    sprintf(resultname, "ON-OFF_%d.fits",i);

    check( res = xsh_subtract_dark(on_f, off_f, resultname, instr));
    check(cpl_frameset_insert(result, res));
  }
  
  cleanup:
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
      xsh_free_frameset(&result);
      xsh_free_frame(&res);
    }
    return result; 
}
/*---------------------------------------------------------------------------*/
/** 
 * @brief subtract the master dark frame from PRE frame
 * @param frame frame to be sutracted
 * @param dark the master dark frame
 * @param filename the result frame filename
 * @param instr instrument containing the arm , mode and lamp in use
 * @return the subtracted frame
 */
/*---------------------------------------------------------------------------*/
cpl_frame * xsh_subtract_dark(cpl_frame * frame,cpl_frame * dark,
  const char* filename, xsh_instrument* instr)
{
  /* MUST BE DEALLOCATED in caller */
  cpl_frame * result = NULL;
  /* ALLOCATED locally */
  xsh_pre* xframe = NULL;
  xsh_pre* xdark = NULL;
  /* Others */
  double exptime = 0.0;  
  const char* tag=NULL;

  /* check input parameters */
  XSH_ASSURE_NOT_NULL(frame);
  XSH_ASSURE_NOT_NULL(dark);
  XSH_ASSURE_NOT_NULL(filename);
  XSH_ASSURE_NOT_NULL(instr);

  /* load the frames */
  check(xframe = xsh_pre_load(frame, instr));
  check(xdark = xsh_pre_load(dark, instr));
 
  /* compute the master to the same exptime */
  if (xsh_instrument_get_arm(instr) != XSH_ARM_NIR) {
    exptime = xframe->exptime;
    assure(exptime > 0,CPL_ERROR_ILLEGAL_INPUT,
      "EXPTIME must be greater than 0 : %f",exptime);
    check(xsh_pre_multiply_scalar(xdark,exptime));
  }
  /* do the subtract operation */
  check(xsh_pre_subtract(xframe,xdark));
  /* save result */
  tag=cpl_frame_get_tag(frame);
  check(result = xsh_pre_save (xframe, filename, tag,1));
  check(cpl_frame_set_tag (result,tag));
  
  cleanup:
    if (cpl_error_get_code () != CPL_ERROR_NONE) {
      xsh_free_frame(&result);
    }
    xsh_pre_free(&xframe);
    xsh_pre_free(&xdark);
  return result;
}

/*---------------------------------------------------------------------------*/
/** 
  @brief
    Subtract the background from PRE frame
  @param[in] frame
    The frame from we subtract the background
  @param[in] edges_order_tab_frame
    The table where the orders positions are
  @param[in] background_par
    The background_img parameters structure
  @param[in] instr 
    The instrument containing the arm , mode and lamp in use
  @param[in] prefix
  @param[out] grid_frame grid sampling point frame
  @param[out] backg_frame background frame

  @return
    The subtracted frame
*/
/*---------------------------------------------------------------------------*/
cpl_frame* xsh_subtract_background( cpl_frame *frame, 
				    cpl_frame *edges_order_tab_frame,
				    xsh_background_param *background_par,
				    xsh_instrument *instr,
				    const char *prefix,
                                    cpl_frame **grid_frame,
                                    cpl_frame **backg_frame)
{
  /* MUST BE DEALLOCATED in caller */
  cpl_frame *result = NULL;
  /* ALLOCATED locally */
  xsh_order_list *orderlist = NULL;
  xsh_pre *pre = NULL;
  xsh_grid *grid = NULL;
  cpl_image *background_img = NULL;
  /* Others */
  float *data = NULL;
  int incr = 0, i = 0, y =0, lost = 0;
  int grid_size_x =0, grid_size_y =0;  
  double *tab = NULL;
  int x = 0;
  float x1 = 0, x2 =0, dx = 0;
  double medflux = 0.0;
  char *fname = NULL;
  char *pcatg = NULL;
  cpl_image *background_smo = NULL;
  int smooth_x = -1;
  int smooth_y = -1;
  cpl_propertylist *plist = NULL;
  cpl_table *grid_tbl = NULL;
    cpl_image* ima=NULL;

  /* check parameters */
  XSH_ASSURE_NOT_NULL( instr);
  XSH_ASSURE_NOT_NULL( frame);
  XSH_ASSURE_NOT_NULL( edges_order_tab_frame);
  XSH_ASSURE_NOT_NULL( background_par);
  /*
  XSH_ASSURE_NOT_NULL( grid_frame);
  XSH_ASSURE_NOT_NULL( backg_frame);
  */

  XSH_ASSURE_NOT_ILLEGAL(background_par->sampley > 0);
  XSH_ASSURE_NOT_ILLEGAL(background_par->radius_x >= 0);
  XSH_ASSURE_NOT_ILLEGAL(background_par->radius_y >= 0);
 
  /* Parameters */
  
  /* load data */
  check( pre = xsh_pre_load(frame, instr));
  check( orderlist = xsh_order_list_load( edges_order_tab_frame, instr));
  xsh_order_list_set_bin_x( orderlist, pre->binx);
  xsh_order_list_set_bin_y( orderlist, pre->biny);
  check( data = cpl_image_get_data_float(pre->data));

  /* create the grid */
  grid_size_x = orderlist->size+5;
  grid_size_y = background_par->sampley+2;

  smooth_x = background_par->smooth_x;
  smooth_y = background_par->smooth_y;

  check( grid = xsh_grid_create( grid_size_x*grid_size_y));

  /* allocate share memory for medflux */
  XSH_MALLOC(tab,double,(2*background_par->radius_x+1)*(2*background_par->radius_y+1));

  /* compute incr along Y axis */
  incr = (int)ceil((float)pre->ny/(float)background_par->sampley);

  /* Step 1 : Fill the grid */
  for( y=1; y <= (pre->ny-1+incr); y += incr){
    int sizey =  2 * background_par->radius_y+1;
    int oy = -background_par->radius_y;
    
    /* force to do the last line */
    if( (y+incr) > pre->ny){
      y = pre->ny;
    }
  
    /* special cases in Y */
    if((y == 1) || (y == pre->ny)){
      sizey = background_par->radius_y+1;
      oy = 0;
    }

    /* X = 1 */
    check( medflux = xsh_pre_data_window_sample_flux_pa(pre,
							1, y+oy, 
							2 * background_par->radius_x+1,
                                                        sizey, 
                                                        tab,
                                                        background_par->method,
							background_par->min_frac_grid));
    /* special case for NIR */
    if ( xsh_instrument_get_arm(instr) == XSH_ARM_NIR){
      if ( (y <= orderlist->list[orderlist->size-2].starty) || 
        y >= orderlist->list[orderlist->size-2].endy){
        /* add to grid */
        check( xsh_grid_add(grid, 1, y, medflux));
      }
    }
    else{
      check( xsh_grid_add(grid, 1, y, medflux));
    }

    /* Points in [orderN->edgeup, order0->edgeup] */
    for( i=orderlist->size-1; i > 0; i--){
      int starty, endy;
      
      starty = orderlist->list[i-1].starty;
      endy = orderlist->list[i-1].endy;


      if(starty == -999 ) {
          /* this condition can happen only if qc_mode=TRUE and some order 
             has edges with S/N less than the user defined threshold
             it is used for QC in order to make recipe more robust and 
             generate all QC. Normal users should set qc_mode=FALSE. 
          */
	xsh_msg("PROBLEMS");
          check(xsh_grid_add(grid, 0, y, 0.));
          background_par->method="poly";
          lost++;
          break;
      }

      check( x1 = xsh_order_list_eval( orderlist, 
        orderlist->list[i].edguppoly, y));
      check( x2 = xsh_order_list_eval( orderlist, 
        orderlist->list[i-1].edglopoly, y));
      /* compute x */
      x = floor( (x1+x2)/2.0+0.5);
      dx = (x2-x1)/2.0;
      xsh_msg_dbg_high("x %d dx %f", x, dx); 

      if ( xsh_instrument_get_arm(instr) == XSH_ARM_NIR){
        if ( y <= starty || y >= endy){
          if (x >1 && x <= pre->nx){
            check(xsh_grid_add(grid, x, y, 0.0));
          }
          continue;
        }
      }

      if ( (x > (background_par->radius_x+1)) && ((2*background_par->radius_x+1)< dx)){
        /* estimate the median flux */
        check(medflux = xsh_pre_data_window_sample_flux_pa(pre,
							   x-background_par->radius_x, 
                                                           y+oy , 
                                                           2*background_par->radius_x+1,
                                                           sizey, 
                                                           tab,
                                                           background_par->method,
							   background_par->min_frac_grid));
        check( xsh_grid_add(grid, x, y, medflux));
      }
      else{
        lost++;
      }
      /* special case X in ]0, orderN->edgelow] */
      if ( i == (orderlist->size-1)){
        check( x2 = xsh_order_list_eval( orderlist, 
          orderlist->list[i].edglopoly, y));

        x = x2-dx;

        if(dx<0) {
           xsh_msg("Monitor dx=%d x=%g y=%d",x,dx,y);
        }


        if ( (x > (background_par->radius_x+1)) ){
          check(medflux = xsh_pre_data_window_sample_flux_pa(pre,
                                                             x-background_par->radius_x,
                                                             y+oy , 
                                                             2*background_par->radius_x+1, 
                                                             sizey, 
                                                             tab,
                                                             background_par->method,
							     background_par->min_frac_grid));
          check(xsh_grid_add(grid, x, y, medflux));
        }
      }
      else{
        lost++;
      }
      /* special case X in ]order0->edgeup, pre->nx] */ 
      if ( i == 1){
        check( x1 = xsh_order_list_eval( orderlist, 
          orderlist->list[0].edguppoly, y));
        /* special case in VIS */
        if ( xsh_instrument_get_arm(instr) == XSH_ARM_VIS){
          check( x2 = xsh_order_list_eval( orderlist,      
            orderlist->list[0].edglopoly, y));
          x = x1+(x1-x2)*1.2;
        }
        else{
          x=x1+dx;
        }
        if ( (x > (background_par->radius_x+1)) && (x+background_par->radius_x < pre->nx) ){
          check(medflux = xsh_pre_data_window_sample_flux_pa(pre,
                                                             x-background_par->radius_x,
                                                             y+oy ,
                                                             2*background_par->radius_x+1,
                                                             sizey,
                                                             tab,
                                                             background_par->method,
							     background_par->min_frac_grid));
          check(xsh_grid_add(grid, x, y, medflux));
        }
        else {
          lost++;
        }
      } /* if i==1 */
      // }/* check over y: starty<y<endy */
    }/* end loop over i (orders) */
    /* X=pre->nx */
    check(medflux = xsh_pre_data_window_sample_flux_pa(pre,
						       pre->nx-background_par->radius_x, 
						       y+oy, 
						       background_par->radius_x+1, 
						       sizey, 
						       tab,
						       background_par->method,
						       background_par->min_frac_grid));

    check(xsh_grid_add(grid, pre->nx, y, medflux)); 
  }

  xsh_msg_dbg_low("Nb of lost points %d ",lost);
  xsh_msg_dbg_low("Nb of grid points %d ",xsh_grid_get_index(grid));

  /* sort the grid points */
  check( xsh_grid_sort(grid));

  /* generate background_img */
  if(strcmp(background_par->method,"minimum")==0) {
    /* do a fit spline 1d */
    xsh_msg("generate background image: spline");
    check( background_img = cpl_image_new( pre->nx, pre->ny, CPL_TYPE_DOUBLE));
    check(xsh_image_fit_spline(background_img, grid));
  } else if(strcmp(background_par->method,"median")==0) {
    /* do a fit spline 1d */
    xsh_msg("generate background image: spline");
    check( background_img = cpl_image_new( pre->nx, pre->ny, CPL_TYPE_DOUBLE));
    check(xsh_image_fit_spline(background_img, grid));
  } else if (strcmp(background_par->method,"poly")==0) {
    xsh_msg("generate background image: poly");
    check(ima=cpl_image_cast(pre->data,CPL_TYPE_DOUBLE));
    check( background_img = xsh_background_poly( ima, grid));
    xsh_free_image(&ima);
    /*
    check(cpl_image_save(background_img,"background.fits",CPL_BPP_IEEE_FLOAT,
			 NULL,CPL_IO_DEFAULT));
    */
  }

  if( smooth_y>0) {
    check( background_smo = xsh_image_smooth_mean_y( background_img, 
      smooth_y));
    /*
    check(cpl_image_save(background_smo,"smooth_y.fits",CPL_BPP_IEEE_FLOAT,
		       NULL,CPL_IO_DEFAULT));
    */
    xsh_free_image( &background_img);
    background_img= cpl_image_duplicate( background_smo);
    xsh_free_image( &background_smo);
  }

  if( smooth_x>0) {
    check(background_smo = xsh_image_smooth_mean_x( background_img,
      smooth_x));
    /*
    check( cpl_image_save( background_smo, "smooth_x.fits",
      CPL_BPP_IEEE_FLOAT, NULL, CPL_IO_DEFAULT));
    */
    xsh_free_image( &background_img);  
    background_img = cpl_image_duplicate( background_smo);
    xsh_free_image( &background_smo);
  }

 
  XSH_NAME_PREFIX_LAMP_MODE_ARM( pcatg, prefix, "_BACK", "", instr);
  XSH_NAME_PREFIX_LAMP_MODE_ARM( fname, prefix, "_BACK", ".fits", instr);
  plist=cpl_propertylist_new();
  check( xsh_pfits_set_pcatg( plist, pcatg));
 
  check( cpl_image_save( background_img, fname, CPL_BPP_IEEE_FLOAT,
    plist, CPL_IO_DEFAULT));
  xsh_free_propertylist( &plist);

  check( *backg_frame = xsh_frame_product( fname, pcatg, 
    CPL_FRAME_TYPE_IMAGE, CPL_FRAME_GROUP_CALIB,
    CPL_FRAME_LEVEL_FINAL));


  check( cpl_image_subtract( pre->data, background_img));

  XSH_FREE( fname);
  XSH_FREE( pcatg);
  XSH_NAME_PREFIX_LAMP_MODE_ARM( fname, prefix, "_SUB_BACK", ".fits", instr);
  XSH_NAME_PREFIX_LAMP_MODE_ARM( pcatg, prefix, "_SUB_BACK", "", instr);

  check( xsh_pfits_set_pcatg( pre->data_header, pcatg));

  check( result = xsh_pre_save( pre, fname, pcatg, 1));
  check(cpl_frame_set_tag (result, pcatg));
  check(cpl_frame_set_type (result, CPL_FRAME_TYPE_IMAGE));
  check(cpl_frame_set_group (result, CPL_FRAME_GROUP_CALIB));
  check(cpl_frame_set_level (result, CPL_FRAME_LEVEL_FINAL));

  grid_tbl = xsh_grid2table(grid);
 
  XSH_FREE( fname);
  XSH_FREE( pcatg);
  XSH_NAME_PREFIX_LAMP_MODE_ARM( fname, prefix, "_GRID_BACK", ".fits", instr);
  XSH_NAME_PREFIX_LAMP_MODE_ARM( pcatg, prefix, "_GRID_BACK", "", instr);

  plist = cpl_propertylist_new();
  check( xsh_pfits_set_pcatg( plist, pcatg));
  check( cpl_table_save( grid_tbl, plist, NULL, fname, CPL_IO_DEFAULT));
  xsh_free_propertylist( &plist);
  xsh_free_table( &grid_tbl);
  *grid_frame =xsh_frame_product(fname,pcatg,CPL_FRAME_TYPE_TABLE, 
				CPL_FRAME_GROUP_CALIB,CPL_FRAME_LEVEL_FINAL);
 
  cleanup:
    if (cpl_error_get_code () != CPL_ERROR_NONE) {
      xsh_free_frame(&result);
    }
    xsh_free_propertylist(&plist);
    xsh_order_list_free( &orderlist);
    XSH_FREE( tab);
    XSH_FREE( fname);
    XSH_FREE( pcatg);
    xsh_grid_free( &grid);
    xsh_pre_free( &pre);
    xsh_free_image( &background_img);
    return result;
}










/*----------------------------------------------------------------------------*/
/**
   @brief   Subtract the previously defined background 
   @param   image            Image to be background subtracted
   @param   background_im    Background image, may be NULL. If non-NULL,
                             this is updated to contain the flux
                 values that were actually subtracted.
   @param   background_pol   Background polynomial, may be NULL
   @return  CPL_ERROR_NONE iff OK.

   Exactly one of @em background_im and @em background_pol must be non-NULL.
*/
/*----------------------------------------------------------------------------*/
    
static cpl_error_code
subtract_background(cpl_image *image, cpl_image *background_im, 
            const polynomial *background_pol)
{
    int nx, ny;
    int x, y;

    double *image_data;
    double *background_data = NULL;
    passure(image != NULL, " ");
    /* Exactly one of 'background_im' and 'background_pol' must be non-NULL */

    /* For efficiency, don't call cpl_image_get() */
   
    assure(cpl_image_count_rejected(image) == 0, 
       CPL_ERROR_UNSUPPORTED_MODE, "Input image contains bad pixels");
    assure(cpl_image_get_type(image) == CPL_TYPE_DOUBLE,
       CPL_ERROR_UNSUPPORTED_MODE, 
       "Input image is of type %s. double expected", 
        xsh_tostring_cpl_type(cpl_image_get_type(image)));


    image_data = cpl_image_get_data_double(image);
    background_data = cpl_image_get_data_double(background_im);

    nx = cpl_image_get_size_x(image);
    ny = cpl_image_get_size_y(image);

    for (y = 1; y <= ny; y++)
    {
        for (x = 1; x <= nx; x++)
        {
            double back;
            double flux, new_flux;
            
 
	    /* Evaluate at (x,y) */
            back = xsh_polynomial_evaluate_2d(background_pol,x,y);
             
            /* Slow: flux = cpl_image_get(image, x, y, &pis_rejected);  */
            flux = image_data[(x-1) + (y-1) * nx];
            new_flux = flux-back;            
            image_data[(x-1) + (y-1) * nx] = new_flux; 
            background_data[(x-1)+(y-1)*nx]=xsh_max_double(0, flux - new_flux);

        }
    }/* for each pixel... */

  cleanup:
    return cpl_error_get_code();
}



#define DEGX 2
#define DEGY 2
#define KAPPA 4.

static cpl_image*
xsh_background_poly(cpl_image* image, xsh_grid* grid)
{
/*
cpl_table  *t          = NULL;
            t = cpl_table_new( (nx/stepx + 1)*(cpl_table_get_nrow(orders) + 1) );
            cpl_table_new_column(t, "X", CPL_TYPE_INT);
            cpl_table_new_column(t, "Y", CPL_TYPE_INT);
            cpl_table_new_column(t, "Z", CPL_TYPE_DOUBLE);
*/
      
   /* Sampling done. Fit poly. */
   polynomial* background=NULL;
   double mse=0, rmse=0;                   /* mse, rms of fit */
   int total_clipped = 0;
   cpl_table* t=NULL;
   cpl_image* bkg_img=NULL;

   check(t=xsh_grid2table(grid));
   //check(cpl_table_save(t,NULL,NULL,"grid_table.fits",CPL_IO_DEFAULT));

   {
      int n_clipped;
      do {
         assure( cpl_table_get_nrow(t) > (DEGX + 1)*(DEGY + 1), 
                 CPL_ERROR_ILLEGAL_OUTPUT,
                 "Too few sample points available (%d point(s)) to make the fit "
                 "(more than %d points needed). "
                 "Increase number of sample points or increase kappa",
                 cpl_table_get_nrow(t),  (DEGX + 1)*(DEGY + 1));
        
         /* Fit, calculate Zfit */
         xsh_polynomial_delete(&background);
         check_msg( background = xsh_polynomial_regression_2d(
                   t, "X", "Y", "INT", NULL,
                   DEGX, DEGY, "INTfit", NULL, NULL, &mse,
                   NULL, NULL, -1, -1),
                "Error fitting polynomial");
        
         /* Residual := INT - INTfit */
         cpl_table_duplicate_column(t, "Residual", t, "INT");
         cpl_table_subtract_columns(t, "Residual", "INTfit");
        
         /* Compute residuals w.r.t. median of INT 
            (i.e. subtract median(residual) from all residuals),
            then get stdev based on this new mean/median value.
            This is to make kappa sigma clipping more robust */

         cpl_table_subtract_scalar(t, "Residual", 
                                   cpl_table_get_column_median(t, "Residual"));
         rmse = cpl_table_get_column_stdev(t, "Residual");
         xsh_msg_dbg_medium("rmse=%g",rmse);
         /* One-sided kappa-sigma clipping */
         if (KAPPA > 0)
         {
            check_msg( n_clipped = xsh_select_table_rows(
                      t,  "Residual", CPL_GREATER_THAN, KAPPA * rmse),
                   "Error selecting rows");
         }
         else
         {
            n_clipped = 0;
         }
            
         total_clipped += n_clipped;
        
         xsh_msg_dbg_medium("RMS = %f. %d of %d points rejected in kappa-sigma clipping",
                        rmse, n_clipped, cpl_table_get_nrow(t));
            
         cpl_table_erase_selected(t);

         if (n_clipped > 0)
         {
            cpl_table_erase_column(t, "INTfit");
            cpl_table_erase_column(t, "Residual");
         }
        
      } while (n_clipped > 0);
   }


   /* Try to do some quality checking of the background subtraction.
      The number of rejected points (the signal) 
      is often around 10-20 %  */
   {
      double percentage = 
         100.0 * ( (double)total_clipped ) / (total_clipped + cpl_table_get_nrow(t));
        
      if (KAPPA > 0) {
         xsh_msg_dbg_medium("%d of %d points (%.2f %%) were rejected in "
                 "kappa-sigma clipping. RMS = %.2f ADU", 
                 total_clipped,
                 cpl_table_get_nrow(t) + total_clipped,
                 percentage,
                 sqrt(mse));
      }
        
      /* For grid sampling: */
      /* AMO USED HERE ????????????????????????? */
      /*
      if (orders == NULL && order_locations == NULL) 
      {
         if (total_clipped == 0)
         {
            xsh_msg_warning("No points rejected during background "
                            "estimation. Background subtraction is "
                            "uncertain. Try to decrease KAPPA "
                            "(current value is %f)", KAPPA);
         }
         if (percentage > 40)
         {
            xsh_msg_warning("%f %% of the sample points were "
                            "rejected during "
                            "background estimation", percentage);
         }
      }
      */
   }
   bkg_img=cpl_image_duplicate(image);

   check_msg( subtract_background(image,bkg_img, background),
              "Error subtracting background polynomial");

  cleanup:
    xsh_free_table( &t);
    xsh_polynomial_delete(&background);
    return bkg_img;

}



/**@}*/
